
/*

  charset implementation

*/

#include <stdlib.h>
#include <string.h>

#include "video32.h"
#include "charset.h"

extern int goshift;
extern int multiply;

tCharset::tCharset(int nspr,int nchr,int nt16,int nt32)
{
	long int c;

	nSprites=nspr;
	nChars=nchr;
	nTiles16=nt16;
	nTiles32=nt32;

	// allocate graphic charsets memory space

	sprites=new unsigned char* [nSprites];
	chars=new unsigned char* [nChars];
	tiles16=new unsigned char* [nTiles16];
	tiles32=new unsigned char* [nTiles32];

	for (c=0;c<nSprites;c++)
	{
		sprites[c]=new unsigned char [16*16];
		memset(sprites[c],0,16*16);
	}
	for (c=0;c<nChars;c++)
	{
		chars[c]=new unsigned char [8*8];
		memset(chars[c],0,8*8);
	}
	for (c=0;c<nTiles16;c++)
	{
		tiles16[c]=new unsigned char [16*16];
		memset(tiles16[c],0,16*16);
	}
	for (c=0;c<nTiles32;c++)
	{
		tiles32[c]=new unsigned char [32*32];
		memset(tiles32[c],0,32*32);
	}
}

tCharset::~tCharset()
{
	//
	// free decoded charsets
	//

	if (sprites!=0)
	{
		for (int j=0;j<nSprites;j++)
		{
			delete [] sprites[j];
		}

		delete [] sprites;
	}

	if (chars!=0)
	{
		for (int j=0;j<nChars;j++)
		{
			delete [] chars[j];
		}

		delete [] chars;
	}

	if (tiles16!=0)
	{
		for (int j=0;j<nTiles16;j++)
		{
			delete [] tiles16[j];
		}

		delete [] tiles16;
	}

	if (tiles32!=0)
	{
		for (int j=0;j<nTiles32;j++)
		{
			delete [] tiles32[j];
		}

		delete [] tiles32;
	}
}

void tCharset::printSprite(unsigned long int* vb,
							  int charnum,int color,
							  int xp,int yp,
							  int xs,int ys,
							  int flipx,int flipy)
{
	unsigned long int* pbmp=&vb[yp*xs+xp];
	unsigned char* pchar=sprites[charnum];
	unsigned long int* ppal;
	int rx,ry;
	int c;

	for (int y=0;y<16;y++)
	{
		for (int x=0;x<16;x++)
		{
			rx=x;ry=y;

			if (flipx!=0)
			{
				rx=15-x;
			}

			if (flipy!=0)
			{
				ry=15-y;
			}

			c=pchar[rx+ry*16];

			if ( ((xp+x)<0) || ((xp+x)>=xs) || ((yp+y)<0) ) ;
			else if ((yp+y)>=ys) return;
			else if (c!=15)
			{
				ppal=(unsigned long int*)&palette[((c+color*16))*4];
//				*pbmp=(c*16)|((c*16)<<8)|((c*16)<<16);//*ppal;
				*pbmp=*ppal;
			}
			if ((yp+y+1)>=ys) return ;
			pbmp++;
		}
		pbmp+=(xs-16);
	}
}

void tCharset::printChar(unsigned long int* vb,
							  int charnum,int color,
							  int xp,int yp,
							  int xs,int ys,
							  int flipx,int flipy)
{
	unsigned long int* pbmp=&vb[yp*xs+xp];
	unsigned char* pchar=chars[charnum];
	unsigned long int* ppal;
	int rx,ry;
	int c;

	for (int y=0;y<8;y++)
	{
		for (int x=0;x<8;x++)
		{
			rx=x;ry=y;

			if (flipx!=0)
			{
				rx=7-x;
			}

			if (flipy!=0)
			{
				ry=7-y;
			}

			c=pchar[rx+ry*8];

			if ( ((xp+x)<0) || ((xp+x)>=xs) || ((yp+y)<0) ) ;
			else if ((yp+y)>=ys) return;
			else if (c!=15)
			{
				ppal=(unsigned long int*)&palette[(0x200+((c+color*16)))*4];
				*pbmp=*ppal;
//				*pbmp=(c*16)|((c*16)<<8)|((c*16)<<16);
			}
			if ((yp+y+1)>=ys) return ;
			pbmp++;
		}
		pbmp+=(xs-8);
	}
}

void tCharset::printTile16(unsigned long int* vb,
							  int charnum,int color,
							  int xp,int yp,
							  int xs,int ys,
							  int flipx,int flipy)
{
	unsigned long int* pbmp=&vb[yp*xs+xp];
	unsigned char* pchar=tiles16[charnum];
	unsigned long int* ppal;
	int rx,ry;
	int c;

	for (int y=0;y<16;y++)
	{
		for (int x=0;x<16;x++)
		{
			rx=x;ry=y;

			if (flipx!=0)
			{
				rx=15-x;
			}

			if (flipy!=0)
			{
				ry=15-y;
			}

			c=pchar[rx+ry*16];

			if ( ((xp+x)<0) || ((xp+x)>=xs) || ((yp+y)<0) ) ;
			else if ((yp+y)>=ys) return;
			else if (c!=15)
			{
				ppal=(unsigned long int*)&palette[(0x400+((c+color*16)))*4];
				*pbmp=*ppal;
//				*pbmp=(c*16)|((c*16)<<8)|((c*16)<<16);
			}
			if ((yp+y+1)>=ys) return ;
			pbmp++;
		}
		pbmp+=(xs-16);
	}
}

//

void tCharset::printTile32(unsigned long int* vb,
							  int charnum,int color,
							  int xp,int yp,
							  int xs,int ys,
							  int flipx,int flipy)
{
	unsigned long int* pbmp=&vb[yp*xs+xp];
	unsigned char* pchar=tiles32[charnum];
	unsigned long int* ppal;
	int rx,ry;
	int c;

	for (int y=0;y<32;y++)
	{
		for (int x=0;x<32;x++)
		{
			rx=x;ry=y;

			if (flipx!=0)
			{
				rx=31-x;
			}

			if (flipy!=0)
			{
				ry=31-y;
			}

			c=pchar[rx+ry*32];

			if ( ((xp+x)<0) || ((xp+x)>=xs) || ((yp+y)<0) ) ;
			else if ((yp+y)>=ys) return;
			else if (c!=15)
			{
				ppal=(unsigned long int*)&palette[(0x600+((c+color*16)))*4];
//				*pbmp=(c*16)|((c*16)<<8)|((c*16)<<16);
				*pbmp=*ppal;
			}
			if ((yp+y+1)>=ys) return ;
			pbmp++;
		}
		pbmp+=(xs-32);
	}
}
