
/*

	where the fun starts (again)
	7/7/2001

*/


#include "cpucore.h"

// control variables

extern tGui* gui;
extern int do_ss; // do the screenshot
extern int goshift;

static unsigned short int in0=0xffff;
static unsigned short int in1=0xffff;
static unsigned short int in2=0xfffe;

static int charbase,tile16base,tile32base;
static int charpad,tile16pad,tile32pad;
static int layercontrol;

static const int sndbufsize=44100/60;

// memory regions

static unsigned char* rom68k;
static unsigned char* crypt;
static unsigned char* cps1ram;
static unsigned char* cps1eprom;
static unsigned char* qsoundram1;
static unsigned char* output_port;
static unsigned char* cps2output_port;
static unsigned char* gfxram;

static unsigned char* extraram;
static unsigned char* objram;
static unsigned char objectoutput[0x0c];

static short int* sndbuf[2];
static short int* tempsoundbuf;
static short int* realsoundbuf;

// eepro-t
// thanks again to mame team

static struct EEPROM_interface cps2_eeprom_interface =
{
	6,		/* address bits */
	16,		/* data bits */
	"0110",	/*  read command */
	"0101",	/* write command */
	"0111"	/* erase command */
};

// cps2 decrypt

void cps2decrypt()
{
	int i;

	for (i=0;i<0x100000;i++)
	{
		rom68k[i+0x100000]=rom68k[i]^rom68k[i+0x100000];
	}

	memcpy(rom68k,rom68k+0x100000,0x100000);
	memset(rom68k+0x100000,0,0x300000);
}

// cpu init

void cps2cpucore_init(tRomSpace* roms)
{
	sndbuf[0]=new short int [sndbufsize];
	sndbuf[1]=new short int [sndbufsize];

	// stereo sound buffer
	realsoundbuf=new short int [sndbufsize*2];
	tempsoundbuf=new short int [sndbufsize];

	rom68k=new unsigned char [0x400000];

	crypt=new unsigned char [0x400000];

	cps1ram=new unsigned char [0x10000];
	cps1eprom=new unsigned char [0x400];
	qsoundram1=new unsigned char [0x2000];

	extraram=new unsigned char [0x4000];
	objram=new unsigned char [0x8000];

	output_port=new unsigned char [0x100];
	cps2output_port=new unsigned char [0x100];
	
	gfxram=new unsigned char [0x30000];

	// init memory areas

	memset(rom68k,0,0x400000);
	memset(crypt,0,0x400000);

	memcpy(rom68k,roms->romCPU1,0x180000);
	memcpy(crypt,roms->romCPU1,0x180000);
	memset(crypt+0x100000,0,0x300000);

	// decrypt cps2 code
	cps2decrypt();

	memset(objram,0,0x8000);
	memset(objectoutput,0,0x0c);
	memset(gfxram,0,0x30000);
	memset(output_port,0,0x100);
	memset(cps2output_port,0,0x100);
	memset(cps1ram,0,0x10000);
	memset(cps1eprom,0,0x400);
	memset(qsoundram1,0,0x2000);

	// set up values

	layercontrol=roms->layercontrol;
	charbase=roms->charbase;
	tile16base=roms->tile16base;
	tile32base=roms->tile32base;
	charpad=roms->charpad;
	tile16pad=roms->tile16pad;
	tile32pad=roms->tile32pad;

	// init 68000 cpu

	m68k_set_cpu_type(M68K_CPU_TYPE_68000);
	m68k_pulse_reset();

	// init eeprom

	EEPROM_init(&cps2_eeprom_interface);
}

void cps2cpucore_free()
{
	delete [] rom68k;
	delete [] crypt;
	delete [] cps1ram;
	delete [] cps1eprom;
	delete [] qsoundram1;
	delete [] output_port;
	delete [] cps2output_port;
	delete [] gfxram;

	delete [] extraram;
	delete [] objram;

	delete [] sndbuf[0];
	delete [] sndbuf[1];
	delete [] realsoundbuf;
	delete [] tempsoundbuf;
}


void cps2cpucore_emulate(int snd)
{
	int res;

	res=m68k_execute(196666);

	qsoundram1[0x1fff] = 0x77;		// hack to make qsound
									// pass his test
	m68k_set_irq(2);
}

static void update_controls()
{
	if (gui->ginput.coin1==1)
	{
		in2=in2&(~0x1000);
	}
	else
	{
		in2=in2|0x1000;
	}

	if (gui->ginput.start1==1)
	{
		in2=in2&(~0x100);
	}
	else
	{
		in2=in2|0x100;
	}

	if (gui->ginput.p1left==1)
	{
		in0=in0&(~0x02);
	}
	else
	{
		in0=in0|0x02;
	}

	if (gui->ginput.p1right==1)
	{
		in0=in0&(~0x01);
	}
	else
	{
		in0=in0|0x01;
	}

	if (gui->ginput.p1up==1)
	{
		in0=in0&(~0x08);
	}
	else
	{
		in0=in0|0x08;
	}

	if (gui->ginput.p1down==1)
	{
		in0=in0&(~0x04);
	}
	else
	{
		in0=in0|0x04;
	}

	if (gui->ginput.p1button[0]==1)
	{
		in0=in0&(~0x10);
	}
	else
	{
		in0=in0|0x10;
	}

	if (gui->ginput.p1button[1]==1)
	{
		in0=in0&(~0x20);
	}
	else
	{
		in0=in0|0x20;
	}
}

//
// general helper functions
//

__inline int READ_WORD(void *dst)
{
	unsigned char *p = (unsigned char *) dst;
	return (p[0] << 8) | p[1];
}

// special reads/writes

__inline int specREAD8(unsigned char* mem,unsigned long int addr,unsigned long int base)
{
	return mem[(addr^1)-base];
}

__inline int specREAD16(unsigned char* mem,unsigned long int addr,unsigned long int base)
{
	return ((mem[addr-base+1]<<8)|mem[addr-base]);
}

__inline unsigned long int specREAD32(unsigned char* mem,unsigned long int addr,unsigned long int base)
{
	return (
			(mem[addr-base+1]<<24)|
			(mem[addr-base+0]<<16)|
			(mem[addr-base+3]<<8)|
			(mem[addr-base+2])
			);
}

__inline void specWRITE8(unsigned char* mem,unsigned long int addr,unsigned long int base,unsigned long int val)
{
	mem[(addr^1)-base]=(unsigned char)(val>>8);
}

__inline void specWRITE16(unsigned char* mem,unsigned long int addr,unsigned long int base,unsigned long int val)
{
	mem[addr-base+1]=(unsigned char)(val>>8);
	mem[addr-base  ]=(unsigned char)val;
}

__inline void specWRITE32(unsigned char* mem,unsigned long int addr,unsigned long int base,unsigned long int val)
{
	mem[addr-base+1]=(unsigned char)(val>>24);
	mem[addr-base+0]=(unsigned char)(val>>16);
	mem[addr-base+3]=(unsigned char)(val>>8);
	mem[addr-base+2]=(unsigned char)val;
}

//

__inline int READ8(unsigned char* mem,unsigned long int addr,unsigned long int base)
{
	return mem[addr-base];
}

__inline int READ16(unsigned char* mem,unsigned long int addr,unsigned long int base)
{
		return ((mem[addr-base]<<8)|mem[addr-base+1]);
}

__inline unsigned long int READ32(unsigned char* mem,unsigned long int addr,unsigned long int base)
{
		return (
				(mem[addr-base]<<24)|
				(mem[addr-base+1]<<16)|
				(mem[addr-base+2]<<8)|
				(mem[addr-base+3])
				);
}

__inline void WRITE8(unsigned char* mem,unsigned long int addr,unsigned long int base,unsigned long int val)
{
	mem[addr-base]=(unsigned char)val;
}

__inline void WRITE16(unsigned char* mem,unsigned long int addr,unsigned long int base,unsigned long int val)
{
	mem[addr-base]=(unsigned char)(val>>8);
	mem[addr-base+1]=(unsigned char)val;
}

__inline void WRITE32(unsigned char* mem,unsigned long int addr,unsigned long int base,unsigned long int val)
{
	mem[addr-base  ]=(unsigned char)(val>>24);
	mem[addr-base+1]=(unsigned char)(val>>16);
	mem[addr-base+2]=(unsigned char)(val>>8);
	mem[addr-base+3]=(unsigned char)(val);
}

// special writes on output ports

__inline void WRITE8opp(unsigned char* mem,unsigned long int addr,unsigned long int base,unsigned long int val)
{
	mem[(addr^1)-base]=(unsigned char)(val>>8);
}

__inline void WRITE16opp(unsigned char* mem,unsigned long int addr,unsigned long int base,unsigned long int val)
{
	mem[addr-base+1]=(unsigned char)(val>>8);
	mem[addr-base  ]=(unsigned char)val;
}

__inline void WRITE32opp(unsigned char* mem,unsigned long int addr,unsigned long int base,unsigned long int val)
{
	mem[addr-base+1]=(unsigned char)(val>>24);
	mem[addr-base+0]=(unsigned char)(val>>16);
	mem[addr-base+3]=(unsigned char)(val>>8);
	mem[addr-base+2]=(unsigned char)(val);
}

//
// read-write handlers
//

// pc-relativee stuph

unsigned int cps2_m68kri16(unsigned int address)
{
	if ( (address>=0) && (address<=0x3fffff) )
	{
		return specREAD16(rom68k,address,0);
	}

	return 0;
}

unsigned int cps2_m68kri32(unsigned int address)
{
	return ((m68k_read_immediate_16(address) << 16) | m68k_read_immediate_16((address)+2));
}

unsigned int cps2_m68krpc8(unsigned int address)
{
	if ( (address>=0) && (address<=0x3fffff))
	{
		return ((m68k_read_immediate_16(address&~1)>>(8*(1-(address & 1))))&0xff);
	}
	else
	{
		return m68k_read_memory_8(address);
	}
}

unsigned int cps2_m68krpc16(unsigned int address)
{
	if ( (address>=0) && (address<=0x3fffff))
	{
		return m68k_read_immediate_16(address);
	}
	else
	{
		return m68k_read_memory_16(address);
	}
}

unsigned int cps2_m68krpc32(unsigned int address)
{
	if ( (address>=0) && (address<=0x3fffff))
	{
		return m68k_read_immediate_32(address);
	}
	else
	{
		return m68k_read_memory_32(address);
	}
}

unsigned int cps2_m68kr8(unsigned int address)
{
	if ( (address>=0x0) && (address<=0x3fffff) )
	{
		return specREAD8(crypt,address,0);
	}
	else if ( (address>=0x400000) && (address<=0x40000b) )
	{
		return READ8(objectoutput,address,0x400000);
	}
	else if ( (address>=0x618000) && (address<=0x619fff) )
	{
		return READ8(qsoundram1,address,0x618000);
	}
	else if ( (address>=0x660000) && (address<=0x663fff) )
	{
		return READ8(extraram,address,0x660000);
	}
	else if ( (address>=0x900000) && (address<=0x92ffff) )
	{
		return READ8(gfxram,address,0x900000);
	}
	else if (address==0x804001)
	{
		return in0&0xff;
	}
	else if (address==0x804000)
	{
		return ((in0>>8)&0xff);
	}
	else if ( (address>=0x804010) && (address<=0x804011) )
	{
		return in1;
	}
	else if (address==0x804021) 
	{
		unsigned short int res=((in2&0xfffe)| EEPROM_read_bit());
		return res&0xff;
	}
	else if (address==0x804020) 
	{
		unsigned short int res=((in2&0xfffe)| EEPROM_read_bit());
		return ((res>>8)&0xff);
	}
	else if ( (address>=0x804030) && (address<=0x804031) )
	{
	    return 0xe021;
	}
	else if ( (address>=0x8040b0) && (address<=0x8040b3) )
	{
		return 0xffff;
	}
	else if ( (address>=0x708000) && (address<=0x70ffff) )
	{
		return READ8(objram,address,0x708000);
	}
	else if ( (address>=0xff0000) && (address<=0xffffff) )
	{
		return READ8(cps1ram,address,0xff0000);
	}

	return 0;
}

unsigned int cps2_m68kr16(unsigned int address)
{
	if ( (address>=0x0) && (address<=0x3fffff) )
	{
		return specREAD16(crypt,address,0);
	}
	else if ( (address>=0x400000) && (address<=0x40000b) )
	{
		return READ16(objectoutput,address,0x400000);
	}
	else if ( (address>=0x618000) && (address<=0x619fff) )
	{
		return READ16(qsoundram1,address,0x618000);
	}
	else if ( (address>=0x660000) && (address<=0x663fff) )
	{
		return READ16(extraram,address,0x660000);
	}
	else if ( (address>=0x900000) && (address<=0x92ffff) )
	{
		return READ16(gfxram,address,0x900000);
	}
	else if ( (address>=0x804000) && (address<=0x804001) )
	{
		return in0;
	}
	else if ( (address>=0x804010) && (address<=0x804011) )
	{
		return in1;
	}
	else if ( (address>=0x804020) && (address<=0x804021) )
	{
		return ((in2&0xfffe)| EEPROM_read_bit());
	}
	else if ( (address>=0x804030) && (address<=0x804031) )
	{
	    return 0xe021;
	}
	else if ( (address>=0x8040b0) && (address<=0x8040b3) )
	{
		return 0xffff;
	}
	else if ( (address>=0x708000) && (address<=0x70ffff) )
	{
		return READ16(objram,address,0x708000);
	}
	else if ( (address>=0xff0000) && (address<=0xffffff) )
	{
		return READ16(cps1ram,address,0xff0000);
	}

	return 0;
}

unsigned int cps2_m68kr32(unsigned int address)
{
	if ( (address>=0x0) && (address<=0x3fffff) )
	{
		return specREAD32(crypt,address,0);
	}
	else if ( (address>=0x400000) && (address<=0x40000b) )
	{
		return READ32(objectoutput,address,0x400000);
	}
	else if ( (address>=0x618000) && (address<=0x619fff) )
	{
		return READ32(qsoundram1,address,0x618000);
	}
	else if ( (address>=0x660000) && (address<=0x663fff) )
	{
		return READ32(extraram,address,0x660000);
	}
	else if ( (address>=0x900000) && (address<=0x92ffff) )
	{
		return READ32(gfxram,address,0x900000);
	}
/*	else if ( (address>=0x804000) && (address<=0x804001) )
	{
		return in0;
	}
	else if ( (address>=0x804010) && (address<=0x804011) )
	{
		return in1;
	}
	else if ( (address>=0x804020) && (address<=0x804021) )
	{
		return ((in2&0xfffe)| EEPROM_read_bit());
	}
*/	else if ( (address>=0x804030) && (address<=0x804031) )
	{
	    return 0xe021;
	}
	else if ( (address>=0x8040b0) && (address<=0x8040b3) )
	{
		return 0xffff;
	}
	else if ( (address>=0x708000) && (address<=0x70ffff) )
	{
		return READ32(objram,address,0x708000);
	}
	else if ( (address>=0xff0000) && (address<=0xffffff) )
	{
		return READ32(cps1ram,address,0xff0000);
	}

	return 0;
}

void cps2_m68kw8(unsigned int address, unsigned int value)
{
	if ( (address>=0x900000) && (address<=0x92ffff) )
	{
		WRITE8(gfxram,address,0x900000,value);
	}
	else if ( (address>=0x400000) && (address<=0x40000b) )
	{
		WRITE8(objectoutput,address,0x400000,value);
	}
	else if ( (address>=0x618000) && (address<=0x619fff) )
	{
		WRITE8(qsoundram1,address,0x618000,value);
	}
	else if ( (address>=0x660000) && (address<=0x663fff) )
	{
		WRITE8(extraram,address,0x660000,value);
	}
	else if ( (address>=0x800100) && (address<=0x8001ff) )
	{
		WRITE8opp(output_port,address,0x800100,value);
	}
	else if ( (address>=0x804040) && (address<=0x804041) )
	{
        EEPROM_write_bit(value & 0x1000);
        EEPROM_set_clock_line((value & 0x2000) ? ASSERT_LINE : CLEAR_LINE);
        EEPROM_set_cs_line((value & 0x4000) ? CLEAR_LINE : ASSERT_LINE);
	}
	else if ( (address>=0x804100) && (address<=0x8041ff) )
	{
		WRITE8opp(cps2output_port,address,0x804100,value);
	}
	else if ( (address>=0x708000) && (address<=0x70ffff) )
	{
		WRITE8(objram,address,0x708000,value);
	}
	else if ( (address>=0xff0000) && (address<=0xffffff) )
	{
		WRITE8(cps1ram,address,0xff0000,value);
	}
}

void cps2_m68kw16(unsigned int address, unsigned int value)
{
	if ( (address>=0x900000) && (address<=0x92ffff) )
	{
		WRITE16(gfxram,address,0x900000,value);
	}
	else if ( (address>=0x400000) && (address<=0x40000b) )
	{
		WRITE16(objectoutput,address,0x400000,value);
	}
	else if ( (address>=0x618000) && (address<=0x619fff) )
	{
		WRITE16(qsoundram1,address,0x618000,value);
	}
	else if ( (address>=0x660000) && (address<=0x663fff) )
	{
		WRITE16(extraram,address,0x660000,value);
	}
	else if ( (address>=0x800100) && (address<=0x8001ff) )
	{
		WRITE16opp(output_port,address,0x800100,value);
	}
	else if ( (address>=0x804040) && (address<=0x804041) )
	{
        EEPROM_write_bit(value & 0x1000);
        EEPROM_set_clock_line((value & 0x2000) ? ASSERT_LINE : CLEAR_LINE);
        EEPROM_set_cs_line((value & 0x4000) ? CLEAR_LINE : ASSERT_LINE);
	}
	else if ( (address>=0x804100) && (address<=0x8041ff) )
	{
		WRITE16opp(cps2output_port,address,0x804100,value);
	}
	else if ( (address>=0x708000) && (address<=0x70ffff) )
	{
		WRITE16(objram,address,0x708000,value);
	}
	else if ( (address>=0xff0000) && (address<=0xffffff) )
	{
		WRITE16(cps1ram,address,0xff0000,value);
	}
}

void cps2_m68kw32(unsigned int address, unsigned int value)
{
	if ( (address>=0x900000) && (address<=0x92ffff) )
	{
		WRITE32(gfxram,address,0x900000,value);
	}
	else if ( (address>=0x400000) && (address<=0x40000b) )
	{
		WRITE32(objectoutput,address,0x400000,value);
	}
	else if ( (address>=0x618000) && (address<=0x619fff) )
	{
		WRITE32(qsoundram1,address,0x618000,value);
	}
	else if ( (address>=0x660000) && (address<=0x663fff) )
	{
		WRITE32(extraram,address,0x660000,value);
	}
	else if ( (address>=0x800100) && (address<=0x8001ff) )
	{
		WRITE32opp(output_port,address,0x800100,value);
	}
	else if ( (address>=0x804040) && (address<=0x804041) )
	{
        EEPROM_write_bit(value & 0x1000);
        EEPROM_set_clock_line((value & 0x2000) ? ASSERT_LINE : CLEAR_LINE);
        EEPROM_set_cs_line((value & 0x4000) ? CLEAR_LINE : ASSERT_LINE);
	}
	else if ( (address>=0x804100) && (address<=0x8041ff) )
	{
		WRITE32opp(cps2output_port,address,0x804100,value);
	}
	else if ( (address>=0x708000) && (address<=0x70ffff) )
	{
		WRITE32(objram,address,0x708000,value);
	}
	else if ( (address>=0xff0000) && (address<=0xffffff) )
	{
		WRITE32(cps1ram,address,0xff0000,value);
	}
}

//
//
//

static void cpucore_build_palette(tCharset* chset)
{
	// get palette base

	unsigned short int* popp=(unsigned short int*)&output_port[0x0a];
	int base=*popp*256;
	base&= ~(0x2000-1);
	unsigned char* ppal=(unsigned char*)&gfxram[base&0x3ffff];

	int offset;

	for (offset = 0; offset < 32*32*8; offset++)
	{
		int palette = READ_WORD(&ppal[offset*2]);

		int red, green, blue, bright;

		bright= (palette>>12);
		if (bright) bright += 2;

		red   = ((palette>>8)&0x0f) * bright;
		green = ((palette>>4)&0x0f) * bright;
		blue  = (palette&0x0f) * bright;

		// store

		chset->palette[offset*4+2]=red&0xff;
		chset->palette[offset*4+1]=green&0xff;
		chset->palette[offset*4+0]=blue&0xff;
		chset->palette[offset*4+3]=0;
	}
}


/*

  scroll 1

*/

void cps2cpucore_drawScroll1(unsigned long int* video,tCharset* chset)
{
	// base of scroll1 is at output ports 0x02-0x03
	// scroll1 x and y are at 0x0c and 0x0e

	unsigned short int* popbs1=(unsigned short int*)&output_port[0x02];
	unsigned short int* pops1x=(unsigned short int*)&output_port[0x0c];
	unsigned short int* pops1y=(unsigned short int*)&output_port[0x0e];
//	unsigned short int* popbs1=(unsigned short int*)&cps2output_port[0x02];
//	unsigned short int* pops1x=(unsigned short int*)&cps2output_port[0x0c];
//	unsigned short int* pops1y=(unsigned short int*)&cps2output_port[0x0e];

	int x,y, offs, offsx, offsy, sx, sy, ytop;
	int code,realcode,attr,cl,ch;

	int scroll1x=(*pops1x)&0xffff;
	int scroll1y=(*pops1y)&0xffff;

	int base=*popbs1*256;
	base&= ~(0x4000-1);

	unsigned char* c_pgfx=(unsigned char*)&gfxram[base&0x3ffff];

	FILE* ss;
	if (do_ss==1)
	{
		ss=fopen("ss1.bin","wb");
	}

	// draw layer

	int scrlxrough=(scroll1x>>3)+4;
	int scrlyrough=(scroll1y>>3);

	sx=-(scroll1x&0x07)-32;
	ytop=-(scroll1y&0x07)-16;

	for (x=0; x<0x35; x++)
	{
		 sy=ytop;
		 offsx=(scrlxrough+x)*0x80;
		 offsx&=0x1fff;

		 for (y=0; y<0x20; y++)
		 {
			int n=scrlyrough+y;
			offsy=( (n&0x1f)*4 | ((n&0x20)*0x100)) & 0x3fff;
			offs=offsy+offsx;
			offs &= 0x3fff;

			// read code and attrib
			
			code=READ_WORD(&c_pgfx[offs]);
			attr=READ_WORD(&c_pgfx[offs+2]);

			cl =code &0xff;	ch =(code >>8)&0xff;

//			charbase=goshift;
			realcode=(code-charbase)&charpad;

//			if ( (realcode!=0x20) && (code!=0) )
			{
				chset->printChar(video,realcode,attr&0x1f,sx,sy,SCRX,SCRY,attr&0x20,attr&0x40);
			}

			if (do_ss==1)
			{
				fwrite(&ch,1,1,ss); fwrite(&cl,1,1,ss);
			}

			sy+=8;
		 }
		 sx+=8;
	}

	if (do_ss==1)
	{
		fclose(ss);
	}
}

/*

  Scroll 2

*/

void cps2cpucore_drawScroll2(unsigned long int* video,tCharset* chset)
{
	// base of scroll2 is at output ports 0x04-0x05
	// scroll2 x and y are at 0x10 and 0x12

	unsigned short int* popbs2=(unsigned short int*)&output_port[0x04];
	unsigned short int* pops2x=(unsigned short int*)&output_port[0x10];
	unsigned short int* pops2y=(unsigned short int*)&output_port[0x12];

	int base=*popbs2*256;
	base&= ~(0x4000-1);

	unsigned char* c_pgfx=(unsigned char*)&gfxram[base&0x3ffff];

	int scroll2x=*pops2x+64;
	int scroll2y=*pops2y+64-16;

	int sx, sy;
	int nxoffset=(scroll2x&0x0f)+32;    
	int nyoffset=(scroll2y&0x0f);    
	int nx=(scroll2x>>4);	  
	int ny=(scroll2y>>4)-4;	

	FILE* ss;
	if (do_ss==1)
	{
		ss=fopen("ss2.bin","wb");
	}

	for (sx=0; sx<0x32/2+4; sx++)
	{
		for (sy=0; sy<0x09*2-2; sy++)
		{
			int offsy, offsx, offs, colour, code,cl,ch,realcode;
			int n;
			n=ny+sy+2;
			offsy  = ((n&0x0f)*4 | ((n&0x30)*0x100))&0x3fff;
			offsx=((nx+sx)*0x040)&0xfff;
			offs=offsy+offsx;
			offs &= 0x3fff;

			code=  READ_WORD(&c_pgfx[offs]);
			colour=READ_WORD(&c_pgfx[offs+2]);
			cl =code &0xff;	ch =(code >>8)&0xff;

			realcode=(code-tile16base)&tile16pad;

			if (do_ss==1)
			{
				fwrite(&ch,1,1,ss); fwrite(&cl,1,1,ss);
			}

			chset->printTile16(video,realcode,colour&0x1f,
							   (16*sx)-(nxoffset-32),
							   (16*sy)-nyoffset,
							   SCRX,SCRY,colour&0x20,colour&0x40);
		}
	}

	if (do_ss==1)
	{
		fclose(ss);
	}
}

/*

  Spiritezz

*/

void cps2cpucore_drawSprites(unsigned long int* video,tCharset* chset)
{
	unsigned char* c_pgfx=(unsigned char*)&objram[0];
	unsigned short int* base=(unsigned short int*)c_pgfx;

	int cps1_lso;
	int i;
	int offset=6;

	int found=0;

	FILE* ss;
	if (do_ss==1)
	{
		ss=fopen("sss.bin","wb");
	}

	while ( (offset < 0x1000) && (!found) )
	{
		int colour1=READ_WORD(&c_pgfx[offset]);
		int colour2=READ_WORD(&c_pgfx[offset+4]);

		if ( (colour1 == 0xff00) || (colour2==0x8000) )
		{
			cps1_lso=offset-6-8;
			found=1;
		}
		else
		{
			offset+=8;
		}
	}
	if (!found) cps1_lso=0x1000-8;

	int code2;
	int cl,ch;

	int xoffs = 0x20-0x0040;

	for (i=0; i<=cps1_lso; i+=4)
	{
		int x=READ_WORD(&c_pgfx[i]);
		int y=READ_WORD(&c_pgfx[i+2]);

		if (x && y)
		{
			int code=READ_WORD(&c_pgfx[i+4]);
			int colour=READ_WORD(&c_pgfx[i+6]);
			int col=colour&0x1f;

			y &= 0x1ff;
			if (y > 450) y -= 0x200;

			x &= 0x1ff;
			if (x > 450) x -= 0x200;

			x-=0x40;
			y-=0x10;

			if (colour & 0xff00)
			{
				int nx=(colour & 0x0f00) >> 8;
				int ny=(colour & 0xf000) >> 12;
				int nxs,nys,sx,sy;
				nx++;
				ny++;

				if (colour & 0x40)
				{
					if (colour &0x20)
					{
						for (nys=0; nys<ny; nys++)
						{
							for (nxs=0; nxs<nx; nxs++)
							{
								sx = x+nxs*16;
								sy = y+nys*16;
								if (sx > 450) sx -= 0x200;
								if (sy > 450) sy -= 0x200;

								code2=(code+(nx-1)-nxs+0x10*(ny-1-nys));
								cl =code2&0xff;ch=(code2>>8)&0xff;
								if (do_ss==1)
								{
									fwrite(&ch,1,1,ss); fwrite(&cl,1,1,ss);
								}

//								code2+=goshift;

								if (code2<chset->nSprites)
								{
									chset->printSprite(video,
										code2,col,
										sx,sy,SCRX,SCRY,
										1,1);
								}
							}
						}
					}
					else
					{
						for (nys=0; nys<ny; nys++)
						{
							for (nxs=0; nxs<nx; nxs++)
							{
								sx = x+nxs*16;
								sy = y+nys*16;
								if (sx > 450) sx -= 0x200;
								if (sy > 450) sy -= 0x200;

								code2=(code+nxs+0x10*(ny-1-nys));
								cl =code2&0xff;ch=(code2>>8)&0xff;
								if (do_ss==1)
								{
									fwrite(&ch,1,1,ss); fwrite(&cl,1,1,ss);
								}

//								code2+=goshift;

								if (code2<chset->nSprites)
								{
									chset->printSprite(video,
										code2,col,
										sx,sy,SCRX,SCRY,
										0,1);
								}
							}
						}
					}
				}
				else
				{
					if (colour &0x20)
					{
						for (nys=0; nys<ny; nys++)
						{
							for (nxs=0; nxs<nx; nxs++)
							{
								sx = x+nxs*16;
								sy = y+nys*16;
								if (sx > 450) sx -= 0x200;
								if (sy > 450) sy -= 0x200;

								code2=(code+(nx-1)-nxs+0x10*nys);
								cl =code2&0xff;ch=(code2>>8)&0xff;
								if (do_ss==1)
								{
									fwrite(&ch,1,1,ss); fwrite(&cl,1,1,ss);
								}
//								code2+=goshift;

								if (code2<chset->nSprites)
								{
									chset->printSprite(video,
										code2,col,
										sx,sy,SCRX,SCRY,
										1,0);
								}
							}
						}
					}
					else
					{
						for (nys=0; nys<ny; nys++)
						{
							for (nxs=0; nxs<nx; nxs++)
							{
								sx = x+nxs*16;
								sy = y+nys*16;
								if (sx > 450) sx -= 0x200;
								if (sy > 450) sy -= 0x200;

								code2=(code+nxs+0x10*nys);
								cl =code2&0xff;ch=(code2>>8)&0xff;
								if (do_ss==1)
								{
									fwrite(&ch,1,1,ss); fwrite(&cl,1,1,ss);
								}
//								code2+=goshift;

								if (code2<chset->nSprites)
								{
									chset->printSprite(video,
										code2,col,
										sx,sy,SCRX,SCRY,
										0,0);
								}
							}
						}
					}
				}
			}
			else
			{
				if (code<chset->nSprites)
				{
					chset->printSprite(video,code,col,
						x,y,SCRX,SCRY,
						colour&0x20,colour&0x40);
				}
			}
		}

	}

	if (do_ss==1)
	{
		fclose(ss);
	}
}

/* main draw function */

void cps2cpucore_drawScreen(unsigned long int* video,tCharset* chset)
{
	update_controls();
	cpucore_build_palette(chset);
	cps2cpucore_drawScroll2(video,chset);
	cps2cpucore_drawSprites(video,chset);
	cps2cpucore_drawScroll1(video,chset);
}