
/*

  gui implementation

*/

#include "gui.h"


tGui::tGui(int s,int gxs,int gys,char* f8x8name)
{
	chset=0;
	exitp=0;
	state=s;
	emulating=0;
	guiXsize=gxs;
	guiYsize=gys;

	// reset controls

	ginput.mouseXpos=0;
	ginput.mouseYpos=0;
	ginput.mousebutton1=0;
	ginput.mousebutton2=0;

	for (int i=0;i<6;i++)
	{
		ginput.p1button[i]=0;
		ginput.p2button[i]=0;
	}

	ginput.coin1=ginput.coin2=0;
	ginput.start1=ginput.start2=0;
	ginput.p1up=ginput.p2up=0;
	ginput.p1down=ginput.p2down=0;
	ginput.p1left=ginput.p2left=0;
	ginput.p1right=ginput.p2right=0;

	//
	// init emulation related things
	//

	// store font

	font8x8=new tFont(64,8,8,0xcecece,f8x8name);

	// add components
	
	compList.resize(3);

	compList[0]=new tBackground(0x151535,guiXsize,guiYsize);
	compList[1]=new tSysbar(0x202050,0x9292c2,12,guiXsize,guiYsize);
	compList[2]=new tMenu(0x9292c2,guiXsize,guiYsize);

	// create first menu items

	tMenuItem mi1("File",	1,0x9292c2,-1);
	tMenuItem mi2("Display",2,0x9292c2,-1);
	tMenuItem mi3("Cpu",	3,0x9292c2,-1);
	tMenuItem mi4("About",	-1,0x9292c2,1);
	tMenuItem mi5("Exit",	-1,0x9292c2,2);

	tMenuItem mi6("Open",-1,0x9292c2,3);
	tMenuItem mi7("Free",-1,0x9292c2,4);

	tMenuItem mi8("Windows GDI",-1,0x9292c2,5);
	tMenuItem mi9("DX7 windowed",-1,0x9292c2,6);
	tMenuItem mi10("DX7 fullscreen",-1,0x9292c2,7);

	tMenuItem mi11("Reset",-1,0x9292c2,8);
	tMenuItem mi12("Pause",-1,0x9292c2,9);
	tMenuItem mi13("Save state",-1,0x9292c2,10);
	tMenuItem mi14("Load state",-1,0x9292c2,11);

	tSubMenu sm1(2,210,0x202050,guiXsize,guiYsize,0);
	tSubMenu sm2(75,174,0x202050,guiXsize,guiYsize,0);
	tSubMenu sm3(75,174,0x202050,guiXsize,guiYsize,0);
	tSubMenu sm4(75,210,0x202050,guiXsize,guiYsize,0);

	sm1.addItem(mi1);
	sm1.addItem(mi2);
	sm1.addItem(mi3);
	sm1.addItem(mi4);
	sm1.addItem(mi5);

	sm2.addItem(mi6);
	sm2.addItem(mi7);

	sm3.addItem(mi8);
	sm3.addItem(mi9);
	sm3.addItem(mi10);

	sm4.addItem(mi11);
	sm4.addItem(mi12);
	sm4.addItem(mi13);
	sm4.addItem(mi14);

	tMenu* mp=(tMenu*)compList[2];
	mp->addItem(sm1);
	mp->addItem(sm2);
	mp->addItem(sm3);
	mp->addItem(sm4);
}

//
// open a rom and start emulation
// (returns 1 if any problem occurs)
//

int tGui::openRom()
{
	BOOL result;
	OPENFILENAME ofn;
	char szFile[2048];      

	memset(szFile,0,2048);

	// Initialize OPENFILENAME
	ZeroMemory(&ofn, sizeof(OPENFILENAME));
	ofn.lStructSize = sizeof(OPENFILENAME);
	ofn.hwndOwner = 0;
	ofn.lpstrFile = szFile;
	ofn.nMaxFile = sizeof(szFile);
	ofn.lpstrFilter = "*.ZIP\0";
	ofn.nFilterIndex = 0;
	ofn.lpstrFileTitle = NULL;
	ofn.nMaxFileTitle = 0;
	ofn.lpstrInitialDir = NULL;
	ofn.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST;

	result=GetOpenFileName(&ofn);

	if (result==FALSE)
	{
		// no rom selected
		return 0;
	}

	// otherwise open the selected rom
	// and start emulation

	if (romset.readRoms(ofn.lpstrFile))
	{
		MessageBox(NULL,"can't read roms","error",MB_OK);
		return 1;
	}

	// set romset type

	romsettype=romset.romsettype;

	// choose which cpucore to use
	// (will switch this to virtual functions asap)

	if (romsettype==ROMSET_CPS1)
	{
		m68kri16=cps1_m68kri16;
		m68kri32=cps1_m68kri32;
		m68krpc8=cps1_m68krpc8;
		m68krpc16=cps1_m68krpc16;
		m68krpc32=cps1_m68krpc32;
		m68kr8=cps1_m68kr8;
		m68kr16=cps1_m68kr16;
		m68kr32=cps1_m68kr32;
		m68kw8=cps1_m68kw8;
		m68kw16=cps1_m68kw16;
		m68kw32=cps1_m68kw32;

		// decode charset
		chset=new tCharset(romset.nspr,romset.nchars,romset.ntile16,romset.ntile32);
		romset.decodeCharset(chset,romsettype);

		// init cpucore
		cps1cpucore_init(&romset);
	}
	else if (romsettype==ROMSET_CPS2)
	{
		m68kri16=cps2_m68kri16;
		m68kri32=cps2_m68kri32;
		m68krpc8=cps2_m68krpc8;
		m68krpc16=cps2_m68krpc16;
		m68krpc32=cps2_m68krpc32;
		m68kr8=cps2_m68kr8;
		m68kr16=cps2_m68kr16;
		m68kr32=cps2_m68kr32;
		m68kw8=cps2_m68kw8;
		m68kw16=cps2_m68kw16;
		m68kw32=cps2_m68kw32;

		// decode charset
		chset=new tCharset(romset.nspr,romset.nchars,romset.ntile16,romset.ntile32);
		romset.decodeCharset(chset,romsettype);

		// init cpucore
		cps2cpucore_init(&romset);
	}

	emulating=1;
	return 0;
}

void tGui::redraw(unsigned long int* vbuf)
{
	if (state==1)
	{
		for (int i=0;i<compList.size();i++)
		{
			if (compList[i]->isActive()!=0)
			{
				compList[i]->draw(vbuf,font8x8,&ginput);
			}
		}
	}
}

void tGui::update()
{
	if (state==1)
	{
		// update components
		// basing on user input and mouse movements

		for (int i=0;i<compList.size();i++)
		{
			if (compList[i]->isActive())
			{
				compList[i]->update(&ginput);
			}

			if (i==1)		// sysbar...
			{
				tSysbar* sbp=(tSysbar*)compList[1];

				if (sbp->menuclicked==1)
				{
					compList[2]->active=!compList[2]->active;
				}
			}

			if (i==2)		// sysmenu
			{
				tMenu* mp=(tMenu*)compList[2];
				int act;

				act=mp->getAction();

				// exit
				if (act==2)
				{
					exitp=1;
					mp->deactivate();
				}
				else if (act==3)
				{
					openRom();
					mp->deactivate();
				}

				mp->resetAction();
			}
		}
	}
}

int tGui::exitprogram()
{
	return exitp;
}

tGui::~tGui()
{
	for (int i=0;i<compList.size();i++)
	{
		delete(compList[i]);
	}

	delete(font8x8);

	if (chset!=0) 
	{
		delete chset;
	}
}

//
// user input
//

void tGui::mouseMove(int x,int y)
{
	ginput.mouseXpos=x;
	ginput.mouseYpos=y;
}

void tGui::lButtonDown()
{
	ginput.mousebutton1=1;
}

void tGui::lButtonUp()
{
	ginput.mousebutton1=0;
}

void tGui::rButtonDown()
{
	ginput.mousebutton2=1;

	// the right button should activate/deactivate the gui

	state=!state;
}

void tGui::rButtonUp()
{
	ginput.mousebutton2=0;
}

