 
/*

  cps1 emulation core.

*/

#include "cpucore.h"


// control variables

extern tGui* gui;
extern int do_ss; // do the screenshot

static int in0=0xff;	 // 0 input port
static int in1=0xffff;	 // player controls 
static int in2=0xffff;	 // player controls 
static int drawlayer[3]; // draw layers?

static unsigned long int protport,protval;
static unsigned long int fac1,fac2,multlo,multhi;
static int charbase,tile16base,tile32base;
static int charpad,tile16pad,tile32pad;
static int layercontrol;

static int sound_latch=0;
static int z80sndreg=0;
static int sndfade=0;
static int sndbufsize;
static OKIM6295interface okim;
static int snddivider=2;
static short int* psndbuf[2];

// memory regions

static unsigned char* rom68k;
static unsigned char* cps1ram;
static unsigned char* cps1eprom;
static unsigned char* qsoundram1;
static unsigned char* output_port;
static unsigned char* gfxram;

static unsigned char* romadpcm;

static unsigned char* z80rom;
static unsigned char* z80prg;

static short int* sndbuf[2];
static short int* tempsoundbuf;
static short int* realsoundbuf;
static short int* okimsoundbuf[2];

// z80

static CONTEXTMZ80 ctx1;

// z80 read/write functions

unsigned char z80rd(unsigned long int A,struct MemoryReadByte* mwb)
{
	if ( (A>=0xd000) && (A<0xd800) )
	{
		return z80prg[A];
	}
	else if (A==0xf001)
	{
		return YM2151ReadStatus(0);
	}
	else if (A==0xf002)
	{
		// OKI status
		return OKIM6295_status_r(0);
	}
	else if (A==0xf008)
	{
		return sound_latch;
	}
	else if (A==0xf00a)
	{
		return sndfade;
	}

	return 0;
}


void z80wr(unsigned long int A,unsigned char V,struct MemoryWriteByte* mwb)
{
	if ( (A>=0xd000) && (A<0xd800) )
	{
		z80prg[A]=V;
	}
	else if (A==0xf000)
	{
		z80sndreg=V;
	}
	else if (A==0xf001)
	{
		YM2151WriteReg(0,z80sndreg,V);
	}
	else if (A==0xf002)
	{
		// OKI data write
		OKIM6295_data_w(0,V);
	}
	else if (A==0xf004)
	{
		// set bank

		int bank=(V*0x4000)&0x7fff;

		memcpy(z80prg+0x8000,z80rom+0x8000+bank,0x4000);
	}
}

// z80 handlers

struct z80PortRead z80ReadPorts[] =
{
	{-1,-1,0}
};

struct z80PortWrite z80WritePorts[] =
{
	{-1,-1,0}
};

struct MemoryReadByte z80readmem[] =
{
	{0xc000,0xffff,z80rd,0},
	{-1,-1,0,0}
};
  
struct MemoryWriteByte z80writemem[] =
{
	{0x0000,0xffff,z80wr,0},
	{-1,-1,0,0}
};

// ym2151 interrupt handler

static void ym2151irq()
{
	mz80SetContext(&ctx1);
	mz80int(0x38);
	mz80GetContext(&ctx1);
}

// cpu init

void cps1cpucore_init(tRomSpace* roms)
{
	int i;

	sndbufsize=gui->guiinfo.soundfreq/60;

	z80prg=new unsigned char [0x10000];
	z80rom=new unsigned char [0x10000];

	sndbuf[0]=new short int [sndbufsize];
	sndbuf[1]=new short int [sndbufsize];

	// stereo sound buffer
	realsoundbuf=new short int [sndbufsize*2];
	tempsoundbuf=new short int [sndbufsize];

	okimsoundbuf[0]=new short int [sndbufsize];
	okimsoundbuf[1]=new short int [sndbufsize];

	psndbuf[0]=new short int [sndbufsize*snddivider];
	psndbuf[1]=new short int [sndbufsize*snddivider];

	// memory space
	rom68k=new unsigned char [0x200000];
	cps1ram=new unsigned char [0x10000];
	cps1eprom=new unsigned char [0x400];
	qsoundram1=new unsigned char [0x2000];
	output_port=new unsigned char [0x100];
	gfxram=new unsigned char [0x30000];

	romadpcm=new unsigned char [0x40000];

	// init memory areas

	memcpy(rom68k,roms->romCPU1,0x200000);
	memcpy(z80rom,roms->romCPU2,0x10000);

	memset(z80prg,0,0x10000);
	memcpy(z80prg,z80rom,0xc000);

	memset(gfxram,0,0x30000);
	memset(output_port,0,0x100);
	memset(cps1ram,0,0x10000);
	memset(cps1eprom,0,0x400);

	memcpy(romadpcm,roms->romadpcm,0x40000);

	// set up values

	protport=roms->protport;
	protval=roms->protval;
	multlo=roms->multlo;
	multhi=roms->multhi;
	fac1=roms->fac1;
	fac2=roms->fac2;
	layercontrol=roms->layercontrol;
	charbase=roms->charbase;
	tile16base=roms->tile16base;
	tile32base=roms->tile32base;
	charpad=roms->charpad;
	tile16pad=roms->tile16pad;
	tile32pad=roms->tile32pad;

	for (i=0;i<3;i++)
	{
		drawlayer[i]=1;
	}

	// init 68000 cpu

	m68k_set_cpu_type(M68K_CPU_TYPE_68000);
	m68k_pulse_reset();

	// init z80

	mz80init();

	ctx1.z80Base=z80prg;
	ctx1.z80IoRead=z80ReadPorts;
	ctx1.z80IoWrite=z80WritePorts;
	ctx1.z80MemRead=z80readmem;
	ctx1.z80MemWrite=z80writemem;

	mz80SetContext(&ctx1);
	mz80reset();
	mz80GetContext(&ctx1);

	// init ym2151
	YM2151Init(1,3579580/2,gui->guiinfo.soundfreq,16);
	YM2151SetIrqHandler(0,ym2151irq);

	// init OKIM

	okim.num=1;
	okim.frequency[0]=7576;
	OKIM6295_sh_start(&okim,romadpcm,gui->guiinfo.soundfreq);
}


void cps1cpucore_free()
{
	delete [] rom68k;
	delete [] z80rom;
	delete [] cps1ram;
	delete [] cps1eprom;
	delete [] qsoundram1;
	delete [] output_port;
	delete [] gfxram;
	delete [] z80prg;

	delete [] romadpcm;

	delete [] sndbuf[0];
	delete [] sndbuf[1];
	delete [] realsoundbuf;
	delete [] tempsoundbuf;
	delete [] okimsoundbuf[0];
	delete [] okimsoundbuf[1];
	delete [] psndbuf[0];
	delete [] psndbuf[1];

	YM2151Shutdown();
	OKIM6295_sh_stop();
}


void cps1cpucore_emulate(int snd)
{
	int i;
	int res;
	unsigned long int dwres;

	if (!snd)
	{
		res=m68k_execute(166666);
		m68k_set_irq(2);
	}
	else
	{
		int pos=0;

		for (i=0;i<snddivider;i++)
		{
			res=m68k_execute(166666/snddivider);

			mz80SetContext(&ctx1);
			dwres=mz80exec(66666/snddivider);	
			if (dwres!=0x80000000)
			{
				MessageBox(NULL,"z80 bad op","error",MB_OK);
			}
			mz80GetContext(&ctx1);

			YM2151UpdateOne(0,sndbuf,sndbufsize);

			for (int p=0;p<sndbufsize;p++)
			{
				psndbuf[0][pos]=sndbuf[0][p];
				psndbuf[1][pos]=sndbuf[1][p];
				pos++;
			}
		}

		OKIM6295_sh_update(okimsoundbuf,sndbufsize);

		for (i=0;i<sndbufsize;i++)
		{
			realsoundbuf[i*2]=(psndbuf[0][i*snddivider]+okimsoundbuf[0][i])/2;
			realsoundbuf[i*2+1]=(psndbuf[1][i*snddivider]+okimsoundbuf[1][i])/2;
		}

		feedBuffer(realsoundbuf,sndbufsize*2);

		m68k_set_irq(2);
	}

}


//
// emulation functions
//

static void update_controls()
{
	if (gui->ginput.coin1==1)
	{
		in0=in0&(~1);
	}
	else
	{
		in0=in0|1;
	}

	if (gui->ginput.start1==1)
	{
		in0=in0&(~0x10);
	}
	else
	{
		in0=in0|0x10;
	}

	if (gui->ginput.p1left==1)
	{
		in1=in1&(~0x02);
	}
	else
	{
		in1=in1|0x02;
	}

	if (gui->ginput.p1right==1)
	{
		in1=in1&(~0x01);
	}
	else
	{
		in1=in1|0x01;
	}

	if (gui->ginput.p1up==1)
	{
		in1=in1&(~0x08);
	}
	else
	{
		in1=in1|0x08;
	}

	if (gui->ginput.p1down==1)
	{
		in1=in1&(~0x04);
	}
	else
	{
		in1=in1|0x04;
	}

	if (gui->ginput.p1button[0]==1)
	{
		in1=in1&(~0x10);
	}
	else
	{
		in1=in1|0x10;
	}

	if (gui->ginput.p1button[1]==1)
	{
		in1=in1&(~0x20);
	}
	else
	{
		in1=in1|0x20;
	}

	if (gui->ginput.p1button[2]==1)
	{
		in1=in1&(~0x40);
	}
	else
	{
		in1=in1|0x40;
	}

	if (gui->ginput.p1button[3]==1)
	{
		in2=in2&(~0x01);
	}
	else
	{
		in2=in2|0x01;
	}

	if (gui->ginput.p1button[4]==1)
	{
		in2=in2&(~0x02);
	}
	else
	{
		in2=in2|0x02;
	}

	if (gui->ginput.p1button[5]==1)
	{
		in2=in2&(~0x04);
	}
	else
	{
		in2=in2|0x04;
	}

	// system keys

	for (int i=0;i<3;i++)
	{
		if (gui->ginput.layeron[i]==1)
		{
			drawlayer[i]=!drawlayer[i];
			gui->ginput.layeron[i]=0;	//ack
		}
	}

}

static __inline int read_player_controls_in1()
{
	return in1;
}

static __inline int read_player_controls_in2()
{
	return in2;
}

static __inline int cps1_input_r(int port)
{
	int sm=0x00;
//	int sm=0x44;		// test mode

	switch (port/2)
	{
		case (0x00):
			{
				return in0&(~sm);		//IN0
				break;
			}
		case (0x01):
			{
				return 0xff;
				break;
			}
		case (0x02):
			{
				return 0xff;
				break;
			}
		case (0x03):
			{
				return in0&(~sm);		//IN0
				break;
			}
		case (0x04):
			{
				return 0xff;
				break;
			}
		return 0xff;
	}

	return 0xff;
}

static __inline int cps1_multiply(unsigned long int offset)
{
	// fac1,fac2,multlo,multhi
	// 0x4e,0x4c,0x4a,0x48

	if (offset == multlo)	
	{
		unsigned short int* pop4e=(unsigned short int*)&output_port[fac1];
		unsigned short int* pop4c=(unsigned short int*)&output_port[fac2];

		return ((*pop4e)*(*pop4c))&0xffff;
	}

	if (offset == multhi)	
	{
		unsigned short int* pop4e=(unsigned short int*)&output_port[fac1];
		unsigned short int* pop4c=(unsigned short int*)&output_port[fac2];

		return ((*pop4e)*(*pop4c))>>16;
	}

	return 0;
}

//

//
// general helper functions
//

__inline int READ_WORD(void *dst)
{
	unsigned char *p = (unsigned char *) dst;
	return (p[0] << 8) | p[1];
}

// special reads/writes

__inline int specREAD8(unsigned char* mem,unsigned long int addr,unsigned long int base)
{
	return mem[(addr^1)-base];
}

__inline int specREAD16(unsigned char* mem,unsigned long int addr,unsigned long int base)
{
	unsigned short int* pmem=(unsigned short int*)&mem[addr-base];
	return (*pmem);
}

__inline unsigned long int specREAD32(unsigned char* mem,unsigned long int addr,unsigned long int base)
{
	return (
			(mem[addr-base+1]<<24)|
			(mem[addr-base+0]<<16)|
			(mem[addr-base+3]<<8)|
			(mem[addr-base+2])
			);
}

//

__inline int READ8(unsigned char* mem,unsigned long int addr,unsigned long int base)
{
	return mem[addr-base];
}

__inline int READ16(unsigned char* mem,unsigned long int addr,unsigned long int base)
{
	return ((mem[addr-base]<<8)|mem[addr-base+1]);
}

__inline unsigned long int READ32(unsigned char* mem,unsigned long int addr,unsigned long int base)
{
	return (
			(mem[addr-base]<<24)|
			(mem[addr-base+1]<<16)|
			(mem[addr-base+2]<<8)|
			(mem[addr-base+3])
			);

}

__inline void WRITE8(unsigned char* mem,unsigned long int addr,unsigned long int base,unsigned long int val)
{
	mem[addr-base]=(unsigned char)val;
}

__inline void WRITE16(unsigned char* mem,unsigned long int addr,unsigned long int base,unsigned long int val)
{
	mem[addr-base]=(unsigned char)(val>>8);
	mem[addr-base+1]=(unsigned char)val;
}

__inline void WRITE32(unsigned char* mem,unsigned long int addr,unsigned long int base,unsigned long int val)
{
	unsigned char* pmem=&mem[addr-base];

	*pmem=(val>>24);pmem++;
	*pmem=(val>>16);pmem++;
	*pmem=(val>>8); pmem++;
	*pmem=val;
}

// special writes on output ports

__inline void WRITE8opp(unsigned char* mem,unsigned long int addr,unsigned long int base,unsigned long int val)
{
	mem[(addr^1)-base]=(unsigned char)(val>>8);
}

__inline void WRITE16opp(unsigned char* mem,unsigned long int addr,unsigned long int base,unsigned long int val)
{
	unsigned short int* pmem=(unsigned short int*)&mem[addr-base];
	*pmem=val;
}

__inline void WRITE32opp(unsigned char* mem,unsigned long int addr,unsigned long int base,unsigned long int val)
{
	mem[addr-base+1]=(unsigned char)(val>>24);
	mem[addr-base+0]=(unsigned char)(val>>16);
	mem[addr-base+3]=(unsigned char)(val>>8);
	mem[addr-base+2]=(unsigned char)(val);
}

//
// read-write handlers
//

unsigned int cps1_m68kri16(unsigned int address)
{
	return cps1_m68kr16(address);
}

unsigned int cps1_m68kri32(unsigned int address)
{
	return cps1_m68kr32(address);
}

unsigned int cps1_m68krpc8(unsigned int address)
{
	return cps1_m68kr8(address);
}

unsigned int cps1_m68krpc16(unsigned int address)
{
	return cps1_m68kr16(address);
}

unsigned int cps1_m68krpc32(unsigned int address)
{
	return cps1_m68kr32(address);
}

unsigned int  cps1_m68kr8(unsigned int address)
{
	if (address<=0x1fffff)
	{
		return specREAD8(rom68k,address,0);
	}
	else if (address>=0xff0000)
	{
		return READ8(cps1ram,address,0xff0000);
	}
	else if ( (address>=0x900000) && (address<=0x92ffff) )
	{
		return READ8(gfxram,address,0x900000);
	}
	else if ( (address>=0x800000) && (address<=0x800001) )
	{
		return read_player_controls_in1();
	}
	else if ( (address>=0x800010) && (address<=0x800011) )
	{
		return read_player_controls_in1();
	}
	else if ( (address>=0x800176) && (address<=0x800177) )
	{
		return read_player_controls_in2();
	}
	else if ( (address>=0x800018) && (address<=0x80001f) )
	{
		return (cps1_input_r(address-0x800018));
	}

	return 0;
}

unsigned int cps1_m68kr16(unsigned int address)
{
	if (address<=0x1fffff)
	{
		return specREAD16(rom68k,address,0);
	}
	else if (address>=0xff0000)
	{
		return READ16(cps1ram,address,0xff0000);
	}
	else if ( (address>=0x900000) && (address<=0x92ffff) )
	{
		return READ16(gfxram,address,0x900000);
	}
	else if ( (address>=0x800000) && (address<=0x800001) )
	{
		return read_player_controls_in1();
	}
	else if ( (address>=0x800010) && (address<=0x800011) )
	{
		return read_player_controls_in1();
	}
	else if ( (address>=0x800176) && (address<=0x800177) )
	{
		return read_player_controls_in2();
	}
	else if ( (address>=0x800018) && (address<=0x80001f) )
	{
		return (cps1_input_r(address-0x800018));
	}
	else if (address==multlo)
	{
		return cps1_multiply(multlo);
	}
	else if (address==multhi)
	{
		return cps1_multiply(multhi);
	}
	else if (address==protport)
	{
		return protval;
	}

	return 0;
}

unsigned int cps1_m68kr32(unsigned int address)
{
	if  (address<=0x1fffff)
	{
		return specREAD32(rom68k,address,0);
	}
	else if (address>=0xff0000)
	{
		return READ32(cps1ram,address,0xff0000);
	}
	else if ( (address>=0x900000) && (address<=0x92ffff) )
	{
		return READ32(gfxram,address,0x900000);
	}
	else if ( (address>=0x800000) && (address<=0x800001) )
	{
		return read_player_controls_in1();
	}
	else if ( (address>=0x800010) && (address<=0x800011) )
	{
		return read_player_controls_in1();
	}
	else if ( (address>=0x800176) && (address<=0x800177) )
	{
		return read_player_controls_in2();
	}
	else if ( (address>=0x800018) && (address<=0x80001f) )
	{
		return (cps1_input_r(address-0x800018));
	}

	return 0;
}

void cps1_m68kw8(unsigned int address, unsigned int value)
{
	if (address>=0xff0000)
	{
		WRITE8(cps1ram,address,0xff0000,value);
	}
	else if ( (address>=0x900000) && (address<=0x92ffff) )
	{
		WRITE8(gfxram,address,0x900000,value);
	}
	else if ( (address>=0x800180) && (address<=0x800181) )
	{
		sound_latch=value&0xff;
	}
	else if ( (address>=0x800188) && (address<=0x800189) )
	{
		sndfade=value&0xff;
	}
	else if ( (address>=0x800100) && (address<=0x8001ff) )
	{
		WRITE8opp(output_port,address,0x800100,value);
	}
}

void cps1_m68kw16(unsigned int address, unsigned int value)
{
	if (address>=0xff0000)
	{
		WRITE16(cps1ram,address,0xff0000,value);
	}
	else if ( (address>=0x900000) && (address<=0x92ffff) )
	{
		WRITE16(gfxram,address,0x900000,value);
	}
	else if ( (address>=0x800180) && (address<=0x800181) )
	{
		sound_latch=value&0xff;
	}
	else if ( (address>=0x800188) && (address<=0x800189) )
	{
		sndfade=value&0xff;
	}
	else if ( (address>=0x800100) && (address<=0x8001ff) )
	{
		WRITE16opp(output_port,address,0x800100,value);
	}
}

void cps1_m68kw32(unsigned int address, unsigned int value)
{
	if (address>=0xff0000)
	{
		WRITE32(cps1ram,address,0xff0000,value);
	}
	else if ( (address>=0x900000) && (address<=0x92ffff) )
	{
		WRITE32(gfxram,address,0x900000,value);
	}
	else if ( (address>=0x800180) && (address<=0x800181) )
	{
		sound_latch=value&0xff;
	}
	else if ( (address>=0x800188) && (address<=0x800189) )
	{
		sndfade=value&0xff;
	}
	else if ( (address>=0x800100) && (address<=0x8001ff) )
	{
		WRITE32opp(output_port,address,0x800100,value);
	}
}

/*

  palette builder

*/

void cps1cpucore_build_palette(tCharset* chset)
{
	// get palette base

	unsigned short int* popp=(unsigned short int*)&output_port[0x0a];
	int base=*popp*256;
	base&= ~(0x2000-1);
	unsigned char* ppal=(unsigned char*)&gfxram[base&0x3ffff];

	int offset;

	for (offset = 0; offset < 32*32*8; offset++)
	{
		int palette = READ_WORD(&ppal[offset*2]);

		int red, green, blue, bright;

		bright= (palette>>12);
		if (bright) bright += 2;

		red   = ((palette>>8)&0x0f) * bright;
		green = ((palette>>4)&0x0f) * bright;
		blue  = (palette&0x0f) * bright;

		// store

		chset->palette[offset*4+2]=red&0xff;
		chset->palette[offset*4+1]=green&0xff;
		chset->palette[offset*4+0]=blue&0xff;
		chset->palette[offset*4+3]=0;
	}
}

/*

  scroll 1

*/

void cps1cpucore_drawScroll1(unsigned long int* video,tCharset* chset)
{
	// base of scroll1 is at output ports 0x02-0x03
	// scroll1 x and y are at 0x0c and 0x0e

	unsigned short int* popbs1=(unsigned short int*)&output_port[0x02];
	unsigned short int* pops1x=(unsigned short int*)&output_port[0x0c];
	unsigned short int* pops1y=(unsigned short int*)&output_port[0x0e];

	int x,y, offs, offsx, offsy, sx, sy, ytop;
	int code,realcode,attr,cl,ch;

	int scroll1x=(*pops1x)&0xffff;
	int scroll1y=(*pops1y)&0xffff;

	int base=*popbs1*256;
	base&= ~(0x4000-1);

	unsigned char* c_pgfx=(unsigned char*)&gfxram[base&0x3ffff];

	FILE* ss;
	if (do_ss==1)
	{
		ss=fopen("ss1.bin","wb");
	}

	// draw layer

	int scrlxrough=(scroll1x>>3)+4;
	int scrlyrough=(scroll1y>>3);

	sx=-(scroll1x&0x07)-32;
	ytop=-(scroll1y&0x07)-16;

	for (x=0; x<53; x++)
	{
		 sy=ytop;
		 offsx=(scrlxrough+x)*0x80;
		 offsx&=0x1fff;

		 for (y=0; y<30; y++)
		 {
			int n=scrlyrough+y;
			offsy=( (n&0x1f)*4 | ((n&0x20)*0x100)) & 0x3fff;
			offs=offsy+offsx;
			offs &= 0x3fff;

			// read code and attrib
			
			code=READ_WORD(&c_pgfx[offs]);
			attr=READ_WORD(&c_pgfx[offs+2]);

			realcode=(code-charbase)&charpad;

			if ( (realcode!=0x20) && (code!=0) )
			{
				chset->printChar(video,realcode,attr&0x1f,sx,sy,attr&0x20,attr&0x40);
			}

			if (do_ss==1)
			{
				cl =code &0xff;	ch =(code >>8)&0xff;
				fwrite(&ch,1,1,ss); fwrite(&cl,1,1,ss);
			}

			sy+=8;
		 }
		 sx+=8;
	}

	if (do_ss==1)
	{
		fclose(ss);
	}
}

/*

  Scroll 2 (normal and rowscrolled)

*/

void cps1cpucore_drawScroll2(unsigned long int* video,tCharset* chset)
{
	// base of scroll2 is at output ports 0x04-0x05
	// scroll2 x and y are at 0x10 and 0x12

	unsigned short int* popbs2=(unsigned short int*)&output_port[0x04];
	unsigned short int* pops2x=(unsigned short int*)&output_port[0x10];
	unsigned short int* pops2y=(unsigned short int*)&output_port[0x12];

	int base=*popbs2*256;
	base&= ~(0x4000-1);

	unsigned char* c_pgfx=(unsigned char*)&gfxram[base&0x3ffff];

	int scroll2x=*pops2x+64;
	int scroll2y=*pops2y+64-16;

	int sx, sy;
	int nxoffset=(scroll2x&0x0f)+32;    
	int nyoffset=(scroll2y&0x0f);    
	int nx=(scroll2x>>4);	  
	int ny=(scroll2y>>4)-4;	

	FILE* ss;
	if (do_ss==1)
	{
		ss=fopen("ss2.bin","wb");
	}

	for (sx=0; sx<0x32/2+4; sx++)
	{
		for (sy=0; sy<0x09*2-2; sy++)
		{
			int offsy, offsx, offs, colour, code,cl,ch,realcode;
			int n;
			n=ny+sy+2;
			offsy  = ((n&0x0f)*4 | ((n&0x30)*0x100))&0x3fff;
			offsx=((nx+sx)*0x040)&0xfff;
			offs=offsy+offsx;
			offs &= 0x3fff;

			code=  READ_WORD(&c_pgfx[offs]);
			colour=READ_WORD(&c_pgfx[offs+2]);
			cl =code &0xff;	ch =(code >>8)&0xff;

			realcode=(code-tile16base)&tile16pad;

			if (do_ss==1)
			{
				fwrite(&ch,1,1,ss); fwrite(&cl,1,1,ss);
			}

			chset->printTile16(video,realcode,colour&0x1f,
							   (16*sx)-(nxoffset-32),
							   (16*sy)-nyoffset,
							   colour&0x20,colour&0x40);
		}
	}

	if (do_ss==1)
	{
		fclose(ss);
	}
}

void cps1cpucore_drawScroll2rs(unsigned long int* video,tCharset* chset)
{
	// base of scroll2 is at output ports 0x04-0x05
	// scroll2 x and y are at 0x10 and 0x12

	unsigned short int* popbs2=(unsigned short int*)&output_port[0x04];
	unsigned short int* pops2x=(unsigned short int*)&output_port[0x10];
	unsigned short int* pops2y=(unsigned short int*)&output_port[0x12];

	int base=*popbs2*256;
	base&= ~(0x4000-1);

	unsigned char* c_pgfx=(unsigned char*)&gfxram[base&0x3ffff];

	// we should get rowscroll ram base (0x08)
	// and rowscroll offset (0x20)

	unsigned short int* prowscroll=(unsigned short int*)&output_port[0x08];
	base=*prowscroll*256;
	base&=~(0x800-1);
	unsigned char* prs=(unsigned char*)&gfxram[base&0x3ffff];

	unsigned short int* prowscrolloffs=(unsigned short int*)&output_port[0x20];
	int rowscrolloffs=*prowscrolloffs;

	//

	int scroll2x=*pops2x+64;
	int scroll2y=*pops2y;

	int sx, sy;
	int nxoffset=(scroll2x&0x0f)+32;    
	int nyoffset=(scroll2y&0x0f);    
	int nx=(scroll2x>>4);	  
	int ny=(scroll2y>>4)-4;	

	//

	int scrollx[1024];
	int scrly=-scroll2y;


	for (int i=0;i<256 ;i++)
	{
		scrollx[(i - scrly) & 0x3ff] = (( -(READ_WORD(&prs[(2*(i + rowscrolloffs)) & 0x7ff]))));
	}

	//

	FILE* ss;
	if (do_ss==1)
	{
		ss=fopen("ss2.bin","wb");
	}

	for (sx=0x2a; sx>=0; sx--)
	{
		int n=ny;
		for (sy=0x12-1; sy>=0; sy--)
		{
			int realcode,ch,cl;
			int offsy, offsx, offs, colour, code;

			n=ny+sy+2;
			offsy  = ((n&0x0f)*4 | ((n&0x30)*0x100))&0x3fff;
			offsx=((nx+sx)*0x040)&0xfff;
			offs=offsy+offsx;
			offs &= 0x3fff;

			code=  READ_WORD(&c_pgfx[offs]);
			colour=READ_WORD(&c_pgfx[offs+2]);

			realcode=(code-tile16base)&tile16pad;

			if (do_ss==1)
			{
				cl =code &0xff;	ch =(code >>8)&0xff;
				fwrite(&ch,1,1,ss); fwrite(&cl,1,1,ss);
			}

			chset->printTile16rs(video,realcode,colour&0x1f,
							   (16*sx)+32-nxoffset,
							   (16*sy)-nyoffset-48,
							   colour&0x20,colour&0x40,scrollx,(-scrly+0x10)&0x3ff);
		}
	}

	if (do_ss==1)
	{
		fclose(ss);
	}
}

/*

  Scroll 3

*/

void cps1cpucore_drawScroll3(unsigned long int* video,tCharset* chset)
{
	// base of scroll3 is at output ports 0x06-0x07
	// scroll3 x and y are at 0x14 and 0x16

	unsigned short int* popbs3=(unsigned short int*)&output_port[0x06];
	unsigned short int* pops3x=(unsigned short int*)&output_port[0x14];
	unsigned short int* pops3y=(unsigned short int*)&output_port[0x16];

	int base=(*popbs3)*256;
	base&= ~(0x4000-1);

	unsigned char* c_pgfx=(unsigned char*)&gfxram[base&0x3ffff];

	int scroll3x=*pops3x+32;
	int scroll3y=*pops3y+32;

	int sx,sy;
	int nxoffset=scroll3x&0x1f;
	int nyoffset=scroll3y&0x1f;
	int nx=(scroll3x>>5)+1;
	int ny=(scroll3y>>5)-1;

	FILE* ss;
	if (do_ss==1)
	{
		ss=fopen("ss3.bin","wb");
	}

	for (sx=0; sx<15; sx++)
	{
		for (sy=0; sy<9; sy++)
		{
			int offsy, offsx, offs, colour, code,realcode;
			int n;
			n=ny+sy;
			offsy  = ((n&0x07)*4 | ((n&0xf8)*0x0100))&0x3fff;
			offsx=((nx+sx)*0x020)&0x7ff;
			offs=offsy+offsx;
			offs &= 0x3fff;

			colour=READ_WORD(&c_pgfx[offs+2]);
			code=READ_WORD(&c_pgfx[offs]);

			if (do_ss==1)
			{
				int cl,ch;

				cl=code&0xff; 
				ch=(code>>8)&0xff;

				fwrite(&ch,1,1,ss); fwrite(&cl,1,1,ss);
			}

//			if (code!=0)
			{
				realcode=(code-tile32base)&tile32pad;

				chset->printTile32(video,realcode,colour&0x1f,
								   32*sx-nxoffset,
								   32*sy-16-nyoffset,
								   colour&0x20,colour&0x40);
			}
		}
	}

	if (do_ss==1)
	{
		fclose(ss);
	}
}

/*

  Spiritezz

*/

void cps1cpucore_drawSprites(unsigned long int* video,tCharset* chset)
{
	// obj ram base
	unsigned short int* popbss=(unsigned short int*)&output_port[0x00];

	int base=*popbss*256;
	base&= ~(0x800-1);

	unsigned char* c_pgfx=(unsigned char*)&gfxram[base&0x3ffff];

	int cps1_lso;
	int i;
	int offset=6;

	int found=0;

	while ( (offset < 0x4000) && (!found) )
	{
		int colour=READ_WORD(&c_pgfx[offset]);
		if (colour == 0xff00)
		{
			cps1_lso=offset-6-8;
			found=1;
		}
		else
		{
			offset+=8;
		}
	}
	if (!found) cps1_lso=0x4000-8;

	// draw (new)

	int code2;

	for (i=cps1_lso; i>=0; i-=8)
	{
		int x=READ_WORD(&c_pgfx[i]);
		int y=READ_WORD(&c_pgfx[i+2]);

		if (x && y)
		{
			int code=READ_WORD(&c_pgfx[i+4]);
			int colour=READ_WORD(&c_pgfx[i+6]);
			int col=colour&0x1f;

			// 3wonders
			//if (code>=0x2a00) code+=0x4000;

			y &= 0x1ff;
			if (y > 450) y -= 0x200;

			x &= 0x1ff;
			if (x > 450) x -= 0x200;

			x-=0x40;
			y-=0x10;

			if (colour & 0xff00)
			{
				int nx=(colour & 0x0f00) >> 8;
				int ny=(colour & 0xf000) >> 12;
				int nxs,nys,sx,sy;
				nx++;
				ny++;

				if (colour & 0x40)
				{
					if (colour &0x20)
					{
						for (nys=0; nys<ny; nys++)
						{
							for (nxs=0; nxs<nx; nxs++)
							{
								sx = x+nxs*16;
								sy = y+nys*16;
								if (sx > 450) sx -= 0x200;
								if (sy > 450) sy -= 0x200;

								code2=(code+(nx-1)-nxs+0x10*(ny-1-nys));
								
								if (code2<chset->nSprites)
								{
									chset->printSprite(video,
										code2,col,
										sx,sy,
										1,1);
								}
							}
						}
					}
					else
					{
						for (nys=0; nys<ny; nys++)
						{
							for (nxs=0; nxs<nx; nxs++)
							{
								sx = x+nxs*16;
								sy = y+nys*16;
								if (sx > 450) sx -= 0x200;
								if (sy > 450) sy -= 0x200;

								code2=(code+nxs+0x10*(ny-1-nys));

								if (code2<chset->nSprites)
								{
									chset->printSprite(video,
										code2,col,
										sx,sy,
										0,1);
								}
							}
						}
					}
				}
				else
				{
					if (colour &0x20)
					{
						for (nys=0; nys<ny; nys++)
						{
							for (nxs=0; nxs<nx; nxs++)
							{
								sx = x+nxs*16;
								sy = y+nys*16;
								if (sx > 450) sx -= 0x200;
								if (sy > 450) sy -= 0x200;

								code2=(code+(nx-1)-nxs+0x10*nys);

								if (code2<chset->nSprites)
								{
									chset->printSprite(video,
										code2,col,
										sx,sy,
										1,0);
								}
							}
						}
					}
					else
					{
						for (nys=0; nys<ny; nys++)
						{
							for (nxs=0; nxs<nx; nxs++)
							{
								sx = x+nxs*16;
								sy = y+nys*16;
								if (sx > 450) sx -= 0x200;
								if (sy > 450) sy -= 0x200;

								code2=(code+nxs+0x10*nys);

								if (code2<chset->nSprites)
								{
									chset->printSprite(video,
										code2,col,
										sx,sy,
										0,0);
								}
							}
						}
					}
				}
			}
			else
			{
				if (code<chset->nSprites)
				{
					chset->printSprite(video,code,col,
						x,y,
						colour&0x20,colour&0x40);
				}
			}
		}

	}
}

/* helper draw function */

static void draw_layer(int l,unsigned long int* video,tCharset* chset)
{
	if (l==0)
	{
		cps1cpucore_drawSprites(video,chset);
	}
	else if (l==1)
	{
		if (drawlayer[0])
		{
			cps1cpucore_drawScroll1(video,chset);
		}
	}
	else if (l==2)
	{
		if (drawlayer[1])
		{
			unsigned short int* prson=(unsigned short int*)&output_port[0x22];
			int rson=*prson&1;

			if (rson)
			{
				cps1cpucore_drawScroll2rs(video,chset);
			}
			else
			{
				cps1cpucore_drawScroll2(video,chset);
			}
		}
	}
	else if (l==3)
	{
		if (drawlayer[2])
		{
			cps1cpucore_drawScroll3(video,chset);
		}
	}
}


/* main draw function */

void cps1cpucore_drawScreen(unsigned long int* video,tCharset* chset)
{
	// layer control states what's the layer to draw first
	// and in which order the others

	int l0,l1,l2,l3;
	unsigned short int* lcp=(unsigned short int*)&output_port[layercontrol];

	update_controls();

	cps1cpucore_build_palette(chset);

	l0=((*lcp)>>0x06)&3;
	l1=((*lcp)>>0x08)&3;
	l2=((*lcp)>>0x0a)&3;
	l3=((*lcp)>>0x0c)&3;

	draw_layer(l0,video,chset);
	draw_layer(l1,video,chset);
	draw_layer(l2,video,chset);
	draw_layer(l3,video,chset);
}
