
/*

  gui implementation

*/

#include "gui.h"


tGui::tGui(int s,int gxs,int gys,char* inifile)
{
	int i;

	chset=0;
	exitp=0;
	state=s;
	emulating=0;
	guiXsize=gxs;
	guiYsize=gys;

	// reset controls

	ginput.mouseXpos=0;
	ginput.mouseYpos=0;
	ginput.mousebutton1=0;
	ginput.mousebutton2=0;

	for (i=0;i<6;i++)
	{
		ginput.p1button[i]=0;
		ginput.p2button[i]=0;
	}

	ginput.coin1=ginput.coin2=0;
	ginput.start1=ginput.start2=0;
	ginput.p1up=ginput.p2up=0;
	ginput.p1down=ginput.p2down=0;
	ginput.p1left=ginput.p2left=0;
	ginput.p1right=ginput.p2right=0;

	for (i=0;i<3;i++)
	{
		ginput.layeron[i]=0;
	}

	//
	// init emulation related things
	//

	// inifile

	readIniFile(inifile);

	// store font

	font8x8=new tFont(64,8,8,0xcecece);

	// add components

	int ds=guiinfo.doublesize;
	
	compList.resize(3);

	compList[0]=new tBackground(0x151535,guiXsize,guiYsize);
	compList[1]=new tSysbar(0x202050,0x9292c2,12,guiXsize,guiYsize,ds);
	compList[2]=new tMenu(0x9292c2,guiXsize,guiYsize);

	// create first menu items

	tMenuItem mi1("File",	1,0x9292c2,-1);
	tMenuItem mi2("Display",2,0x9292c2,-1);
	tMenuItem mi3("Cpu",	3,0x9292c2,-1);
	tMenuItem mi4("About",	-1,0x9292c2,1);
	tMenuItem mi5("Exit",	-1,0x9292c2,2);

	tMenuItem mi6("Open",-1,0x9292c2,3);
	tMenuItem mi7("Free",-1,0x9292c2,4);

	tMenuItem mi8("Windows GDI",-1,0x9292c2,5);
	tMenuItem mi9("DX7 windowed",-1,0x9292c2,6);
	tMenuItem mi10("DX7 fullscreen",-1,0x9292c2,7);

	tMenuItem mi11("Reset",-1,0x9292c2,8);
	tMenuItem mi12("Pause",-1,0x9292c2,9);
	tMenuItem mi13("Save state",-1,0x9292c2,10);
	tMenuItem mi14("Load state",-1,0x9292c2,11);

	tSubMenu sm1(2,210,0x202050,guiXsize,guiYsize,0,ds);
	tSubMenu sm2(75,174,0x202050,guiXsize,guiYsize,0,ds);
	tSubMenu sm3(75,174,0x202050,guiXsize,guiYsize,0,ds);
	tSubMenu sm4(75,210,0x202050,guiXsize,guiYsize,0,ds);

	sm1.addItem(mi1);
	sm1.addItem(mi2);
	sm1.addItem(mi3);
	sm1.addItem(mi4);
	sm1.addItem(mi5);

	sm2.addItem(mi6);
	sm2.addItem(mi7);

	sm3.addItem(mi8);
	sm3.addItem(mi9);
	sm3.addItem(mi10);

	sm4.addItem(mi11);
	sm4.addItem(mi12);
	sm4.addItem(mi13);
	sm4.addItem(mi14);

	tMenu* mp=(tMenu*)compList[2];
	mp->addItem(sm1);
	mp->addItem(sm2);
	mp->addItem(sm3);
	mp->addItem(sm4);
}

//
// open a rom and start emulation
// (returns 1 if any problem occurs)
//

int tGui::openRom()
{
	BOOL result;
	OPENFILENAME ofn;
	char szFile[2048];      

	memset(szFile,0,2048);

	// Initialize OPENFILENAME
	ZeroMemory(&ofn, sizeof(OPENFILENAME));
	ofn.lStructSize = sizeof(OPENFILENAME);
	ofn.hwndOwner = 0;
	ofn.lpstrFile = szFile;
	ofn.nMaxFile = sizeof(szFile);
	ofn.lpstrFilter = "*.ZIP\0";
	ofn.nFilterIndex = 0;
	ofn.lpstrFileTitle = NULL;
	ofn.nMaxFileTitle = 0;
	ofn.lpstrInitialDir = NULL;
	ofn.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST;

	result=GetOpenFileName(&ofn);

	if (result==FALSE)
	{
		// no rom selected
		return 0;
	}

	// otherwise open the selected rom
	// and start emulation

	if (romset.readRoms(ofn.lpstrFile))
	{
		MessageBox(NULL,"can't read roms","error",MB_OK);
		return 1;
	}

	// set romset type

	romsettype=romset.romsettype;

	// choose which cpucore to use
	// (will switch this to virtual functions asap)

	if (romsettype==ROMSET_CPS1)
	{
		m68kri16=cps1_m68kri16;
		m68kri32=cps1_m68kri32;
		m68krpc8=cps1_m68krpc8;
		m68krpc16=cps1_m68krpc16;
		m68krpc32=cps1_m68krpc32;
		m68kr8=cps1_m68kr8;
		m68kr16=cps1_m68kr16;
		m68kr32=cps1_m68kr32;
		m68kw8=cps1_m68kw8;
		m68kw16=cps1_m68kw16;
		m68kw32=cps1_m68kw32;

		// decode charset
		chset=new tCharset(romset.nspr,romset.nchars,romset.ntile16,romset.ntile32);
		romset.decodeCharset(chset,romsettype);

		// init cpucore
		cps1cpucore_init(&romset);
	}
	else if (romsettype==ROMSET_CPS2)
	{
		m68kri16=cps2_m68kri16;
		m68kri32=cps2_m68kri32;
		m68krpc8=cps2_m68krpc8;
		m68krpc16=cps2_m68krpc16;
		m68krpc32=cps2_m68krpc32;
		m68kr8=cps2_m68kr8;
		m68kr16=cps2_m68kr16;
		m68kr32=cps2_m68kr32;
		m68kw8=cps2_m68kw8;
		m68kw16=cps2_m68kw16;
		m68kw32=cps2_m68kw32;

		// decode charset
		chset=new tCharset(romset.nspr,romset.nchars,romset.ntile16,romset.ntile32);
		romset.decodeCharset(chset,romsettype);

		// init cpucore
		cps2cpucore_init(&romset);
	}

	emulating=1;
	start_count();	//start counting seconds of emulation

	return 0;
}

void tGui::redraw(unsigned long int* vbuf)
{
	if (state==1)
	{
		for (int i=0;i<compList.size();i++)
		{
			if (compList[i]->isActive()!=0)
			{
				compList[i]->draw(vbuf,font8x8,&ginput);
			}
		}
	}
}

void tGui::update()
{
	if (state==1)
	{
		// update components
		// basing on user input and mouse movements

		for (int i=0;i<compList.size();i++)
		{
			if (compList[i]->isActive())
			{
				compList[i]->update(&ginput);
			}

			if (i==1)		// sysbar...
			{
				tSysbar* sbp=(tSysbar*)compList[1];

				if (sbp->menuclicked==1)
				{
					compList[2]->active=!compList[2]->active;
				}
			}

			if (i==2)		// sysmenu
			{
				tMenu* mp=(tMenu*)compList[2];
				int act;

				act=mp->getAction();

				// exit
				if (act==2)
				{
					exitp=1;
					mp->deactivate();
				}
				else if (act==3)
				{
					openRom();
					mp->deactivate();
				}

				mp->resetAction();
			}
		}
	}
}

int tGui::exitprogram()
{
	return exitp;
}

tGui::~tGui()
{
	for (int i=0;i<compList.size();i++)
	{
		delete(compList[i]);
	}

	delete(font8x8);

	if (chset!=0) 
	{
		delete chset;
	}
}

/*

  functions to deal with .ini files

*/

void tGui::readIniFile(char* inifile)
{
	int err;
	char pval[256];
	int ipval;

	err=getParameter(inifile,"[sound]","soundon",pval);
	if (err!=0)
	{
		MessageBox(0,"Error: can't read soundon var. from .ini file","error",MB_OK);
		return ;
	}
	else
	{
		ipval=atoi(pval);
		if (ipval) guiinfo.soundon=1;
		else guiinfo.soundon=0;
	}

	//

	err=getParameter(inifile,"[sound]","soundfreq",pval);
	if (err!=0)
	{
		MessageBox(0,"Error: can't read soundfreq var. from .ini file","error",MB_OK);
		return ;
	}
	else
	{
		ipval=atoi(pval);
		guiinfo.soundfreq=ipval;
	}

	//
	err=getParameter(inifile,"[video]","frameskip",pval);
	if (err!=0)
	{
		MessageBox(0,"Error: can't read frameskip var. from .ini file","error",MB_OK);
		return ;
	}
	else
	{
		ipval=atoi(pval);
		guiinfo.frameskip=ipval;
		if ((ipval<=0) || (ipval>20)) guiinfo.frameskip=1;
	}

	//
	err=getParameter(inifile,"[video]","doublesize",pval);
	if (err!=0)
	{
		MessageBox(0,"Error: can't read doublesize var. from .ini file","error",MB_OK);
		return ;
	}
	else
	{
		ipval=atoi(pval);
		guiinfo.doublesize=ipval;
	}

	//

	err=getParameter(inifile,"[video]","trimspeed",pval);
	if (err!=0)
	{
		MessageBox(0,"Error: can't read trimspeed var. from .ini file","error",MB_OK);
		return ;
	}
	else
	{
		ipval=atoi(pval);
		guiinfo.trimspeed=ipval;
	}
}

int tGui::ltrim(char* str)
{
    int found=0;
    char* pstr=str;

    while (found==0)
    {
        if (*pstr==' ')
        {
            pstr++;
        }
        else found=1;
    }

    strcpy(str,pstr);

    return 0;
}

int tGui::rtrim(char* str)
{
    unsigned int sl;
    unsigned int pos;
    int found=0;

    sl=strlen(str);
    if (sl==0) return 0;

    pos=sl-1;

    while ( (found==0) && (pos>0) )
    {
        if ( (str[pos]==' ') || (str[pos]==0x0a) || (str[pos]==0x0d) )
        {
            str[pos]='\0';
            pos--;
        }
        else
        {
            found=1;
        }
    }

    return 0;
}

int tGui::getParameter(char* filename,char* sectionname,
                       char* paramname,char* paramvalue)
{
    FILE* inifile;
    int sectionFound=0;
    int paramFound=0;
    int goOut=0;
    
    if ( (filename==NULL) || (sectionname==NULL) || (paramname==NULL) || (paramvalue==NULL) ) return 1;

    inifile=fopen(filename,"rb");
    if (inifile==NULL) return 1;

    /* scan the file until you don't find a matching section or until it ends */
    while (!goOut)
    {
        char line[256];
        char* bufptr;
        
        bufptr=(char*)fgets((char*)line,256,inifile);
        if (bufptr==NULL)
        {
            goOut=1;
        }
        else
        {
            /* rtrim and ltrim the line found */
            ltrim(line);
            rtrim(line);
                
            /* check if it's a comment */
            if (line[0]==';') ;
            else
            {
                /* check if it's a section and matches */
                if (!strcmp(line,sectionname))
                {
                    sectionFound=1;
                    /* printf("trovata la sezione %s\n",line); */
                    goOut=1;
                }
            }
        }
    }

    if (sectionFound!=0)
    {
        goOut=0;
        
        while (!goOut)
        {
            /*
               scan the inifile until:
               1) it ends
               2) you find "[" (another section)
               3) you find the parameter
            */
        
            char line[256];
            char* bufptr;
        
            bufptr=(char*)fgets(line,256,inifile);
            if (bufptr==NULL)
            {
                goOut=1;
                /* printf("eof?\n"); */
            }
            else
            {
                /* rtrim and ltrim the line found */
                ltrim(line);
                rtrim(line);
                
                /* printf("line trimmed\n"); */
                
                /* check if it's a comment */
                if (line[0]==';') ;
                if (line[0]=='[')
                {
                    goOut=1;
                }
                else
                {
                    /* check if there's a parameter and if it matches */
                    char lvalue[256];
                    char rvalue[256];
                    char* eqptr;
                
                    eqptr=(char*)strstr(line,"=");
                    if (eqptr==NULL)
                    {
                        /* skip the line (it would be a syntax error really) */
                    }
                    else
                    {
                        strcpy(rvalue,eqptr+1);
                        *eqptr='\0';
                        strcpy(lvalue,line);
                        
                        ltrim(lvalue); rtrim(lvalue);
                        ltrim(rvalue); rtrim(rvalue);
                        
                        /* printf("rtrimmed\n"); */
                        /* printf("paramname:%s lvalue:%s\n",paramname,lvalue); */
                        
                        /* is it our parameter? */
                        if (!strcmp(paramname,lvalue))
                        {
                            /* allocate paramvalue */
                        
                            strcpy(paramvalue,rvalue);
                            goOut=1;
                            paramFound=1;
                        }
                    }
                }
            }
        
        }
    }

    fclose(inifile);
    return 0;
}

//
// user input
//

void tGui::mouseMove(int x,int y)
{
	ginput.mouseXpos=x;
	ginput.mouseYpos=y;
}

void tGui::lButtonDown()
{
	ginput.mousebutton1=1;
}

void tGui::lButtonUp()
{
	ginput.mousebutton1=0;
}

void tGui::rButtonDown()
{
	ginput.mousebutton2=1;

	// the right button should activate/deactivate the gui

	state=!state;
}

void tGui::rButtonUp()
{
	ginput.mousebutton2=0;
}

