
/*

  functions to deal with blitting,windows....

*/


#include "video32.h"
#include "gui.h"

extern tGui* gui;

/*

  local variables

*/

static HWND hwin;
static RECT g_rcScreen;
static HINSTANCE hinst; 
static int g_videomode;
static int oldresX,oldresY,oldresBPP;

static unsigned long int* addbuf;

static LPDIRECTDRAW7 g_pDD = NULL;	//ddraw obj
static LPDIRECTDRAWSURFACE7 g_pDDSPrimary = NULL;// DirectDraw primary surface
static LPDIRECTDRAWSURFACE7 g_pDDSBack = NULL;   // DirectDraw back surface

static const char* appName="a.c.e. v1.2c - CPS1 and CPS2 emulator";

/*

  videoput32 - blit onto the screen

*/

void videoput32(unsigned long int* buf)
{
	if (g_videomode==0)
	{
		DDSURFACEDESC2              ddsd;
		DDSURFACEDESC2              ddsd2;
		HRESULT hres;
		unsigned long int* pBits;
		unsigned long int* pSrc=buf;

		ddsd.dwSize = sizeof(ddsd);
		ddsd2.dwSize = sizeof(ddsd2);

		g_pDDSBack->GetSurfaceDesc(&ddsd2);
		hres=g_pDDSBack->Lock(NULL,&ddsd,DDLOCK_WAIT,NULL);
		if (hres==DD_OK)
		{
			pBits=(unsigned long int*)ddsd.lpSurface;
			int ptc=ddsd2.lPitch/4;
			unsigned long int* ppbits;

			for (int i=0;i<SCRY;i++)
			{
				int yv=ptc*i;
				ppbits=&pBits[yv];

				memcpy(ppbits,pSrc,SCRX*4);
				pSrc+=SCRX;
			}
		
			g_pDDSBack->Unlock(NULL);
		}

		// flip
		g_pDDSPrimary->Flip(0,0);
	}
	if (g_videomode==1)
	{
		DDSURFACEDESC2              ddsd;
		DDSURFACEDESC2              ddsd2;
		HRESULT hres;
		unsigned long int* pSrc=buf;

		ddsd.dwSize = sizeof(ddsd);
		ddsd2.dwSize = sizeof(ddsd2);

		g_pDDSBack->GetSurfaceDesc(&ddsd2);
		hres=g_pDDSBack->Lock(NULL,&ddsd,DDLOCK_WAIT,NULL);

		if (oldresBPP==16)
		{
			if (hres==DD_OK)
			{
				unsigned short int* pBits=(unsigned short int*)ddsd.lpSurface;
				unsigned char* pps=(unsigned char*)pSrc;

				for (int i=0;i<SCRY;i++)
				{
					unsigned short int* ppBits=&pBits[i*ddsd2.lPitch/2];
					for (int l=0;l<SCRX;l++)
					{
						int r,g,b;
						unsigned short int fcol;

						r=pps[0]>>3;
						g=(pps[1]>>2)<<5;
						b=(pps[2]>>3)<<11;

						fcol=r|g|b;

						*ppBits=fcol;
						ppBits++;
						pSrc++;
						pps+=4;
					}
				}
			
				g_pDDSBack->Unlock(NULL);
			}
		}
		else if (oldresBPP==32)
		{
			if (hres==DD_OK)
			{
				unsigned long int* pBits=(unsigned long int*)ddsd.lpSurface;
				int ptch=ddsd2.lPitch/4;
				int yoff=0;
				unsigned long int* ppbits;

				for (int i=0;i<SCRY;i++)
				{
					ppbits=&pBits[yoff];

					memcpy(ppbits,pSrc,SCRX*4);
					pSrc+=SCRX;
					yoff+=ptch;
				}
			
				g_pDDSBack->Unlock(NULL);
			}
		}


		// flip screen
		hres=g_pDDSPrimary->Blt(&g_rcScreen,g_pDDSBack,0,DDBLT_WAIT,NULL);
	}
	if (g_videomode==2)
	{
		// copy the source into addbuf
		memcpy(addbuf,buf,4*SCRX*SCRY);

		char bibuf[sizeof (BITMAPINFOHEADER)+12];
		BITMAPINFO &bi=*(BITMAPINFO*)&bibuf;
		BITMAPINFOHEADER &bih=bi.bmiHeader;
		bih.biSize=sizeof(bih);
		bih.biWidth=SCRX;
		bih.biHeight=-SCRY;
		bih.biPlanes=1;
		bih.biBitCount=32;
		bih.biCompression=BI_BITFIELDS;
		bih.biSizeImage=0;
		bih.biXPelsPerMeter=0;
		bih.biYPelsPerMeter=0;
		bih.biClrUsed=0;
		bih.biClrImportant=0;
		((unsigned long*)bi.bmiColors)[0]=0x00FF0000;
		((unsigned long*)bi.bmiColors)[1]=0x0000FF00;
		((unsigned long*)bi.bmiColors)[2]=0x000000FF;

		HDC dc=GetDC(hwin);
		RECT r;
		GetClientRect(hwin, &r);
		StretchDIBits(dc,r.left,r.top,r.right,r.bottom,
					  0,0,SCRX,SCRY,addbuf,&bi,0,SRCCOPY);
		ReleaseDC(hwin,dc);

		memset(addbuf,0,SCRX*SCRY*4);
	}

	memset(buf,0x0,SCRX*SCRY*4);
}

//
// window proc
//

int do_ss=0; // do screenshot
int goshift=0;
int multiply=0;

long CALLBACK WindowProc( HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam )
{
	int nVirtKey;
	int lKeyData;

	switch (message)
	{
		case WM_CREATE:
		{
			break;
		}
		case WM_DESTROY: 
		{
			PostQuitMessage(0);
			break;
		}
		case WM_PAINT:
		{
			break;
		}
		case WM_MOVE:
		{
			if (g_videomode==1)
			{
				GetClientRect(hWnd, &g_rcScreen);
				ClientToScreen(hWnd, (POINT*)&g_rcScreen.left);
				ClientToScreen(hWnd, (POINT*)&g_rcScreen.right);
			}
			break;
		}
		case WM_LBUTTONDOWN:
		{
			// inform the gui of this
			gui->lButtonDown();
			break;
		}
		case WM_LBUTTONUP:
		{
			// inform the gui of this
			gui->lButtonUp();
			break;
		}
		case WM_RBUTTONDOWN:
		{
			// inform the gui of this
			gui->rButtonDown();
			break;
		}
		case WM_RBUTTONUP:
		{
			// inform the gui of this
			gui->rButtonUp();
			break;
		}
		case WM_MOUSEMOVE:
		{
			int xpos,ypos;

			xpos=LOWORD(lParam);
			ypos=HIWORD(lParam);
			gui->mouseMove(xpos,ypos);

			break;
		}
		case WM_KEYDOWN:
		{
			// for some keys we should control
			// if the key is being just repeated

			nVirtKey=(int)wParam;
			lKeyData = lParam;   

			if (nVirtKey=='1')
			{
				gui->ginput.layeron[0]=1;	
			}
			if (nVirtKey=='2')
			{
				gui->ginput.layeron[1]=1;	
			}
			if (nVirtKey=='3')
			{
				gui->ginput.layeron[2]=1;	
			}
			if (nVirtKey==VK_F8)
			{
				do_ss=1;
			}
			if (nVirtKey==VK_F4)
			{
				goshift-=0x10;
			}
			if (nVirtKey==VK_F5)
			{
				goshift+=0x10;
			}
			if (nVirtKey==VK_F6)
			{
				multiply-=0x10;
			}
			if (nVirtKey==VK_F7)
			{
				multiply+=0x10;
			}

			if (nVirtKey==VK_F3)	//coin p1
			{
				gui->ginput.coin1=1;
			}
			if (nVirtKey==VK_F1)	//start p1
			{
				gui->ginput.start1=1;
			}

			if (nVirtKey==VK_LEFT)
			{
				gui->ginput.p1left=1;
			}
			if (nVirtKey==VK_RIGHT)
			{
				gui->ginput.p1right=1;
			}
			if (nVirtKey==VK_UP)
			{
				gui->ginput.p1up=1;
			}
			if (nVirtKey==VK_DOWN)
			{
				gui->ginput.p1down=1;
			}
			// buttons
			if (nVirtKey=='A')
			{
				gui->ginput.p1button[0]=1;
			}
			if (nVirtKey=='S')
			{
				gui->ginput.p1button[1]=1;
			}
			if (nVirtKey=='D')
			{
				gui->ginput.p1button[2]=1;
			}
			if (nVirtKey=='Z')
			{
				gui->ginput.p1button[3]=1;
			}
			if (nVirtKey=='X')
			{
				gui->ginput.p1button[4]=1;
			}
			if (nVirtKey=='C')
			{
				gui->ginput.p1button[5]=1;
			}


			break;
		}
		case WM_KEYUP:
		{
			nVirtKey=(int)wParam;
			lKeyData = lParam;   

			if (nVirtKey==VK_F8)
			{
				do_ss=0;
			}
			if (nVirtKey==VK_F3)	//coin p1
			{
				gui->ginput.coin1=0;
			}
			if (nVirtKey==VK_F1)	//start p1
			{
				gui->ginput.start1=0;
			}

			if (nVirtKey==VK_LEFT)
			{
				gui->ginput.p1left=0;
			}
			if (nVirtKey==VK_RIGHT)
			{
				gui->ginput.p1right=0;
			}
			if (nVirtKey==VK_DOWN)
			{
				gui->ginput.p1down=0;
			}
			if (nVirtKey==VK_UP)
			{
				gui->ginput.p1up=0;
			}
			// buttons
			if (nVirtKey=='A')
			{
				gui->ginput.p1button[0]=0;
			}
			if (nVirtKey=='S')
			{
				gui->ginput.p1button[1]=0;
			}
			if (nVirtKey=='D')
			{
				gui->ginput.p1button[2]=0;
			}
			if (nVirtKey=='Z')
			{
				gui->ginput.p1button[3]=0;
			}
			if (nVirtKey=='X')
			{
				gui->ginput.p1button[4]=0;
			}
			if (nVirtKey=='C')
			{
				gui->ginput.p1button[5]=0;
			}

			break;
		}
	}

	return DefWindowProc(hWnd, message, wParam, lParam);
}

//
// initDisplay
// videomode: 0 dx fullscreen, 1 dx window, 2 gdi window
//

int initDisplay(int nCmdShow,HWND* rhwnd,unsigned long int** video,int videomode,int ds)
{
	WNDCLASS wc;
	HRESULT hres;
    DDSURFACEDESC2 ddsd;
    DDSCAPS2            ddscaps;
    LPDIRECTDRAWCLIPPER pClipper;

	// store videomode

	g_videomode=videomode;

	// video buffer allocation

	*video=new unsigned long int [SCRX*SCRY];
	memset(*video,0,SCRX*SCRY*4);

	// init window

	if (videomode==0) // dx fullscreen
	{
		//display window

		wc.style=CS_HREDRAW|CS_VREDRAW;
		wc.lpfnWndProc=WindowProc;
		wc.cbClsExtra=0;
		wc.cbWndExtra=0;
		wc.hInstance=hinst;
		wc.hIcon=LoadIcon(hinst,NULL);
		wc.hCursor=LoadCursor(0, IDC_ARROW);
		wc.hbrBackground=GetStockObject(BLACK_BRUSH);
		wc.lpszMenuName=0;
		wc.lpszClassName="emu";
		RegisterClass(&wc);

		RECT r;
		r.top=0;
		r.left=0;
		r.bottom=FSWINY;
		r.right=FSWINX;

		unsigned long style=WS_POPUP;
		unsigned long styleex=0;//WS_EX_TOPMOST;

		hwin = CreateWindowEx(styleex,"emu",appName,style,
							r.left,r.top,r.right,r.bottom,NULL,NULL,hinst,NULL);

		if (!hwin) 
		{
			MessageBox(0,"error","unable to create window",MB_OK);
			return 1;
		}

		ShowWindow(hwin, nCmdShow);
		UpdateWindow(hwin);
		SetFocus(hwin);
	}
	if (videomode==1) //dx window
	{
		//display window

		wc.style=CS_HREDRAW|CS_VREDRAW;
		wc.lpfnWndProc=WindowProc;
		wc.cbClsExtra=0;
		wc.cbWndExtra=0;
		wc.hInstance=hinst;
		wc.hIcon=LoadIcon(hinst,NULL);
		wc.hCursor=LoadCursor(0, IDC_ARROW);
		wc.hbrBackground=GetStockObject(BLACK_BRUSH);
		wc.lpszMenuName=0;
		wc.lpszClassName="emu";
		RegisterClass(&wc);

		RECT r;
		r.top=0;
		r.left=0;

		if (ds)
		{
			r.bottom=WIN_Y*2;
			r.right=WIN_X*2;
		}
		else
		{
			r.bottom=WIN_Y;
			r.right=WIN_X;
		}
//		r.bottom=WIN_Y;
//		r.right=WIN_X;

		unsigned long style=WS_OVERLAPPED;//WS_POPUP;
		unsigned long styleex=0;//WS_EX_TOPMOST;

		hwin = CreateWindowEx(styleex,"emu",appName,style,
							r.left,r.top,r.right,r.bottom,NULL,NULL,hinst,NULL);

		if (!hwin) 
		{
			MessageBox(0,"error","unable to create window",MB_OK);
			return 1;
		}

		ShowWindow(hwin, nCmdShow);
		UpdateWindow(hwin);
		SetFocus(hwin);
	}
	if (videomode==2)
	{
		WNDCLASS wc;

		wc.style=CS_HREDRAW|CS_VREDRAW;
		wc.lpfnWndProc=WindowProc;
		wc.cbClsExtra=0;
		wc.cbWndExtra=0;
		wc.hInstance=hinst;
		wc.hIcon=LoadIcon(hinst,NULL);
		wc.hCursor=LoadCursor(0, IDC_ARROW);
		wc.hbrBackground=GetStockObject(BLACK_BRUSH);
		wc.lpszMenuName=0;
		wc.lpszClassName="emu";
		RegisterClass(&wc);

		RECT r;
		r.top=0;
		r.left=0;

		if (ds)
		{
			r.bottom=r.top+SCRY*2;
			r.right=r.left+SCRX*2;
		}
		else
		{
			r.bottom=r.top+SCRY;
			r.right=r.left+SCRX;
		}

		unsigned long style=WS_OVERLAPPEDWINDOW;
		unsigned long styleex=0;

		hwin = CreateWindowEx(styleex,"emu",appName,style,
							r.left,r.top,r.right,r.bottom,NULL,NULL,hinst,NULL);

		if (!hwin) 
		{
		  return 1;
		}

		addbuf=new unsigned long int [SCRX*SCRY];
		memset(addbuf,0,SCRX*SCRY*4);

		ShowWindow(hwin, nCmdShow);
		UpdateWindow(hwin);
	}

	// init direct draw

	if (videomode==0)
	{
		HANDLE hdll;
		typedef HRESULT (WINAPI *ddc_proc)(GUID FAR *lpGUID,
										   LPVOID *lplpDD,
										   REFIID iid,
										   IUnknown FAR *pUnkOuter);
		ddc_proc ddcreateFunc;

	    hdll=LoadLibrary("ddraw.dll");
		if (hdll==NULL)
		{
			MessageBox(NULL,"error","unable to load ddraw.dll",MB_OK);
			return 1;
		}

		ddcreateFunc=(ddc_proc)GetProcAddress(hdll, "DirectDrawCreateEx");
		if (ddcreateFunc == NULL)
		{
			MessageBox(NULL,"error","unable to attach DirectDrawCreateEx",MB_OK);
			return 1;
		}

		// create dx7 interface object
/*		hres=DirectDrawCreateEx(NULL,(VOID**)&g_pDD,IID_IDirectDraw7,NULL);*/
		hres=ddcreateFunc(NULL,(VOID**)&g_pDD,IID_IDirectDraw7,NULL);
		if (hres!=DD_OK)
		{
			MessageBox(NULL,"error","unable to create DX7 interface",MB_OK);
			return 1;
		}

		// free dll
        FreeLibrary(hdll);

		// try to guess which videomode and bpp are we using

		DDSURFACEDESC2 sd2;
		memset(&sd2,0,sizeof(sd2));
		sd2.dwSize=sizeof(sd2);
		hres=g_pDD->GetDisplayMode(&sd2);
		if (hres!=DD_OK)
		{
			MessageBox(NULL,"error","unable to get video mode",MB_OK);
			return 1;
		}

		oldresX=sd2.dwWidth;
		oldresY=sd2.dwHeight,
		oldresBPP=sd2.ddpfPixelFormat.dwRGBBitCount;

		// set cooperative level
		hres=g_pDD->SetCooperativeLevel(hwin,DDSCL_EXCLUSIVE|DDSCL_FULLSCREEN);
		if (hres!=DD_OK)
		{
			MessageBox(NULL,"error","unable to set cooperative level",MB_OK);
			return 1;
		}

		// set videomode
//		hres=g_pDD->SetDisplayMode(WIN_X,WIN_Y,32,0,0);
		hres=g_pDD->SetDisplayMode(FSWINX,FSWINY,32,0,0);
		if (hres!=DD_OK)
		{
			MessageBox(NULL,"error","unable to set fullscreen display mode",MB_OK);
			return 1;
		}

		// Create the primary surface with 1 back buffer
		ZeroMemory(&ddsd, sizeof(ddsd));
		ddsd.dwSize = sizeof(ddsd);
		ddsd.dwHeight=FSWINY;
		ddsd.dwWidth=FSWINX;
		ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
		ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE |
                			  DDSCAPS_FLIP |
                			  DDSCAPS_COMPLEX;
        ddsd.dwBackBufferCount = 1;

		hres=g_pDD->CreateSurface(&ddsd,&g_pDDSPrimary,NULL);
		if (hres != DD_OK)
		{
			char err[256];
			sprintf(err,"primary surface err %d",hres);
			g_pDD->SetDisplayMode(oldresX,oldresY,oldresBPP, 0, 0);
			MessageBox(NULL,"error",err,MB_OK);
			return 1;
		}

		// create backbuffer
		ZeroMemory(&ddscaps, sizeof(ddscaps));
        ddscaps.dwCaps = DDSCAPS_BACKBUFFER;
        hres = g_pDDSPrimary->GetAttachedSurface(&ddscaps, &g_pDDSBack);
		if (hres != DD_OK)
		{
			g_pDD->SetDisplayMode(oldresX,oldresY,oldresBPP, 0, 0);
			MessageBox(NULL,"error","secondary surface",MB_OK);
			return 1;
		}

		// blank the displays
		unsigned long int* pBits;

		ddsd.dwSize = sizeof(ddsd);
		hres=g_pDDSBack->Lock(NULL,&ddsd,DDLOCK_WAIT,NULL);
		if (hres==DD_OK)
		{
			pBits=(unsigned long int*)ddsd.lpSurface;
			memset(pBits,0x0,FSWINX*FSWINY*4);	
			g_pDDSBack->Unlock(NULL);
		}
	}
	if (videomode==1)
	{
		HANDLE hdll;
		typedef HRESULT (WINAPI *ddc_proc)(GUID FAR *lpGUID,
										   LPVOID *lplpDD,
										   REFIID iid,
										   IUnknown FAR *pUnkOuter);
		ddc_proc ddcreateFunc;

	    hdll=LoadLibrary("ddraw.dll");
		if (hdll==NULL)
		{
			MessageBox(NULL,"error","unable to load ddraw.dll",MB_OK);
			return 1;
		}

		// attach function
		ddcreateFunc=(ddc_proc)GetProcAddress(hdll, "DirectDrawCreateEx");
		if (ddcreateFunc == NULL)
		{
			MessageBox(NULL,"error","unable to attach DirectDrawCreateEx",MB_OK);
			return 1;
		}

		// create dx7 interface object
/*		hres=DirectDrawCreateEx(NULL,(VOID**)&g_pDD,IID_IDirectDraw7,NULL);*/
		hres=ddcreateFunc(NULL,(VOID**)&g_pDD,IID_IDirectDraw7,NULL);
		if (hres!=DD_OK)
		{
			MessageBox(NULL,"error","unable to create DX7 interface",MB_OK);
			return 1;
		}

		// free dll
        FreeLibrary(hdll);

		// try to guess which videomode and bpp are we using

		DDSURFACEDESC2 sd2;
		memset(&sd2,0,sizeof(sd2));
		sd2.dwSize=sizeof(sd2);
		hres=g_pDD->GetDisplayMode(&sd2);
		if (hres!=DD_OK)
		{
			MessageBox(NULL,"error","unable to get video mode",MB_OK);
			return 1;
		}

		oldresX=sd2.dwWidth;
		oldresY=sd2.dwHeight,
		oldresBPP=sd2.ddpfPixelFormat.dwRGBBitCount;

		if ( (oldresBPP!=16) && (oldresBPP!=32) )
		{
			MessageBox(NULL,"Sorry, only 16 and 32 bpp videomodes are supported in this release.","error",MB_OK);
			return 1;
		}

		// set cooperative level
		hres=g_pDD->SetCooperativeLevel(hwin,DDSCL_NORMAL);
		if (hres!=DD_OK)
		{
			MessageBox(NULL,"error","unable to set cooperative level",MB_OK);
			return 1;
		}

		// Create the primary surface with 1 back buffer
		ZeroMemory(&ddsd, sizeof(ddsd));
		ddsd.dwSize = sizeof(ddsd);
		ddsd.dwHeight=WIN_Y*2;
		ddsd.dwWidth=WIN_X*2;
		ddsd.dwFlags = DDSD_CAPS ;
		ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;

		hres=g_pDD->CreateSurface(&ddsd,&g_pDDSPrimary,NULL);
		if (hres != DD_OK)
		{
			char err[256];
			sprintf(err,"primary surface err %d",hres);
			MessageBox(NULL,"error",err,MB_OK);
			return 1;
		}

		// create clipper
		hres = g_pDD->CreateClipper(0,&pClipper,NULL);
		if (hres != DD_OK)
		{
			MessageBox(NULL,"error","cannot create clipper",MB_OK);
			return 1;
		}
 
		// Associate the clipper with the window
		pClipper->SetHWnd(0,hwin);
		g_pDDSPrimary->SetClipper(pClipper);
		pClipper->Release();
		pClipper = NULL;

		// create backbuffer
		ZeroMemory(&ddsd, sizeof(ddsd));
		ddsd.dwSize = sizeof(ddsd);
		ddsd.dwFlags        = DDSD_WIDTH | DDSD_HEIGHT | DDSD_CAPS;
		ddsd.dwWidth        = WIN_X;
		ddsd.dwHeight       = WIN_Y;
		ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
		hres = g_pDD->CreateSurface(&ddsd, &g_pDDSBack, NULL);
		if (hres != DD_OK)
		{
			MessageBox(NULL,"error","secondary surface",MB_OK);
			return 1;
		}
/*
		// blank the displays
		unsigned long int* pBits;

		ddsd.dwSize = sizeof(ddsd);
		hres=g_pDDSBack->Lock(NULL,&ddsd,DDLOCK_WAIT,NULL);
		if (hres==DD_OK)
		{
			pBits=(unsigned long int*)ddsd.lpSurface;
			memset(pBits,0x0,WIN_X*WIN_Y*4);	
			g_pDDSBack->Unlock(NULL);
		}
*/
	}

	*rhwnd=hwin;
	return 0;
}

/*

  freeDisplay

*/

void freeDisplay(unsigned long int** video)
{
	unsigned long int* vb=*video;

	delete [] vb;

	if ((g_videomode==0) || (g_videomode==1))
	{
		if (g_videomode==0)
		{
			g_pDD->SetDisplayMode(oldresX,oldresY,oldresBPP, 0, 0);
		}

		if (g_pDD != NULL)
		{
			if (g_pDDSPrimary != NULL)
			{
				g_pDDSPrimary->Release();
				g_pDDSPrimary = NULL;
			}
			g_pDD->Release();
			g_pDD = NULL;
		}
	}
	if (g_videomode==2)
	{
		delete [] addbuf;
	}
}
