
/*

	where the fun starts (again)
	7/7/2001

*/


#include "cpucore.h"

//#define PROFILE

// control variables

extern tGui* gui;
extern int do_ss; // do the screenshot
extern int goshift;
extern int multiply;

static unsigned short int in0=0xffff;
static unsigned short int in1=0xffff;
static unsigned short int in2=0xfffe;

static int objram_bank=0;
static int charbase,tile16base,tile32base;
static int charpad,tile16pad,tile32pad;
static int layercontrol;

static signed short int* soundbuf[2];
static const int sndbufsize=44100/60;
static signed short int* realsoundbuf;
static int qsound_data=0;

// memory regions

static unsigned char* rom68k;
static unsigned char* crypt;
static unsigned char* cps1ram;
static unsigned char* qsoundram1;
static unsigned char* output_port;
static unsigned char* gfxram;
static unsigned char* extraram;
static unsigned char* objram1;
static unsigned char* objram2;
static unsigned char objectoutput[0x0c];

static unsigned char* z80prg;
static signed char* qsoundrom;

int r8ff=0;
int r16ff=0;
int r32ff=0;
int r8gfx=0;
int r16gfx=0;
int r32gfx=0;


//

// z80

static CONTEXTMZ80 ctx1;

// z80 read/write functions

static unsigned char z80rd(unsigned long int A,struct MemoryReadByte* mwb)
{
	if ( (A>=0xc000) && (A<=0xcfff) )
	{
		return qsoundram1[A-0xc000];
	}

	return 0x80;
}


static void z80wr(unsigned long int A,unsigned char V,struct MemoryWriteByte* mwb)
{
	if ( (A>=0xc000) && (A<0xd000) )
	{
		qsoundram1[A-0xc000]=V;
	}
	else if (A==0xd000)
	{
		// qsound_data_h_w
		qsound_data=(qsound_data&0xff)|(V<<8);
	}
	else if (A==0xd001)
	{
		// qsound_data_l_w
		qsound_data=(qsound_data&0xff00)|V;
	}
	else if (A==0xd002)
	{
		// qsound_cmd_w
		qsound_set_command(V, qsound_data);
	}
	else if (A==0xd003)
	{
		// qsound_banksw_w
		int bankaddress=0x10000+((V&0x0f)*0x4000);
		if (bankaddress >= 0x40000)
		{
			bankaddress=0x10000;
		}

		memcpy(z80prg+0x8000,z80prg+bankaddress,0x4000);
	}
}

// z80 handlers

static struct z80PortRead z80ReadPorts[] =
{
	{-1,-1,0}
};

static struct z80PortWrite z80WritePorts[] =
{
	{-1,-1,0}
};

static struct MemoryReadByte z80readmem[] =
{
	{0xc000,0xefff,z80rd,0},
	{-1,-1,0,0}
};
  
static struct MemoryWriteByte z80writemem[] =
{
	{0x0000,0xefff,z80wr,0},
	{-1,-1,0,0}
};


// eepro-t
// thanks again to mame team

static struct EEPROM_interface cps2_eeprom_interface =
{
	6,		/* address bits */
	16,		/* data bits */
	"0110",	/*  read command */
	"0101",	/* write command */
	"0111"	/* erase command */
};

// cps2 decrypt

void cps2decrypt()
{
	int i;

	unsigned char* temp=new unsigned char [0x400000];

	memcpy(temp,rom68k,0x400000);

	for (i=0;i<0x400000;i++)
	{
		rom68k[i]=rom68k[i]^crypt[i];
	}

	memcpy(crypt,temp,0x400000);

	delete [] temp;
}

// cpu init

void cps2cpucore_init(tRomSpace* roms)
{
	rom68k=new unsigned char [0x400000];
	crypt=new unsigned char [0x400000];

	cps1ram=new unsigned char [0x10000];
	qsoundram1=new unsigned char [0x1000];
	extraram=new unsigned char [0x4002];
	objram1=new unsigned char [0x2000];
	objram2=new unsigned char [0x2000];
	output_port=new unsigned char [0x100];
	gfxram=new unsigned char [0x30000];

	z80prg=new unsigned char [0x80000];
	qsoundrom=new signed char [0x400000];

	// init memory areas

	memset(rom68k,0,0x400000);
	memset(crypt,0,0x400000);

	memcpy(rom68k,roms->romCPU1,0x400000);
	memcpy(crypt,roms->romCPU1+0x400000,0x400000);

	// decrypt cps2 code
	cps2decrypt();

	// patch out eeprom test of capcom sports club japan
	rom68k[0x27cb8]=0x75;
	rom68k[0x27cb9]=0x4e;

	memset(extraram,0,0x4002);
	memset(objram1,0,0x2000);
	memset(objram2,0,0x2000);
	memset(objectoutput,0,0x0c);
	memset(gfxram,0,0x30000);
	memset(output_port,0,0x100);
	memset(cps1ram,0,0x10000);
	memset(qsoundram1,0,0x1000);

	memset(z80prg,0,0x80000);
	memcpy(z80prg,roms->romCPU2,0x8000);
	memcpy(z80prg+0x10000,roms->romCPU2+0x8000,0x18000);
	memcpy(z80prg+0x28000,roms->romCPU2+0x20000,0x20000);
	memcpy(z80prg+0x8000,roms->romCPU2+0x10000,0x4000);

	memcpy(qsoundrom,roms->romqsound,0x400000);

	// byteswap
	for (int i=0;i<0x400000;i+=2)
	{
		signed char temp=qsoundrom[i];
		qsoundrom[i]=qsoundrom[i+1];
		qsoundrom[i+1]=temp;
	}

	// set up values

	layercontrol=roms->layercontrol;
	charbase=roms->charbase;
	tile16base=roms->tile16base;
	tile32base=roms->tile32base;
	charpad=roms->charpad;
	tile16pad=roms->tile16pad;
	tile32pad=roms->tile32pad;

	// init 68000 cpu

	m68k_set_cpu_type(M68K_CPU_TYPE_68000);
	m68k_pulse_reset();

	// init eeprom

	EEPROM_init(&cps2_eeprom_interface);

	// init z80

	mz80init();

	ctx1.z80Base=z80prg;
	ctx1.z80IoRead=z80ReadPorts;
	ctx1.z80IoWrite=z80WritePorts;
	ctx1.z80MemRead=z80readmem;
	ctx1.z80MemWrite=z80writemem;

	mz80SetContext(&ctx1);
	mz80reset();
	mz80GetContext(&ctx1);

	// init qsound

	soundbuf[0]=new signed short int [sndbufsize];
	soundbuf[1]=new signed short int [sndbufsize];
	realsoundbuf=new signed short int [sndbufsize*2];

	QSound_interface qs;
	qs.clock=3800000;
	qs.samplerom=qsoundrom;
	qs.mixing_level[0]=100;
	qs.mixing_level[1]=100;

	qsound_sh_start(&qs);
}

void cps2cpucore_free()
{
	delete [] rom68k;
	delete [] crypt;
	delete [] cps1ram;
	delete [] qsoundram1;
	delete [] output_port;
	delete [] gfxram;
	delete [] extraram;
	delete [] objram1;
	delete [] objram2;
	delete [] z80prg;
	delete [] qsoundrom;

	delete [] soundbuf[0];
	delete [] soundbuf[1];
	delete [] realsoundbuf;

#ifdef PROFILE
	char profile[1024];
	sprintf(profile,"gfx [%d] [%d] [%d] ff [%d] [%d] [%d]",
		    r8gfx,r16gfx,r32gfx,r8ff,r16ff,r32ff);
	MessageBox(0,profile,"info",MB_OK);
#endif
}


void cps2cpucore_emulate(int snd)
{
	int res;
	unsigned long int dwres;

	if (!snd)
	{
		qsoundram1[0xfff] = 0x77;		
		res=m68k_execute(196666);
		qsoundram1[0xfff] = 0x77;		
		m68k_set_irq(2);
	}
	else
	{
		for (int i=0;i<4;i++)
		{
			res=m68k_execute(196666/4);

			// z80 - 8 mhz

			mz80SetContext(&ctx1);
			dwres=mz80exec(133333/4);	
			if (dwres!=0x80000000)
			{
				MessageBox(NULL,"z80 bad op","error",MB_OK);
			}
			mz80int(0x38);
			mz80GetContext(&ctx1);
		}

		m68k_set_irq(2);

		qsound_update(0,soundbuf,sndbufsize);

		for (i=0;i<sndbufsize;i++)
		{
			realsoundbuf[i*2]=soundbuf[0][i];
			realsoundbuf[i*2+1]=soundbuf[1][i];
		}

		feedBuffer(realsoundbuf,sndbufsize*2);
	}
}

static void update_controls()
{
	if (gui->ginput.coin1==1)
	{
		in2=in2&(~0x1000);
	}
	else
	{
		in2=in2|0x1000;
	}

	if (gui->ginput.start1==1)
	{
		in2=in2&(~0x100);
	}
	else
	{
		in2=in2|0x100;
	}

	if (gui->ginput.p1left==1)
	{
		in0=in0&(~0x02);
	}
	else
	{
		in0=in0|0x02;
	}

	if (gui->ginput.p1right==1)
	{
		in0=in0&(~0x01);
	}
	else
	{
		in0=in0|0x01;
	}

	if (gui->ginput.p1up==1)
	{
		in0=in0&(~0x08);
	}
	else
	{
		in0=in0|0x08;
	}

	if (gui->ginput.p1down==1)
	{
		in0=in0&(~0x04);
	}
	else
	{
		in0=in0|0x04;
	}

	if (gui->ginput.p1button[0]==1)
	{
		in0=in0&(~0x10);
	}
	else
	{
		in0=in0|0x10;
	}

	if (gui->ginput.p1button[1]==1)
	{
		in0=in0&(~0x20);
	}
	else
	{
		in0=in0|0x20;
	}

	if (gui->ginput.p1button[2]==1)
	{
		in0=in0&(~0x40);
	}
	else
	{
		in0=in0|0x40;
	}

	if (gui->ginput.p1button[3]==1)
	{
		in1=in1&(~0x1);
	}
	else
	{
		in1=in1|0x1;
	}

	if (gui->ginput.p1button[4]==1)
	{
		in1=in1&(~0x2);
	}
	else
	{
		in1=in1|0x2;
	}

	if (gui->ginput.p1button[5]==1)
	{
		in1=in1&(~0x4);
	}
	else
	{
		in1=in1|0x4;
	}
}

//
// general helper functions
//

__inline static int READ_WORD(void *dst)
{
	unsigned char *p = (unsigned char *) dst;
	return (p[0] << 8) | p[1];
}

// special reads/writes

__inline static int specREAD8(unsigned char* mem,unsigned long int addr,unsigned long int base)
{
	return mem[(addr^1)-base];
}

__inline static int specREAD16(unsigned char* mem,unsigned long int addr,unsigned long int base)
{
	unsigned short int* pmem=(unsigned short int*)&mem[addr-base];
	return (*pmem);
}

__inline static unsigned long int specREAD32(unsigned char* mem,unsigned long int addr,unsigned long int base)
{
	return (
			(mem[addr-base+1]<<24)|
			(mem[addr-base+0]<<16)|
			(mem[addr-base+3]<<8)|
			(mem[addr-base+2])
			);
}

__inline static void specWRITE8(unsigned char* mem,unsigned long int addr,unsigned long int base,unsigned long int val)
{
	mem[(addr^1)-base]=(unsigned char)(val>>8);
}

__inline static void specWRITE16(unsigned char* mem,unsigned long int addr,unsigned long int base,unsigned long int val)
{
	mem[addr-base+1]=(unsigned char)(val>>8);
	mem[addr-base  ]=(unsigned char)val;
}

__inline static void specWRITE32(unsigned char* mem,unsigned long int addr,unsigned long int base,unsigned long int val)
{
	mem[addr-base+1]=(unsigned char)(val>>24);
	mem[addr-base+0]=(unsigned char)(val>>16);
	mem[addr-base+3]=(unsigned char)(val>>8);
	mem[addr-base+2]=(unsigned char)val;
}

//

__inline static int READ8(unsigned char* mem,unsigned long int addr,unsigned long int base)
{
	return mem[addr-base];
}

__inline static int READ16(unsigned char* mem,unsigned long int addr,unsigned long int base)
{
		return ((mem[addr-base]<<8)|mem[addr-base+1]);
}

__inline static unsigned long int READ32(unsigned char* mem,unsigned long int addr,unsigned long int base)
{
		return (
				(mem[addr-base]<<24)|
				(mem[addr-base+1]<<16)|
				(mem[addr-base+2]<<8)|
				(mem[addr-base+3])
				);
}

__inline static void WRITE8(unsigned char* mem,unsigned long int addr,unsigned long int base,unsigned long int val)
{
	mem[addr-base]=(unsigned char)val;
}

__inline static void WRITE16(unsigned char* mem,unsigned long int addr,unsigned long int base,unsigned long int val)
{
	mem[addr-base]=(unsigned char)(val>>8);
	mem[addr-base+1]=(unsigned char)val;
}

__inline static void WRITE32(unsigned char* mem,unsigned long int addr,unsigned long int base,unsigned long int val)
{
	mem[addr-base  ]=(unsigned char)(val>>24);
	mem[addr-base+1]=(unsigned char)(val>>16);
	mem[addr-base+2]=(unsigned char)(val>>8);
	mem[addr-base+3]=(unsigned char)(val);
}

// special writes on output ports

__inline static void WRITE8opp(unsigned char* mem,unsigned long int addr,unsigned long int base,unsigned long int val)
{
	mem[(addr^1)-base]=(unsigned char)(val>>8);
}

__inline static void WRITE16opp(unsigned char* mem,unsigned long int addr,unsigned long int base,unsigned long int val)
{
	mem[addr-base+1]=(unsigned char)(val>>8);
	mem[addr-base  ]=(unsigned char)val;
}

__inline static void WRITE32opp(unsigned char* mem,unsigned long int addr,unsigned long int base,unsigned long int val)
{
	mem[addr-base+1]=(unsigned char)(val>>24);
	mem[addr-base+0]=(unsigned char)(val>>16);
	mem[addr-base+3]=(unsigned char)(val>>8);
	mem[addr-base+2]=(unsigned char)(val);
}

//
// read-write handlers
//

// pc-relativee stuph

unsigned int cps2_m68kri16(unsigned int address)
{
	if (address<=0x3fffff) 
	{
		return specREAD16(rom68k,address,0);
	}

	return 0;
}

unsigned int cps2_m68kri32(unsigned int address)
{
	return ((m68k_read_immediate_16(address) << 16) | m68k_read_immediate_16((address)+2));
}

unsigned int cps2_m68krpc8(unsigned int address)
{
	if (address<=0x3fffff)
	{
		return ((m68k_read_immediate_16(address&~1)>>(8*(1-(address & 1))))&0xff);
	}

	return m68k_read_memory_8(address);
}

unsigned int cps2_m68krpc16(unsigned int address)
{
	if (address<=0x3fffff)
	{
		return m68k_read_immediate_16(address);
	}

	return m68k_read_memory_16(address);
}

unsigned int cps2_m68krpc32(unsigned int address)
{
	if (address<=0x3fffff)
	{
		return m68k_read_immediate_32(address);
	}

	return m68k_read_memory_32(address);
}

unsigned int cps2_m68kr8(unsigned int address)
{
	if (address<=0x3fffff)
	{
		return specREAD8(crypt,address,0);
	}
	else if ( (address>=0xff0000) && (address<=0xffffff) )
	{
		return READ8(cps1ram,address,0xff0000);
	}
/*	else if ( (address>=0x400000) && (address<=0x40000b) )
	{
		return READ8(objectoutput,address,0x400000);
	}
*/
/*	else if ( (address>=0x660000) && (address<=0x664001) )
	{
		return READ8(extraram,address,0x660000);
	}
	else if ( (address>=0x900000) && (address<=0x92ffff) )
	{
		return READ8(gfxram,address,0x900000);
	}
*/	else if ( (address>=0x618000) && (address<=0x619fff) )
	{
//		r8ff++;
		return (qsoundram1[(address-0x618000)/2]);
	}
	else if (address==0x804001)
	{
		return in0&0xff;
	}
	else if (address==0x804000)
	{
		return ((in0>>8)&0xff);
	}
	else if ( (address>=0x804010) && (address<=0x804011) )
	{
		return in1;
	}
	else if (address==0x804021) 
	{
//		return in2&0xff;
		return 0xff;
	}
	else if (address==0x804020) 
	{
//		unsigned short int res=((in2&0xfffe)| EEPROM_read_bit());
		unsigned short int res=in2&0xfffe;
		return ((res>>8)&0xff);
	}
	else if (address==0x804030)
	{
	    return 0xe0;
	}
	else if (address==0x804031)
	{
	    return 0x21;
	}
	else if ( (address>=0x8040b0) && (address<=0x8040b3) )
	{
		return 0xff;
	}
	else if ( (address>=0x708000) && (address<=0x70ffff) )
	{
		if (objram_bank&1)
		{
			return objram1[address&0x1fff];
		}
		else
		{
			return objram2[address&0x1fff];
		}
	}

	return 0;
}

unsigned int cps2_m68kr16(unsigned int address)
{
	if (address<=0x3fffff)
	{
		return specREAD16(crypt,address,0);
	}
/*	else if ( (address>=0x400000) && (address<=0x40000b) )
	{
		return READ16(objectoutput,address,0x400000);
	}*/
	else if ( (address>=0xff0000) && (address<=0xffffff) )
	{
		return READ16(cps1ram,address,0xff0000);
	}
	else if ( (address>=0x900000) && (address<=0x92ffff) )
	{
		return READ16(gfxram,address,0x900000);
	}
	else if ( (address>=0x618000) && (address<=0x619fff) )
	{
		return (qsoundram1[(address-0x618000)/2]|0xff00);
	}
/*	else if ( (address>=0x660000) && (address<=0x664001) )
	{
		return READ16(extraram,address,0x660000);
	}
*/
	else if ( (address>=0x804000) && (address<=0x804001) )
	{
		return in0;
	}
	else if ( (address>=0x804010) && (address<=0x804011) )
	{
		return in1;
	}
	else if ( (address>=0x804020) && (address<=0x804021) )
	{
		return ((in2&0xfffe)| EEPROM_read_bit());
	}
	else if ( (address>=0x804030) && (address<=0x804031) )
	{
	    return 0xe021;
	}
	else if ( (address>=0x8040b0) && (address<=0x8040b3) )
	{
		return 0xffff;
	}
	else if ( (address>=0x708000) && (address<=0x70ffff) )
	{
		if (objram_bank&1)
		{
			return ((objram1[address&0x1fff]<<8)|
					(objram1[(address+1)&0x1fff]));
		}
		else
		{
			return ((objram1[address&0x1fff]<<8)|
					(objram1[(address+1)&0x1fff]));
		}
	}

	return 0;
}

unsigned int cps2_m68kr32(unsigned int address)
{
	if (address<=0x3fffff)
	{
		return specREAD32(crypt,address,0);
	}
	else if ( (address>=0xff0000) && (address<=0xffffff) )
	{
		return READ32(cps1ram,address,0xff0000);
	}
/*	else if ( (address>=0x400000) && (address<=0x40000b) )
	{
		return READ32(objectoutput,address,0x400000);
	}
	else if ( (address>=0x660000) && (address<=0x664001) )
	{
		return READ32(extraram,address,0x660000);
	}*/
	else if ( (address>=0x900000) && (address<=0x92ffff) )
	{
		return READ32(gfxram,address,0x900000);
	}
/*	else if ( (address>=0x804000) && (address<=0x804001) )
	{
		return in0;
	}
	else if ( (address>=0x804010) && (address<=0x804011) )
	{
		return in1;
	}
	else if ( (address>=0x804020) && (address<=0x804021) )
	{
		return ((in2&0xfffe)| EEPROM_read_bit());
	}
*/	else if ( (address>=0x804030) && (address<=0x804031) )
	{
	    return 0xe021;
	}
	else if ( (address>=0x8040b0) && (address<=0x8040b3) )
	{
		return 0xffff;
	}
	else if ( (address>=0x708000) && (address<=0x70ffff) )
	{
		if (objram_bank&1)
		{
			return READ32(objram1,address&0x1fff,0);
		}
		else
		{
			return READ32(objram2,address&0x1fff,0);
		}
	}

	return 0;
}

void cps2_m68kw8(unsigned int address, unsigned int value)
{
	if (address>=0xff0000)
	{
		WRITE8(cps1ram,address,0xff0000,value);
		return ;
	}
	else if ( (address>=0x900000) && (address<=0x92ffff) )
	{
		WRITE8(gfxram,address,0x900000,value);
		return ;
	}
/*	else if ( (address>=0x400000) && (address<=0x40000b) )
	{
		WRITE8(objectoutput,address,0x400000,value);
		return ;
	}
*/	else if ( (address>=0x618000) && (address<=0x619fff) )
	{
		qsoundram1[(address-0x618000)/2]=value&0xff;
		return ;
	}
/*	else if ( (address>=0x660000) && (address<=0x664001) )
	{
		WRITE8(extraram,address,0x660000,value);
		return ;
	}
*/	else if ( (address>=0x800100) && (address<=0x8001ff) )
	{
		WRITE8opp(output_port,address,0x800100,value);
		return ;
	}
	else if (address==0x804040)
	{
        EEPROM_write_bit(value & 0x1000);
        EEPROM_set_clock_line((value & 0x2000) ? ASSERT_LINE : CLEAR_LINE);
        EEPROM_set_cs_line((value & 0x4000) ? CLEAR_LINE : ASSERT_LINE);
		return ;
	}
	else if ( (address>=0x804100) && (address<=0x8041ff) )
	{
		WRITE8opp(output_port,address,0x804100,value);
		return ;
	}
	else if ( (address>=0x708000) && (address<=0x70ffff) )
	{
		if (objram_bank&1)
		{
//			WRITE8(objram1,address,0x708000,value);
			objram1[address&0x1fff]=value;
			return ;
		}
		else
		{
//			WRITE8(objram2,address,0x708000,value);
			objram2[address&0x1fff]=value;
			return ;
		}
	}
	else if ( (address>=0x8040e0) && (address<=0x8040e1) )
	{
		objram_bank=value&1;
		return;
	}
}

void cps2_m68kw16(unsigned int address, unsigned int value)
{
	if (address>=0xff0000)
	{
		WRITE16(cps1ram,address,0xff0000,value);
		return ;
	}
	else if ( (address>=0x900000) && (address<=0x92ffff) )
	{
		WRITE16(gfxram,address,0x900000,value);
		return ;
	}
	else if ( (address>=0x618000) && (address<=0x619fff) )
	{
		qsoundram1[(address-0x618000)/2]=value;
		return ;
	}
	else if ( (address>=0x800100) && (address<=0x8001ff) )
	{
		WRITE16opp(output_port,address,0x800100,value);
		return ;
	}
	else if (address==0x804040)
	{
        EEPROM_write_bit(value & 0x1000);
        EEPROM_set_clock_line((value & 0x2000) ? ASSERT_LINE : CLEAR_LINE);
        EEPROM_set_cs_line((value & 0x4000) ? CLEAR_LINE : ASSERT_LINE);
		return ;
	}
	else if ( (address>=0x804100) && (address<=0x8041ff) )
	{
		WRITE16opp(output_port,address,0x804100,value);
		return ;
	}
	else if ( (address>=0x708000) && (address<=0x70ffff) )
	{
		if (objram_bank&1)
		{
			WRITE16(objram1,address&0x1fff,0,value);
			return ;
		}
		else
		{
			WRITE16(objram2,address&0x1fff,0,value);
			return ;
		}
	}
/*	else if ( (address>=0x660000) && (address<=0x664001) )
	{
		WRITE16(extraram,address,0x660000,value);
		return ;
	}
*/	else if ( (address>=0x8040e0) && (address<=0x8040e1) )
	{
		objram_bank=value&1;
		return;
	}
/*	else if ( (address>=0x400000) && (address<=0x40000b) )
	{
		WRITE16(objectoutput,address,0x400000,value);
		return ;
	}
*/
}

void cps2_m68kw32(unsigned int address, unsigned int value)
{
	if (address>=0xff0000)
	{
		WRITE32(cps1ram,address,0xff0000,value);
		return ;
	}
	else if ( (address>=0x900000) && (address<=0x92ffff) )
	{
		WRITE32(gfxram,address,0x900000,value);
		return ;
	}
/*	else if ( (address>=0x400000) && (address<=0x40000b) )
	{
		WRITE32(objectoutput,address,0x400000,value);
		return ;
	}
	else if ( (address>=0x660000) && (address<=0x664001) )
	{
		WRITE32(extraram,address,0x660000,value);
		return ;
	}
*/	else if ( (address>=0x800100) && (address<=0x8001ff) )
	{
		WRITE32opp(output_port,address,0x800100,value);
		return ;
	}
	else if ( (address>=0x804100) && (address<=0x8041ff) )
	{
		WRITE32opp(output_port,address,0x804100,value);
		return ;
	}
	else if ( (address>=0x708000) && (address<=0x70ffff) )
	{
		if (objram_bank&1)
		{
			WRITE32(objram1,address&0x1fff,0,value);
			return ;
		}
		else
		{
			WRITE32(objram2,address&0x1fff,0,value);
			return ;
		}
	}
	else if (address==0x804040)
	{
        EEPROM_write_bit(value & 0x1000);
        EEPROM_set_clock_line((value & 0x2000) ? ASSERT_LINE : CLEAR_LINE);
        EEPROM_set_cs_line((value & 0x4000) ? CLEAR_LINE : ASSERT_LINE);
		return ;
	}
	else if ( (address>=0x8040e0) && (address<=0x8040e1) )
	{
		objram_bank=value&1;
		return;
	}
}

//
//
//

static void cpucore_build_palette(tCharset* chset)
{
	// get palette base

	unsigned short int* popp=(unsigned short int*)&output_port[0x0a];
	int base=*popp*256;
	base&= ~(0x2000-1);
	unsigned char* ppal=(unsigned char*)&gfxram[base&0x3ffff];

	int offset;

	for (offset = 0; offset < 32*32*8; offset++)
	{
		int palette = READ_WORD(&ppal[offset*2]);

		int red, green, blue, bright;

		bright= (palette>>12);
		if (bright) bright += 2;

		red   = ((palette>>8)&0x0f) * bright;
		green = ((palette>>4)&0x0f) * bright;
		blue  = (palette&0x0f) * bright;

		// store

		chset->palette[offset*4+2]=red&0xff;
		chset->palette[offset*4+1]=green&0xff;
		chset->palette[offset*4+0]=blue&0xff;
		chset->palette[offset*4+3]=0;
	}
}


/*

  scroll 1

*/

void cps2cpucore_drawScroll1(unsigned long int* video,tCharset* chset)
{
	// base of scroll1 is at output ports 0x02-0x03
	// scroll1 x and y are at 0x0c and 0x0e

	unsigned short int* popbs1=(unsigned short int*)&output_port[0x02];
	unsigned short int* pops1x=(unsigned short int*)&output_port[0x0c];
	unsigned short int* pops1y=(unsigned short int*)&output_port[0x0e];
//	unsigned short int* popbs1=(unsigned short int*)&cps2output_port[0x02];
//	unsigned short int* pops1x=(unsigned short int*)&cps2output_port[0x0c];
//	unsigned short int* pops1y=(unsigned short int*)&cps2output_port[0x0e];

	int x,y, offs, offsx, offsy, sx, sy, ytop;
	int code,realcode,attr,cl,ch;

	int scroll1x=(*pops1x)&0xffff;
	int scroll1y=(*pops1y)&0xffff;

	int base=*popbs1*256;
	base&= ~(0x4000-1);

	unsigned char* c_pgfx=(unsigned char*)&gfxram[base&0x3ffff];

	FILE* ss;
	if (do_ss==1)
	{
		ss=fopen("ss1.bin","wb");
	}

	// draw layer

	int scrlxrough=(scroll1x>>3)+4;
	int scrlyrough=(scroll1y>>3);

	sx=-(scroll1x&0x07)-32;
	ytop=-(scroll1y&0x07)-16;

	for (x=0; x<0x35; x++)
	{
		 sy=ytop;
		 offsx=(scrlxrough+x)*0x80;
		 offsx&=0x1fff;

		 for (y=0; y<0x20; y++)
		 {
			int n=scrlyrough+y;
			offsy=( (n&0x1f)*4 | ((n&0x20)*0x100)) & 0x3fff;
			offs=offsy+offsx;
			offs &= 0x3fff;

			// read code and attrib
			
			code=READ_WORD(&c_pgfx[offs]);
			attr=READ_WORD(&c_pgfx[offs+2]);

			cl =code &0xff;	ch =(code >>8)&0xff;

//			charbase=goshift;
			realcode=(code-charbase)&charpad;

			if ( (realcode!=0x20) && (code!=0) )
			{
				chset->printChar(video,realcode,attr&0x1f,sx,sy,attr&0x20,attr&0x40);
			}

			if (do_ss==1)
			{
				fwrite(&ch,1,1,ss); fwrite(&cl,1,1,ss);
			}

			sy+=8;
		 }
		 sx+=8;
	}

	if (do_ss==1)
	{
		fclose(ss);
	}
}

/*

  Scroll 2

*/

void cps2cpucore_drawScroll2(unsigned long int* video,tCharset* chset)
{
	// base of scroll2 is at output ports 0x04-0x05
	// scroll2 x and y are at 0x10 and 0x12

	unsigned short int* popbs2=(unsigned short int*)&output_port[0x04];
	unsigned short int* pops2x=(unsigned short int*)&output_port[0x10];
	unsigned short int* pops2y=(unsigned short int*)&output_port[0x12];

	int base=*popbs2*256;
	base&= ~(0x4000-1);

	unsigned char* c_pgfx=(unsigned char*)&gfxram[base&0x3ffff];

	int scroll2x=*pops2x+64;
	int scroll2y=*pops2y+64-16;

	int sx, sy;
	int nxoffset=(scroll2x&0x0f)+32;    
	int nyoffset=(scroll2y&0x0f);    
	int nx=(scroll2x>>4);	  
	int ny=(scroll2y>>4)-4;	

	FILE* ss;
	if (do_ss==1)
	{
		ss=fopen("ss2.bin","wb");
	}

	for (sx=0; sx<0x32/2+4; sx++)
	{
		for (sy=0; sy<0x09*2-2; sy++)
		{
			int offsy, offsx, offs, colour, code,cl,ch,realcode;
			int n;
			n=ny+sy+2;
			offsy  = ((n&0x0f)*4 | ((n&0x30)*0x100))&0x3fff;
			offsx=((nx+sx)*0x040)&0xfff;
			offs=offsy+offsx;
			offs &= 0x3fff;

			code=  READ_WORD(&c_pgfx[offs]);
			colour=READ_WORD(&c_pgfx[offs+2]);
			cl =code &0xff;	ch =(code >>8)&0xff;

			realcode=(code-tile16base)&tile16pad;

			if (do_ss==1)
			{
				fwrite(&ch,1,1,ss); fwrite(&cl,1,1,ss);
			}

			chset->printTile16(video,realcode,colour&0x1f,
							   (16*sx)-(nxoffset-32),
							   (16*sy)-nyoffset,
							   colour&0x20,colour&0x40);
		}
	}

	if (do_ss==1)
	{
		fclose(ss);
	}
}

void cps2cpucore_drawScroll2rs(unsigned long int* video,tCharset* chset)
{
	// base of scroll2 is at output ports 0x04-0x05
	// scroll2 x and y are at 0x10 and 0x12

	unsigned short int* popbs2=(unsigned short int*)&output_port[0x04];
	unsigned short int* pops2x=(unsigned short int*)&output_port[0x10];
	unsigned short int* pops2y=(unsigned short int*)&output_port[0x12];

	int base=*popbs2*256;
	base&= ~(0x4000-1);

	unsigned char* c_pgfx=(unsigned char*)&gfxram[base&0x3ffff];

	// we should get rowscroll ram base (0x08)
	// and rowscroll offset (0x20)

	unsigned short int* prowscroll=(unsigned short int*)&output_port[0x08];
	base=*prowscroll*256;
	base&=~(0x800-1);
	unsigned char* prs=(unsigned char*)&gfxram[base&0x3ffff];

	unsigned short int* prowscrolloffs=(unsigned short int*)&output_port[0x20];
	int rowscrolloffs=*prowscrolloffs;

	//

	int scroll2x=*pops2x+64;
	int scroll2y=*pops2y;

	int sx, sy;
	int nxoffset=(scroll2x&0x0f)+32;    
	int nyoffset=(scroll2y&0x0f);    
	int nx=(scroll2x>>4);	  
	int ny=(scroll2y>>4)-4;	

	//

	int scrollx[1024];
	int scrly=-scroll2y;


	for (int i=0;i<256 ;i++)
	{
		scrollx[(i - scrly) & 0x3ff] = (( -(READ_WORD(&prs[(2*(i + rowscrolloffs)) & 0x7ff]))));
	}

	//

	FILE* ss;
	if (do_ss==1)
	{
		ss=fopen("ss2.bin","wb");
	}

	for (sx=0x2a; sx>=0; sx--)
	{
		int n=ny;
		for (sy=0x12-1; sy>=0; sy--)
		{
			int realcode,ch,cl;
			int offsy, offsx, offs, colour, code;

			n=ny+sy+2;
			offsy  = ((n&0x0f)*4 | ((n&0x30)*0x100))&0x3fff;
			offsx=((nx+sx)*0x040)&0xfff;
			offs=offsy+offsx;
			offs &= 0x3fff;

			code=  READ_WORD(&c_pgfx[offs]);
			colour=READ_WORD(&c_pgfx[offs+2]);

			realcode=(code-tile16base)&tile16pad;

			if (do_ss==1)
			{
				cl =code &0xff;	ch =(code >>8)&0xff;
				fwrite(&ch,1,1,ss); fwrite(&cl,1,1,ss);
			}

			chset->printTile16rs(video,realcode,colour&0x1f,
							   (16*sx)+32-nxoffset,
							   (16*sy)-nyoffset-48,
							   colour&0x20,colour&0x40,scrollx,(-scrly+0x10)&0x3ff);
		}
	}

	if (do_ss==1)
	{
		fclose(ss);
	}
}


/*

  Scroll 3

*/

void cps2cpucore_drawScroll3(unsigned long int* video,tCharset* chset)
{
	// base of scroll3 is at output ports 0x06-0x07
	// scroll3 x and y are at 0x14 and 0x16

	unsigned short int* popbs3=(unsigned short int*)&output_port[0x06];
	unsigned short int* pops3x=(unsigned short int*)&output_port[0x14];
	unsigned short int* pops3y=(unsigned short int*)&output_port[0x16];

	int base=(*popbs3)*256;
	base&= ~(0x4000-1);

	unsigned char* c_pgfx=(unsigned char*)&gfxram[base&0x3ffff];

	int scroll3x=*pops3x+32;
	int scroll3y=*pops3y+32;

	int sx,sy;
	int nxoffset=scroll3x&0x1f;
	int nyoffset=scroll3y&0x1f;
	int nx=(scroll3x>>5)+1;
	int ny=(scroll3y>>5)-1;

	FILE* ss;
	if (do_ss==1)
	{
		ss=fopen("ss3.bin","wb");
	}

	for (sx=0; sx<0x32/4+2; sx++)
	{
		for (sy=0; sy<0x20/4+1; sy++)
		{
			int offsy, offsx, offs, colour, code,realcode;
			int n;
			n=ny+sy;
			offsy  = ((n&0x07)*4 | ((n&0xf8)*0x0100))&0x3fff;
			offsx=((nx+sx)*0x020)&0x7ff;
			offs=offsy+offsx;
			offs &= 0x3fff;

			colour=READ_WORD(&c_pgfx[offs+2]);
			code=READ_WORD(&c_pgfx[offs]);

			if (do_ss==1)
			{
				int cl,ch;

				cl=code&0xff; 
				ch=(code>>8)&0xff;

				fwrite(&ch,1,1,ss); fwrite(&cl,1,1,ss);
			}

			realcode=(code-tile32base-goshift)&tile32pad;
			if (realcode!=0)
			{
				chset->printTile32(video,realcode,colour&0x1f,
								   32*sx-nxoffset,
								   32*sy-16-nyoffset,
								   colour&0x20,colour&0x40);
			}
		}
	}

	if (do_ss==1)
	{
		fclose(ss);
	}
}

/*

  Spiritezz

*/

void cps2cpucore_drawSprites(unsigned long int* video,tCharset* chset)
{
	unsigned char* c_pgfx;
	
	if (objram_bank&1)
	{
		c_pgfx=(unsigned char*)&objram1[0];
	}
	else
	{
		c_pgfx=(unsigned char*)&objram2[0];
	}

	unsigned short int* base=(unsigned short int*)c_pgfx;

	int cps1_lso;
	int i;
	int offset=0;

	int found=0;

	FILE* ss;
	if (do_ss==1)
	{
		ss=fopen("sss.bin","wb");
	}

	while ( (offset < 0x1000) && (!found) )
	{
		int colour1=READ_WORD(&c_pgfx[offset+6]);
		int colour2=READ_WORD(&c_pgfx[offset+2]);

		if ( (colour1>=0xff00) || (colour2>=0x8000) )
		{
			cps1_lso=offset-8;
			found=1;
		}
		else
		{
			offset+=8;
		}
	}
	if (!found) cps1_lso=0x1000-8;

	int code2;
	int cl,ch;

	for (i=0; i<=cps1_lso; i+=8)
	{
		signed short int x=READ_WORD(&c_pgfx[i]);
		signed short int y=READ_WORD(&c_pgfx[i+2]);

		x&=0x3ff;
		y&=0x3ff;

		x&=0x3ff;
		y&=0x3ff;
		if (x & 0x200) x |= 0xfc00;
		if (y & 0x200) y |= 0xfc00;

		int code=READ_WORD(&c_pgfx[i+4]);
		int colour=READ_WORD(&c_pgfx[i+6]);
		int col=colour&0x1f;

		x-=0x40;
		y-=0x10;

		if (colour & 0xff00)
		{
			int nx=(colour & 0x0f00) >> 8;
			int ny=(colour & 0xf000) >> 12;
			int nxs,nys,sx,sy;
			nx++;
			ny++;

			if (colour & 0x40)
			{
				if (colour &0x20)
				{
					for (nys=0; nys<ny; nys++)
					{
						for (nxs=0; nxs<nx; nxs++)
						{
							sx = x+nxs*16;
							sy = y+nys*16;
							if (do_ss==1)
							{
								cl =sx&0xff;ch=(sx>>8)&0xff;
								fwrite(&ch,1,1,ss); fwrite(&cl,1,1,ss);
								cl =sy&0xff;ch=(sy>>8)&0xff;
								fwrite(&ch,1,1,ss); fwrite(&cl,1,1,ss);
							}
//							if (sy>450) sy-=0x400;
							
							code2=(code+(nx-1)-nxs+0x10*(ny-1-nys));

							if (code2<chset->nSprites)
							{
								chset->printSprite(video,
									code2,col,
									sx,sy,
									1,1);
							}
						}
					}
				}
				else
				{
					for (nys=0; nys<ny; nys++)
					{
						for (nxs=0; nxs<nx; nxs++)
						{
							sx = x+nxs*16;
							sy = y+nys*16;
							if (do_ss==1)
							{
								cl =sx&0xff;ch=(sx>>8)&0xff;
								fwrite(&ch,1,1,ss); fwrite(&cl,1,1,ss);
								cl =sy&0xff;ch=(sy>>8)&0xff;
								fwrite(&ch,1,1,ss); fwrite(&cl,1,1,ss);
							}
//							if (sy>450) sy-=0x400;

							code2=(code+nxs+0x10*(ny-1-nys));

							if (code2<chset->nSprites)
							{
								chset->printSprite(video,
									code2,col,
									sx,sy,
									0,1);
							}
						}
					}
				}
			}
			else
			{
				if (colour &0x20)
				{
					for (nys=0; nys<ny; nys++)
					{
						for (nxs=0; nxs<nx; nxs++)
						{
							sx = x+nxs*16;
							sy = y+nys*16;
							if (do_ss==1)
							{
								cl =sx&0xff;ch=(sx>>8)&0xff;
								fwrite(&ch,1,1,ss); fwrite(&cl,1,1,ss);
								cl =sy&0xff;ch=(sy>>8)&0xff;
								fwrite(&ch,1,1,ss); fwrite(&cl,1,1,ss);
							}
//							if (sy>450) sy-=0x400;

							code2=(code+(nx-1)-nxs+0x10*nys);

							if (code2<chset->nSprites)
							{
								chset->printSprite(video,
									code2,col,
									sx,sy,
									1,0);
							}
						}
					}
				}
				else
				{
					for (nys=0; nys<ny; nys++)
					{
						for (nxs=0; nxs<nx; nxs++)
						{
							sx = x+nxs*16;
							sy = y+nys*16;
							if (do_ss==1)
							{
								cl =sx&0xff;ch=(sx>>8)&0xff;
								fwrite(&ch,1,1,ss); fwrite(&cl,1,1,ss);
								cl =sy&0xff;ch=(sy>>8)&0xff;
								fwrite(&ch,1,1,ss); fwrite(&cl,1,1,ss);
							}
//							if (sy>450) sy-=0x400;

							code2=(code+nxs+0x10*nys);

							if (code2<chset->nSprites)
							{
								chset->printSprite(video,
									code2,col,
									sx,sy,
									0,0);
							}
						}
					}
				}
			}
		}
		else
		{
//			if (y>450) y-=0x400;

			if (code<chset->nSprites)
			{
				chset->printSprite(video,code,col,
					x,y,
					colour&0x20,colour&0x40);
			}
		}
	}

	if (do_ss==1)
	{
		fclose(ss);
	}
}

/* helper draw function */

static void draw_layer(int l,unsigned long int* video,tCharset* chset)
{
	if (l==1)
	{
		cps2cpucore_drawScroll1(video,chset);
	}
	else if (l==2)
	{
		// rowscroll
		unsigned short int* prson=(unsigned short int*)&output_port[0x22];
		int rson=*prson&1;

		if (rson)
		{
			cps2cpucore_drawScroll2rs(video,chset);
		}
		else
		{
			cps2cpucore_drawScroll2(video,chset);
		}
	}
	else if (l==3)
	{
		cps2cpucore_drawScroll3(video,chset);
	}
}


/* main draw function */

void cps2cpucore_drawScreen(unsigned long int* video,tCharset* chset)
{
	// layer control states what's the layer to draw first
	// and in which order the others

	int l0,l1,l2,l3;
	unsigned short int* lcp=(unsigned short int*)&output_port[0x66];

	l0 = ((*lcp) >> 0x06) & 03;
	l1 = ((*lcp) >> 0x08) & 03;
	l2 = ((*lcp) >> 0x0a) & 03;
	l3 = ((*lcp) >> 0x0c) & 03;

	update_controls();
	cpucore_build_palette(chset);

	if (l0!=0) draw_layer(l0,video,chset);
	if (l1!=0) draw_layer(l1,video,chset);
	if (l2!=0) draw_layer(l2,video,chset);
	if (l3!=0) draw_layer(l3,video,chset);

	cps2cpucore_drawSprites(video,chset);
}

