
/*

  a.c.e.
  Another Capcom (Play System) Emulator
  v1.0

  12/5/2001

*/

//#define DEBUG


#include <stdio.h>
#include <stdlib.h>
#include <windows.h>

#include "component.h"
#include "background.h"
#include "sysbar.h"
#include "menu.h"
#include "video32.h"
#include "sound.h"
#include "romspace.h"
#include "charset.h"
#include "cpucore.h"
#include "timing.h"
#include "gui.h"

// the gui

tGui* gui;


extern int goshift;

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	HWND wh;
	MSG msg;

	int secs;
	char strsecs[256];
	int exit=0;
	int frame=0;

	// init gui
 
	gui=new tGui(0,SCRX,SCRY,"8x8.fnt","ace.ini");

	// init display

	if (initDisplay(nCmdShow,&wh,&gui->videoBuf,1,gui->guiinfo.doublesize)) 
	{
		MessageBox(NULL,"can't do video init","error",MB_OK);
		return 1;
	}

	// init timer

	init_timer();

	// init sound

	if (gui->guiinfo.soundon)
	{
		if (initSound(wh,44100))
		{
			MessageBox(NULL,"can't do sound init","error",MB_OK);
			freeDisplay(&gui->videoBuf);
			shutdownSound();
			delete gui;
			return 1;
		}
	}

	// main loop

	while (!exit)
	{
		if (gui->emulating!=0)
		{
#ifndef DEBUG
			if (gui->romsettype==ROMSET_CPS1)
			{
				cps1cpucore_emulate(gui->guiinfo.soundon);
				if ((frame%gui->guiinfo.frameskip)==0) cps1cpucore_drawScreen(gui->videoBuf,gui->chset);
			}
			else if (gui->romsettype==ROMSET_CPS2)
			{
				cps2cpucore_emulate(gui->guiinfo.soundon);
				if ((frame%gui->guiinfo.frameskip)==0) cps2cpucore_drawScreen(gui->videoBuf,gui->chset);
			}

			gui->update();
			gui->redraw(gui->videoBuf);
			if ((frame%gui->guiinfo.frameskip)==0) videoput32(gui->videoBuf);

			frame+=1;
			trim_speed();
#else 
			int r,c;

			int chr=0;
			int nx=8;
			int ny=8;
			int dim=32;

			chr=0;
			for (r=0;r<ny;r++)
			{	
				for (c=0;c<nx;c++)
				{
					gui->chset->printTile32(
									gui->videoBuf,
									chr+goshift*nx,0,
									c*dim,r*dim,SCRX,SCRY,
									0,0);
					chr++;
				}
			}
			gui->update();
			gui->redraw(gui->videoBuf);
			videoput32(gui->videoBuf);
#endif
		}
		else
		{
			gui->update();
			gui->redraw(gui->videoBuf);
			videoput32(gui->videoBuf);
		}

		if (PeekMessage(&msg, NULL,0,0,PM_NOREMOVE))
		{
			GetMessage (&msg, NULL, 0, 0);
			TranslateMessage (&msg);
			DispatchMessage (&msg);
		}

		if ( (msg.message==WM_QUIT) || (gui->exitprogram()) )
		{
			exit=1;
		}
	}

	// show fps
	secs=secs_elapsed();
	sprintf(strsecs,"fps: %f",(float)((float)frame/(float)secs));
	MessageBox(0,strsecs,"info",MB_OK);

	// dispose video

	freeDisplay(&gui->videoBuf);

	// dispose cpu

	if (gui->romsettype==ROMSET_CPS1)
	{
		cps1cpucore_free();
	}
	else if (gui->romsettype==ROMSET_CPS2)
	{
		cps2cpucore_free();
	}

	// dispose soundz

	if (gui->guiinfo.soundon)
	{
		shutdownSound();
	}

	// dispose gui

	delete gui;

	// exit

	return (msg.wParam);
}
