#ifndef ADPCM_H
#define ADPCM_H

#define MAX_ADPCM 8

#define UINT8 unsigned char
#define INT8 signed char
#define UINT16 unsigned short int
#define INT16 signed short int
#define UINT32 unsigned long int
#define INT32 signed long int

/* a generic ADPCM interface, for unknown chips */

struct ADPCMinterface
{
	int num;			       /* total number of ADPCM decoders in the machine */
	int frequency;             /* playback frequency */
	int region;                /* memory region where the samples come from */
	int mixing_level[MAX_ADPCM];     /* master volume */
};

void ADPCMSetBuffers(const struct ADPCMinterface *msound,UINT8 *region,int banksize);
void ADPCMSetBuffersOne(const struct ADPCMinterface *msound,UINT8 *region,int banksize); // Init 1 chip only, 2 banks (wrestlef...)
int ADPCM_sh_start(const struct ADPCMinterface *msound);
void adpcm_update(int num, INT16 *buffer, int length);
void ADPCM_sh_stop(void);
void ADPCM_sh_update(void);

void ADPCM_play(int num, int offset, int length);
void ADPCM_setvol(int num, int vol);
void ADPCM_stop(int num);
int ADPCM_playing(int num);


/* an interface for the OKIM6295 and similar chips */

#define MAX_OKIM6295 			2
#define MAX_OKIM6295_VOICES		4
#define ALL_VOICES				-1

struct OKIM6295interface
{
	int num;                  		/* total number of chips */
	int frequency[MAX_OKIM6295];	/* playback frequency */
	int region[MAX_OKIM6295];		/* memory region where the sample ROM lives */
	int mixing_level[MAX_OKIM6295];	/* master volume */
};

int OKIM6295_sh_start(const struct OKIM6295interface *msound,unsigned char*);
void OKIM6295_bankswitch(int which, int data);
void OKIM6295_sh_stop(void);
void OKIM6295_sh_update(INT16**,int);
void OKIM6295_set_bank_base(int which, int voice, int base);	/* set voice to ALL_VOICES to set all banks at once */
void OKIM6295_set_frequency(int which, int voice, int frequency);	/* set voice to ALL_VOICES to set all banks at once */

int OKIM6295_status_r(int num);
void OKIM6295_data_w(int num, int data);

#endif
