/*

  gui
  the gui class

*/


#ifndef GUI_H
#define GUI_H

#include <windows.h>

#include <vector>

#include "font.h"
#include "component.h"
#include "video32.h"
#include "romspace.h"
#include "charset.h"
#include "cpucore.h"
#include "timing.h"
#include "video32.h"
#include "background.h"
#include "sysbar.h"
#include "menu.h"
#include "selwindow.h"
#include "gawin.h"


// the class
// gui is the glue between different parts
// of the emulation system
// everything should interact with the gui
// and the gui should interact with everything

struct iniInfo
{
	char rompath[1024];
	int soundon;
	int soundfreq;
	int frameskip;
	int videomode;
	int trimspeed;
};


class tGui
{
private:

	unsigned long int* rotobuf;
	tFont* font8x8;
	std::vector<tComponent*> compList;
	int action;

	int state;			// is gui active?
	int guiXsize;
	int guiYsize;

	int openRom(char*);
	int ltrim(char* str);
	int rtrim(char* str);
	void readIniFile(char* inifile);
	int getParameter(char* filename,char* sectionname,
	                       char* paramname,char* paramvalue);

public:

	unsigned long int* videoBuf;
	int emulating;
	int romsettype;
	tRomSpace romset;
	tCharset* chset;
	guiInput ginput;
	iniInfo guiinfo;
	int frames;
	int orientation;
	int changingmode;
	int exitp;

	tGui(int s,int gxs,int gys,char* inifile);
	~tGui();
	void redraw(unsigned long int* vbuf);
	void update();		// reparse input
	int exitprogram();

	void mouseMove(int x,int y);
	void lButtonDown();
	void lButtonUp();
	void rButtonDown();
	void rButtonUp();
};

#endif