
/*

  menu

*/

#include "menu.h"

extern tGui* gui;

tMenuItem::tMenuItem(std::string n,int childm,int fc,int act)
{
	name=n;
	childMenu=childm;
	fontcolor=fc;
	action=act;
}

std::string tMenuItem::getName()
{
	return name;
}

int tMenuItem::getAction()
{
	return action;
}

void tMenuItem::draw(unsigned long int* vbuf,int xp,int yp,int gxs,int gys,tFont* font)
{

}

///

tSubMenu::tSubMenu(int xp,int yp,int col,int gxs,int gys,int act)
{
	xpos=xp;
	ypos=yp;
	color=col;
	guiXsize=gxs;
	guiYsize=gys;
	active=act;
	highlightedItem=-1;
	activatedmenu=-1;
	action=-1;
}

void tSubMenu::update(guiInput* gi)
{
	int maxlen=0;
	for (int j=0;j<itemList.size();j++)
	{
		std::string cs;

		cs=itemList[j].getName();

		if (cs.size()>maxlen)
		{
			maxlen=cs.size();
		}
	}

	int l=(maxlen+2)*8;

	// calc h

	int h=itemList.size()*(8+4);

	//

	int mxp;
	int myp;

	if ( (gui->guiinfo.videomode==1) || (gui->guiinfo.videomode==2) )
	{
		mxp=gi->mouseXpos/2;
		myp=gi->mouseYpos/2+12;
	}
	else if (gui->guiinfo.videomode==0)
	{
		mxp=gi->mouseXpos;
		myp=gi->mouseYpos;
	}
	else if ( (gui->guiinfo.videomode==3) || (gui->guiinfo.videomode==4) )
	{
		mxp=gi->mouseXpos;
		myp=gi->mouseYpos+20;
	}

	int rxp,ryp;

	if (gui->orientation==VIDEO_ROT90)
	{
		rxp=xpos;
		ryp=ypos+90;

		if (gui->guiinfo.videomode==0) 
		{
			rxp+=200;
			ryp+=70;
		}
		else if ( (gui->guiinfo.videomode==3) || (gui->guiinfo.videomode==4) )
		{
			myp-=65;
		}
	}
	else 
	{
		rxp=xpos;
		ryp=ypos;
	}

	if ( (mxp>rxp) && (mxp<(rxp+l)) && (myp>(ryp-h)) && (myp<ryp) )
	{
		// which item is highlighted?

		highlightedItem=((myp-(ryp-h))/(8+4));

		// a click? 8)

		if (gi->mousebutton1==1)
		{
			gi->mousebutton1=0;

			// a normal or father item?

			int ci=getItemChildMenu(highlightedItem);

			if (ci!=-1)
			{
				activatedmenu=ci;								
			}
			// else manage *the action*
			else
			{
				action=itemList[highlightedItem].getAction();
			}
		}

	}
	else
	{
		highlightedItem=-1;
	}
}

void tSubMenu::addItem(tMenuItem &m)
{
	itemList.push_back(m);
}

int tSubMenu::getItemChildMenu(int n)
{
	return itemList[n].childMenu;
}

int tSubMenu::getAction()
{
	return action;
}

std::string tSubMenu::getItem(int n)
{
	return itemList[n].getName();
}

int tSubMenu::nItems()
{
	return itemList.size();
}

void tSubMenu::resetAction()
{
	action=-1;
}

void tSubMenu::draw(unsigned long int* vbuf,tFont* f8x8,guiInput* gi)
{

}

///

tMenu::tMenu(int fc,int gxs,int gys)
{
	guiXsize=gxs;
	guiYsize=gys;
	active=0;
	action=-1;
	fontcolor=fc;
}

void tMenu::update(guiInput* gi)
{
	// check for mouse-covered menus
	// (update only active menus!)

	for (int i=0;i<menuList.size();i++)
	{
		if (menuList[i].active)
		{
			menuList[i].update(gi);

			if (menuList[i].activatedmenu!=-1)
			{
				// deactivate all submenus,except the one active
				int j;

				for (j=i;j<menuList.size();j++)
				{
					if (j!=menuList[i].activatedmenu)
					{
						menuList[j].active=0;
					}
				}

				menuList[menuList[i].activatedmenu].active=!menuList[menuList[i].activatedmenu].active;
				menuList[i].activatedmenu=-1;
			}
			else // there could be an action
			{
				int act=menuList[i].getAction();

				if (act!=-1)
				{
					action=act;
				}
			}
		}
	}

	// if all the items are unhighlighted
	// and there has been a click
	// we should close all menus

	for (i=0;i<menuList.size();i++)
	{
		if (menuList[i].highlightedItem!=-1)
		{
			return ;
		}
	}

	// so close them

	if (gi->mousebutton1==1)
	{
		for (i=0;i<menuList.size();i++)
		{
			menuList[i].active=0;
		}

		active=0;

		gi->mousebutton1=0;
	}
}

void tMenu::addItem(tSubMenu &sm)
{
	menuList.push_back(sm);
}

void tMenu::resetAction()
{
	for (int i=0;i<menuList.size();i++) 
	{
		menuList[i].resetAction();
	}
	action=-1;
}

void tMenu::deactivate()
{
	for (int i=0;i<menuList.size();i++) 
	{
		menuList[i].active=0;
	}
	active=0;	
}

void tMenu::drawRect(int x1,int y1,int x2,int y2,int col,unsigned long int* vbuf)
{
	int ar,ag,ab,r1,g1,b1,r2,g2,b2,rf,gf,bf;

	r1=col&0xff; g1=(col>>8)&0xff; b1=(col>>16)&0xff;

	for (int r=y1;r<y2;r++)
	{
		for (int c=x1;c<x2;c++)
		{
			int i=r*guiXsize+c;

			ar=vbuf[i]&0xff; ag=(vbuf[i]>>8)&0xff; ab=(vbuf[i]>>16)&0xff;
			r2=clamp(r1+0x20); g2=clamp(g1+0x20); b2=clamp(b1+0x20);

			if ( (r==y1) || (r==y2) || (c==x1) || (c==x2) )
			{
				rf=clamp((ar+r2)/2); gf=clamp((ag+g2)/2); bf=clamp((ab+b2)/2);
			}
			else
			{
				rf=clamp((ar+r1)/2); gf=clamp((ag+g1)/2); bf=clamp((ab+b1)/2);
			}

			vbuf[r*guiXsize+c]=rf|(gf<<8)|(bf<<16);
		}
	}
}

void tMenu::draw(unsigned long int* vbuf,tFont* f8x8,guiInput* gi)
{
	// if this component is active, the first menu is active

	if (active)
	{
		menuList[0].active=1;
	}

	for (int m=0;m<menuList.size();m++)
	{
		if (menuList[m].isActive())
		{
			// draw a rectangle
			// rectangle shoud be:
			// l-> the size of the longest menuitem string + n chars
			// h-> the number of strings*(fonty + 4)

			// calc l

			int maxlen=0;

			for (int j=0;j<menuList[m].nItems();j++)
			{
				std::string cs;

				cs=menuList[m].getItem(j);

				if (cs.size()>maxlen)
				{
					maxlen=cs.size();
				}
			}

			int l=(maxlen+2)*f8x8->xsize;

			// calc h

			int h=menuList[m].nItems()*(f8x8->ysize+4);

			int ryp;

			if (gui->orientation==VIDEO_ROT90)
			{
				ryp=menuList[m].ypos+159;
			}
			else 
			{
				ryp=menuList[m].ypos;
			}

			drawRect(menuList[m].xpos,ryp-h,
					 menuList[m].xpos+l,ryp,
					 0x202050,vbuf);

			// one of the itemz is highlighted?

			if (menuList[m].highlightedItem!=-1)
			{
				drawRect(menuList[m].xpos,
						 (menuList[m].highlightedItem+0)*(f8x8->ysize+4)+ryp-h,
						 menuList[m].xpos+l,
						 (menuList[m].highlightedItem+1)*(f8x8->ysize+4)+ryp-h,
						 0x404070,vbuf);
			}

			// draw all the itemz

			for (int i=0;i<menuList[m].nItems();i++)
			{
				int pos=(ryp-h)+i*(f8x8->ysize+4)+2;
				std::string s=menuList[m].getItem(i);

				f8x8->drawString((char*)s.c_str(),fontcolor,menuList[m].xpos+2,pos,guiXsize,guiYsize,vbuf);

				// if the item has childs
				// draw an arrow on the right

				if (menuList[m].getItemChildMenu(i)!=-1)
				{
					f8x8->drawString(".",fontcolor,menuList[m].xpos+2+l-8,pos,guiXsize,guiYsize,vbuf);
				}
			}
		}
	}
}
