
/*

  code to deal with rom reading

*/
 
#include <stdio.h>
#include <string.h>
#include <windows.h>

#include <string>

#include "romspace.h"
#include "romlist.h"
#include "charset.h"


// constructor

tRomSpace::tRomSpace()
{
	// zero point roms

	romCPU1=0;
	romCPU2=0;

	romadpcm=0;
	romqsound=0;
}

// destructor

tRomSpace::~tRomSpace()
{
	if (romCPU1!=0) delete [] romCPU1;
	if (romCPU2!=0) delete [] romCPU2;

	if (romadpcm!=0) delete [] romadpcm;
	if (romqsound!=0) delete [] romqsound;
}

// generic error in rom reading

void tRomSpace::readError(char* zipfile,char* romfile)
{
	char errmsg[1024];

	if (romfile==0)
	{
		sprintf(errmsg,"Error: can't open %s rom.",zipfile);
	}
	else
	{
		sprintf(errmsg,"Error: can't open %s rom in %s zip.",romfile,zipfile);
	}

	MessageBox(0,errmsg,"error",MB_OK);
}

// zipper reader

int tRomSpace::loadZippedRom(char* zipfile,char* romfile,unsigned long int size,unsigned char* dest)
{
	int err;
	unzFile uf=NULL;

	// open zipfile

	uf = unzOpen(zipfile);
	if (uf==NULL)
	{
		readError(zipfile,0);
		return 1;
	}

	err=unzLocateFile(uf,romfile,0);
	if (err!=UNZ_OK)
	{
		readError(zipfile,romfile);
		return 1;
	}

	err=unzOpenCurrentFile(uf);
	if (err!=UNZ_OK)
	{
		readError(zipfile,romfile);
		return 1;
	}

	err=unzReadCurrentFile(uf,dest,size);
	if (err!=size)
	{
		readError(zipfile,romfile);
		return 1;
	}

	unzCloseCurrentFile(uf);

	// close zipfile

	unzClose(uf);

	return 0;
}

// find romset name by zip name

int tRomSpace::romsetByZipname(char* romzip,int* rs)
{
	int sl=strlen(romzip);
	int found=0;
	int pos;
	char rsname[64];

	if ( (sl<5) || (strcmp(&romzip[sl-4],".zip")) )
	{
		MessageBox(0,"not a zip file","error",MB_OK);
		return 1;		
	}

	// find the /

	pos=sl-5;

	while (!found)
	{
		if (romzip[pos]=='\\')
		{
			found=1;
		}
		else 
		{
			pos--;
		}
	}

	if (!found)
	{
		MessageBox(0,"can't find slash","error",MB_OK);
		return 1;
	}

	// create romset name

	strncpy(rsname,&romzip[pos+1],sl-5-pos);
	rsname[sl-5-pos]='\0';

	// see if it identifies a valide romset

	found=0;
	pos=0;

	while ( (found==0) && (strlen(romsetList[pos].name)!=0) )
	{
		if (!strcmp(rsname,romsetList[pos].name))
		{
			found=1;
		}
		else
		{
			pos++;
		}
	}

	if (found==0)
	{
		MessageBox(0,"romset not supported","error",MB_OK);
		return 1;
	}

	*rs=pos;
	return 0;
}

// get graphic rom size

unsigned long int tRomSpace::getGromSize(tRomSet* rsp)
{
	int i;
	unsigned long int max=0;

	for (i=rsp->ncoderoms;i<(rsp->ncoderoms+rsp->ngfxroms);i++)
	{
		if (rsp->romlist[i].size>max)
		{
			max=rsp->romlist[i].size;
		}
	}

	return (max*rsp->ngfxroms);
}


// read roms

int tRomSpace::readRoms(char* romzip,tCharset** chs)
{
	int err;
	int rs;
	long int i,j;
	unsigned char* temprom;
	tRomSet* rsp;

	// we have to decide which romset to load
	// depending on the zipname
	
	err=romsetByZipname(romzip,&rs);
	if (err!=0)
	{
		return 1;
	}

	rsp=&romsetList[rs];

	//
	// read main cpu roms
	// (68000)
	//

	temprom=new unsigned char [0x400000];
	romCPU1=new unsigned char [0x800000];

	memset(romCPU1,0,0x800000);

	for (i=0;i<rsp->ncoderoms;i++)
	{
		// load rom into temprom

		err=loadZippedRom(romzip,
						  rsp->romlist[i].name,
						  rsp->romlist[i].size,
						  temprom);
		if (err!=0)
		{
			MessageBox(0,"can't load code rom","error",MB_OK);
			return 1;
		}

		// type of rom?

		if (rsp->romlist[i].loadtype==ROM_LOAD_ODD)
		{
			for (j=0;j<rsp->romlist[i].size;j++)
			{
				romCPU1[rsp->romlist[i].position+j*2+1]=temprom[j];
			}
		}
		else if (rsp->romlist[i].loadtype==ROM_LOAD_EVEN)
		{
			for (j=0;j<rsp->romlist[i].size;j++)
			{
				romCPU1[rsp->romlist[i].position+j*2]=temprom[j];
			}
		}
		else if (rsp->romlist[i].loadtype==ROM_LOAD_WIDESWAP)
		{
 			memcpy(romCPU1+rsp->romlist[i].position,
				   temprom,
				   rsp->romlist[i].size);			
		}
		else
		{
			MessageBox(0,"invalid rom load format","error",MB_OK);
			return 1;
		}
	}

	//
	// read the graphic roms
	//

	// delete previous charsets
	if (*chs!=0)
	{
		delete *chs;
	}

	if (rsp->romsettype==ROMSET_CPS1)
	{
		*chs=new tCharset(rsp->nspr,rsp->nchars,rsp->ntile16,rsp->ntile32,CHARSET_CPS1);
	}
	else if (rsp->romsettype==ROMSET_CPS2)
	{
		*chs=new tCharset(rsp->nspr,rsp->nchars,rsp->ntile16,rsp->ntile32,CHARSET_CPS2);
	}

	// calculate and store 
	// total size of graphic roms

	tCharset* pchs=*chs;

	gromsize=getGromSize(rsp);

	pchs->gfxrom=new unsigned char [gromsize];

	for (i=rsp->ncoderoms;i<(rsp->ncoderoms+rsp->ngfxroms);i++)
	{
		if (rsp->romlist[i].loadtype==ROM_LOAD_NORMAL)
		{
			err=loadZippedRom(romzip,
							  rsp->romlist[i].name,
							  rsp->romlist[i].size,
							  pchs->gfxrom+rsp->romlist[i].position);
			if (err!=0)
			{
				MessageBox(0,"can't load graphic rom","error",MB_OK);
				return 1;
			}
		}
		else if (rsp->romlist[i].loadtype==ROM_LOAD_SKIP6)
		{
			err=loadZippedRom(romzip,
							  rsp->romlist[i].name,
							  rsp->romlist[i].size,
							  temprom);
			if (err!=0)
			{
				MessageBox(0,"can't load graphic rom","error",MB_OK);
				return 1;
			}

			// load words skipping 6 bytes
			
			for (int p=0;p<(rsp->romlist[i].size);p+=2)
			{
				pchs->gfxrom[rsp->romlist[i].position+(p/2)*8]=  temprom[p];
				pchs->gfxrom[rsp->romlist[i].position+(p/2)*8+1]=temprom[p+1];
			}
		}
	}

	//
	// read sound roms
	//

	if (rsp->romsettype==ROMSET_CPS1)
	{
		// z80

		romCPU2=new unsigned char [0x10000];

		if (rsp->nsndcoderoms>0)
		{
			i=rsp->ncoderoms+rsp->ngfxroms;

			err=loadZippedRom(romzip,
							  rsp->romlist[i].name,
							  rsp->romlist[i].size,
							  romCPU2+rsp->romlist[i].position);
			if (err!=0)
			{
				MessageBox(0,"can't load sound rom","error",MB_OK);
				return 1;
			}
		}

		// adpcm data

		romadpcm=new unsigned char [0x40000];

		if (rsp->nsndsampleroms>0)
		{
			i++;

			err=loadZippedRom(romzip,
							  rsp->romlist[i].name,
							  rsp->romlist[i].size,
							  romadpcm+rsp->romlist[i].position);
			if (err!=0)
			{
				MessageBox(0,"can't load adpcm rom","error",MB_OK);
				return 1;
			}


			i++;

			err=loadZippedRom(romzip,
							  rsp->romlist[i].name,
							  rsp->romlist[i].size,
							  romadpcm+rsp->romlist[i].position);
			if (err!=0)
			{
				MessageBox(0,"can't load adpcm rom","error",MB_OK);
				return 1;
			}
		}
	}
	else if (rsp->romsettype==ROMSET_CPS2)
	{
		// z80
		romCPU2=new unsigned char [0x40000];

		if (rsp->nsndcoderoms>0)
		{
			for (i=rsp->ncoderoms+rsp->ngfxroms;i<rsp->ncoderoms+rsp->ngfxroms+rsp->nsndcoderoms;i++)
			{
				err=loadZippedRom(romzip,
								  rsp->romlist[i].name,
								  rsp->romlist[i].size,
								  romCPU2+rsp->romlist[i].position);

				if (err!=0)
				{
					MessageBox(0,"can't load z80 sound rom","error",MB_OK);
					return 1;
				}
			}
		}

		// qsound samples

		romqsound=new unsigned char [0x400000];

		if (rsp->nsndsampleroms>0)
		{
			int frm=rsp->ncoderoms+rsp->ngfxroms+rsp->nsndcoderoms;
			int to=rsp->ncoderoms+rsp->ngfxroms+rsp->nsndcoderoms+rsp->nsndsampleroms;

			for (i=frm;i<to;i++)
			{
				err=loadZippedRom(romzip,
								  rsp->romlist[i].name,
								  rsp->romlist[i].size,
								  temprom);

				if (rsp->romlist[i].loadtype==ROM_LOAD_BYTESWAP)
				{
					for (int j=0;j<rsp->romlist[i].size;j+=2)
					{
						unsigned char temp=temprom[j];
						temprom[j]=temprom[j+1];
						temprom[j+1]=temp;
					}
				}

				memcpy(&romqsound[rsp->romlist[i].position],
					   temprom,rsp->romlist[i].size);

				if (err!=0)
				{
					MessageBox(0,"can't load 1st qsound sample rom","error",MB_OK);
					return 1;
				}
			}
		}
	}

	// store emulation related data

	kludge=rsp->kludge;
	orientation=rsp->orientation;
	romsettype=rsp->romsettype;
	nspr=rsp->nspr;
	nchars=rsp->nchars;
	ntile16=rsp->ntile16;
	ntile32=rsp->ntile32;

	sprmagic1=rsp->sprmagic1; sprmagic2=rsp->sprmagic2;
	charmagic1=rsp->charmagic1;
	tile16magic1=rsp->tile16magic1; tile16magic2=rsp->tile16magic2; 
	tile32magic1=rsp->tile32magic1; tile32magic2=rsp->tile32magic2; 

	sproffs=rsp->sproffs;
	charoffs=rsp->charoffs;
	tile16offs=rsp->tile16offs;
	tile32offs=rsp->tile32offs;

	charbase=rsp->charbase;
	tile16base=rsp->tile16base;
	tile32base=rsp->tile32base;

	charpad=rsp->charpad;
	tile16pad=rsp->tile16pad;
	tile32pad=rsp->tile32pad;

	protport=rsp->protport;
	protval=rsp->protval;
	multlo=rsp->multlo;
	multhi=rsp->multhi;
	fac1=rsp->fac1;
	fac2=rsp->fac2;
	layercontrol=rsp->layercontrol;

	// free space

	delete [] temprom;

	return 0;
}

/*

  functions to decode the charset

*/

int tRomSpace::readbit(const unsigned char *src,int bitnum)
{
	return (src[bitnum / 8] >> (7 - bitnum % 8)) & 1;
}

void tRomSpace::decodesprites(int num,unsigned char *src,unsigned char** dest)
{
/*SPRITE_LAYOUT(ffight, 4096*2+512, 0x80000*8, 0x100000*8)*/

	int plane,x,y;
	unsigned char* dp;
	unsigned char* tmpdp;
	int nplanes=4;
	int charincrement=32*8;

	const long int f2=sprmagic1;
	const long int f4=sprmagic2;

	long int planeoffset[4]={ f2+8,f2+0, 8, 0 };

	long int xoffset[16]=	{ f4+0,f4+1,f4+2,f4+3,f4+4,f4+5,f4+6,f4+7,
							  0, 1, 2, 3, 4, 5, 6, 7 };
	long int yoffset[16]=	        { 0*8, 2*8, 4*8, 6*8, 8*8, 10*8, 12*8, 14*8,
		           16*8, 18*8, 20*8, 22*8, 24*8, 26*8, 28*8, 30*8, };


	for (plane = 0;plane < nplanes;plane++)
	{
		int offs;

		offs = num*charincrement+planeoffset[plane];

		for (y = 0;y < 16;y++)
		{
			//dp = gfx->gfxdata->line[num * 16 + y];
			//dp=dest[num]+(y*16);
			tmpdp=dest[num];
			dp=tmpdp+y*16;

			for (x = 0;x < 16;x++)
			{
				int xoffs,yoffs;

				if (plane == 0) dp[x] = 0;
				else dp[x] <<= 1;

				xoffs = x;
				yoffs = y;

				dp[x] += readbit(src,offs + yoffset[yoffs] + xoffset[xoffs]);
			}
		}
	}
}

//

void tRomSpace::decodechars(int num,unsigned char *src,unsigned char** dest)
{
	int plane,x,y;
	unsigned char* dp;
	unsigned char* tmpdp;
	int nplanes=4;
	int charincrement=16*8;

	const long int f2=charmagic1;

	long int planeoffset[4]={ f2+8,f2+0, 8, 0 };
	long int xoffset[8]={0,1,2,3,4,5,6,7};
	long int yoffset[8]={ 0*8, 2*8, 4*8, 6*8, 8*8, 10*8, 12*8, 14*8};

	unsigned char* psrc=src+charoffs;

	for (plane = 0;plane < nplanes;plane++)
	{
		int offs;

		offs = num*charincrement+planeoffset[plane];

		for (y = 0;y < 8;y++)
		{
			tmpdp=dest[num];
			dp=tmpdp+y*8;

			for (x = 0;x < 8;x++)
			{
				int xoffs,yoffs;

				if (plane == 0) dp[x] = 0;
				else dp[x] <<= 1;

				xoffs = x;
				yoffs = y;

				dp[x] += readbit(psrc,offs + yoffset[yoffs] + xoffset[xoffs]);
			}
		}
	}
}

//

void tRomSpace::decodetiles16(int num,unsigned char *src,unsigned char** dest)
{
	int plane,x,y;
	unsigned char* dp;
	unsigned char* tmpdp;
	int nplanes=4;
	int charincrement=32*8;

	const long int f2=tile16magic1;
	const long int f4=tile16magic2;

	long int planeoffset[4]={ f2+8,f2+0, 8, 0 };

	long int xoffset[16]=	{ f4+0,f4+1,f4+2,f4+3,f4+4,f4+5,f4+6,f4+7,
							  0, 1, 2, 3, 4, 5, 6, 7 };
	long int yoffset[16]=	        { 0*8, 2*8, 4*8, 6*8, 8*8, 10*8, 12*8, 14*8,
		           16*8, 18*8, 20*8, 22*8, 24*8, 26*8, 28*8, 30*8, };

	unsigned char* psrc=src+tile16offs;

	for (plane = 0;plane < nplanes;plane++)
	{
		int offs;

		offs = num*charincrement+planeoffset[plane];

		for (y = 0;y < 16;y++)
		{
			tmpdp=dest[num];
			dp=tmpdp+y*16;

			for (x = 0;x < 16;x++)
			{
				int xoffs,yoffs;

				if (plane == 0) dp[x] = 0;
				else dp[x] <<= 1;

				xoffs = x;
				yoffs = y;

				dp[x] += readbit(psrc,offs + yoffset[yoffs] + xoffset[xoffs]);
			}
		}
	}
}


//

/*
TILE32_LAYOUT(tile32layout_ffight, 512+128,     0x080000*8, 0x100000*8)
*/

void tRomSpace::decodetiles32(int num,unsigned char *src,unsigned char** dest)
{
	int plane,x,y;
	unsigned char* dp;
	unsigned char* tmpdp;
	int nplanes=4;
	int charincrement=32*8*4;

	const long int f2=tile32magic1;
	const long int SEP=tile32magic2;

	long int planeoffset[4]={ f2+8,f2+0, 8, 0 };

	long int xoffset[32]={                                                   
					SEP+0,SEP+1,SEP+2,SEP+3, SEP+4,SEP+5,SEP+6,SEP+7,
					0,1,2,3,4,5,6,7,                                 
					16+SEP+0,16+SEP+1,16+SEP+2,                      
					16+SEP+3,16+SEP+4,16+SEP+5,                      
					16+SEP+6,16+SEP+7,                               
					16+0,16+1,16+2,16+3,16+4,16+5,16+6,16+7          
					};                                               

	long int yoffset[32]=
		{                                                          
           0*32, 1*32, 2*32, 3*32, 4*32, 5*32, 6*32, 7*32,         
           8*32, 9*32, 10*32, 11*32, 12*32, 13*32, 14*32, 15*32,   
           16*32, 17*32, 18*32, 19*32, 20*32, 21*32, 22*32, 23*32, 
           24*32, 25*32, 26*32, 27*32, 28*32, 29*32, 30*32, 31*32  
        };

	unsigned char* psrc=src+tile32offs;

	for (plane = 0;plane < nplanes;plane++)
	{
		int offs;

		offs = num*charincrement+planeoffset[plane];

		for (y = 0;y < 32;y++)
		{
			tmpdp=dest[num];
			dp=tmpdp+y*32;

			for (x = 0;x < 32;x++)
			{
				int xoffs,yoffs;

				if (plane == 0) dp[x] = 0;
				else dp[x] <<= 1;

				xoffs = x;
				yoffs = y;

				dp[x] += readbit(psrc,offs + yoffset[yoffs] + xoffset[xoffs]);
			}
		}
	}
}

//
// CPS-2 (makes always cool to write it)
//

void tRomSpace::cps2decodesprites(int num,unsigned char *src,unsigned char** dest)
{
	int x,y;
	unsigned char* psrc=&src[num*16*8+sproffs];

	for (y = 0;y < 16;y++)
	{
		for (x = 0;x < 16;x+=2)
		{
			dest[num][x+y*16]=(*psrc)&0x0f;
			dest[num][x+1+y*16]=(*psrc>>4)&0x0f;
			psrc++;
		}
	}
}

void tRomSpace::cps2decodechars(int num,unsigned char *src,unsigned char** dest)
{
	int x,y;
	unsigned char* psrc=&src[num*8*8+charoffs];

	for (y = 0;y < 8;y++)
	{
		for (x = 0;x < 8;x+=2)
		{
			dest[num][x+y*8]=(*psrc)&0x0f;
			dest[num][x+1+y*8]=((*psrc)>>4)&0x0f;
			psrc++;
		}
		psrc+=4;
	}
}

void tRomSpace::cps2decodetiles16(int num,unsigned char *src,unsigned char** dest)
{
	int x,y;
	unsigned char* psrc=&src[num*16*8+tile16offs];

	for (y = 0;y < 16;y++)
	{
		for (x = 0;x < 16;x+=2)
		{
			dest[num][x+y*16]=(*psrc)&0x0f;
			dest[num][x+1+y*16]=(*psrc>>4)&0x0f;
			psrc++;
		}
	}
}

void tRomSpace::cps2decodetiles32(int num,unsigned char *src,unsigned char** dest)
{
	int x,y;
	unsigned char* psrc=&src[num*64*8+tile32offs];

	for (y = 0;y < 32;y++)
	{
		for (x = 0;x < 32;x+=2)
		{
			dest[num][x+y*32]=(*psrc)&0x0f;
			dest[num][x+1+y*32]=(*psrc>>4)&0x0f;
			psrc++;
		}
	}
}


//

void tRomSpace::cps1_gfx_decode(unsigned char* gRom)
{
	int size=gromsize;
	int i,j,gfxsize;
	unsigned char *cps1_gfx = gRom;


	gfxsize=gromsize/4;

	for (i = 0;i < gfxsize;i++)
	{
		unsigned long int src = cps1_gfx[4*i] + (cps1_gfx[4*i+1]<<8) + (cps1_gfx[4*i+2]<<16) + (cps1_gfx[4*i+3]<<24);
		unsigned long int dwval = 0;

		for (j = 0;j < 8;j++)
		{
			int n = 0;
			unsigned long int mask = (0x80808080 >> j) & src;

			if (mask & 0x000000ff) n |= 1;
			if (mask & 0x0000ff00) n |= 2;
			if (mask & 0x00ff0000) n |= 4;
			if (mask & 0xff000000) n |= 8;

			dwval |= n << (j * 4);
		}
		cps1_gfx[4*i  ] = dwval>>0;
		cps1_gfx[4*i+1] = dwval>>8;
		cps1_gfx[4*i+2] = dwval>>16;
		cps1_gfx[4*i+3] = dwval>>24;
	}
}


//

void tRomSpace::unshuffle(__int64 *buf,int len)
{
	int i;
	__int64 t;

	if (len == 2) return;

	if (len % 4) return;   /* must not happen */

	len /= 2;

	unshuffle(buf,len);
	unshuffle(buf + len,len);

	for (i = 0;i < len/2;i++)
	{
		t = buf[len/2 + i];
		buf[len/2 + i] = buf[len + i];
		buf[len + i] = t;
	}
}

//

void tRomSpace::decodeCharset(tCharset* chset,int rst)
{
	int c;

	//decode:
	// sprites (16x16)
	// chars (8x8)
	// tiles (16x16)
	// tiles (32x32)

	if (rst==ROMSET_CPS1)
	{
		for (c = 0;c < chset->nSprites;c++)
		{
			decodesprites(c,chset->gfxrom,chset->sprites);
		}

		for (c = 0;c < chset->nChars;c++)
		{
			decodechars(c,chset->gfxrom,chset->chars);
		}

		for (c = 0;c < chset->nTiles16;c++)
		{
			decodetiles16(c,chset->gfxrom,chset->tiles16);
		}

		for (c = 0;c < chset->nTiles32;c++)
		{
			decodetiles32(c,chset->gfxrom,chset->tiles32);
		}
	}
	else if (rst==ROMSET_CPS2)
	{
		for (int i = 0;i < gromsize;i += 0x200000)
		{
			unshuffle((__int64 *)(chset->gfxrom + i),0x200000/8);
		}

		cps1_gfx_decode(chset->gfxrom);

		chset->sproffs=sproffs;
		chset->charoffs=charoffs;
		chset->tile16offs=tile16offs;
		chset->tile32offs=tile32offs;
	}
}