
/*

  charset implementation

*/


#include "charset.h"


//#define DEBUG

tCharset::tCharset(int nspr,int nchr,int nt16,int nt32,int type)
{
	long int c;

	gfxrom=0;
	chtype=type;
	nSprites=nspr;
	nChars=nchr;
	nTiles16=nt16;
	nTiles32=nt32;

	sprites=0;
	chars=0;
	tiles16=0;
	tiles32=0;

	if (chtype==CHARSET_CPS1)
	{
		// allocate graphic charsets memory space

		sprites=new unsigned char* [nSprites];
		chars=new unsigned char* [nChars];
		tiles16=new unsigned char* [nTiles16];
		tiles32=new unsigned char* [nTiles32];

		for (c=0;c<nSprites;c++)
		{
			sprites[c]=new unsigned char [16*16];
			memset(sprites[c],0,16*16);
		}
		for (c=0;c<nChars;c++)
		{
			chars[c]=new unsigned char [8*8];
			memset(chars[c],0,8*8);
		}
		for (c=0;c<nTiles16;c++)
		{
			tiles16[c]=new unsigned char [16*16];
			memset(tiles16[c],0,16*16);
		}
		for (c=0;c<nTiles32;c++)
		{
			tiles32[c]=new unsigned char [32*32];
			memset(tiles32[c],0,32*32);
		}
	}
}

tCharset::~tCharset()
{
	if (chtype==CHARSET_CPS1)
	{
		// free decoded charsets

		if (sprites!=0)
		{
			for (int j=0;j<nSprites;j++)
			{
				if (sprites[j]!=0) delete [] sprites[j];
			}

			delete [] sprites;
		}

		if (chars!=0)
		{
			for (int j=0;j<nChars;j++)
			{
				if (chars[j]!=0) delete [] chars[j];
			}

			delete [] chars;
		}

		if (tiles16!=0)
		{
			for (int j=0;j<nTiles16;j++)
			{
				if (tiles16[j]!=0) delete [] tiles16[j];
			}

			delete [] tiles16;
		}

		if (tiles32!=0)
		{
			for (int j=0;j<nTiles32;j++)
			{
				if (tiles32[j]!=0) delete [] tiles32[j];
			}

			delete [] tiles32;
		}
	}
	else if (chtype==CHARSET_CPS2)
	{
		if (gfxrom!=0) 
		{
			delete [] gfxrom;
		}
	}
}

void tCharset::printSprite(unsigned long int* vb,
							  int charnum,int color,
							  int xp,int yp,
							  int flipx,int flipy)
{
	if ( (xp<-16) || (xp>=384) || (yp<-16) || (yp>=224) ) return;

	unsigned long int* pbmp=&vb[yp*384+xp];
	unsigned char* pchar=sprites[charnum];
	unsigned long int* ppal=(unsigned long int*)palette;
	int x,y;
	int c;

	color*=16;

	// have we to do clipping?
	if ( (xp>=0) && (xp<(384-16)) && (yp>=0) && (yp<(224-16)) )
	{
		if ((!flipx) && (!flipy))
		{
			for (y=0;y<16;y++)
			{
				for (x=0;x<16;x++)
				{
					c=*pchar;

					if (c!=15)
					{
#ifdef DEBUG
					*pbmp=(c*16)|((c*16)<<8)|((c*16)<<16);
#else
						*pbmp=ppal[c+color];
#endif
					}
					pbmp++;
					pchar++;
				}
				pbmp+=(384-16);
			}
		}
		else if ((!flipx) && (flipy))
		{
			pchar+=(15*16);
			for (y=0;y<16;y++)
			{
				for (x=0;x<16;x++)
				{
					//c=pchar[x+(15-y)*16];
					c=*pchar;

					if (c!=15)
					{
						*pbmp=ppal[c+color];
					}
					pbmp++;
					pchar++;
				}
				pbmp+=(384-16);
				pchar-=32;
			}
		}
		else if ((flipx) && (!flipy))
		{
			pchar+=15;
			for (y=0;y<16;y++)
			{
				for (x=0;x<16;x++)
				{
					//c=pchar[(15-x)+y*16];
					c=*pchar;

					if (c!=15)
					{
						*pbmp=ppal[c+color];
					}
					pbmp++;
					pchar--;
				}
				pbmp+=(384-16);
				pchar+=32;
			}
		}
		else
		{
			pchar+=(15*16)+15;
			for (y=0;y<16;y++)
			{
				for (x=0;x<16;x++)
				{
					//c=pchar[(15-x)+(15-y)*16];
					c=*pchar;

					if (c!=15)
					{
						*pbmp=ppal[c+color];
					}
					pbmp++;
					pchar--;
				}
				pbmp+=(384-16);
			}
		}
	}
	else
	// clip
	{
		if ((!flipx) && (!flipy))
		{
			for (y=0;y<16;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<16;x++)
				{
					c=*pchar;

					if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
					else if (c!=15)
					{
#ifdef DEBUG
					*pbmp=(c*16)|((c*16)<<8)|((c*16)<<16);
#else
						*pbmp=ppal[c+color];
#endif
					}
					pbmp++;
					pchar++;
				}
				pbmp+=(384-16);
			}
		}
		else if ((!flipx) && (flipy))
		{
			pchar+=15*16;
			for (y=0;y<16;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<16;x++)
				{
					c=*pchar;

					if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
					else if (c!=15)
					{
						*pbmp=ppal[c+color];
					}
					pbmp++;
					pchar++;
				}
				pbmp+=(384-16);
				pchar-=32;
			}
		}
		else if ((flipx) && (!flipy))
		{
			pchar+=15;
			for (y=0;y<16;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<16;x++)
				{
					c=*pchar;

					if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
					else if (c!=15)
					{
						*pbmp=ppal[c+color];
					}
					pbmp++;
					pchar--;
				}
				pbmp+=(384-16);
				pchar+=32;
			}
		}
		else
		{
			pchar+=(15*16+15);
			for (y=0;y<16;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<16;x++)
				{
					c=*pchar;

					if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
					else if (c!=15)
					{
						*pbmp=ppal[c+color];
					}
					pbmp++;
					pchar--;
				}
				pbmp+=(384-16);
			}
		}
	}
}

void tCharset::printChar(unsigned long int* vb,
							  int charnum,int color,
							  int xp,int yp,
							  int flipx,int flipy)
{
	unsigned long int* pbmp=&vb[yp*384+xp];
	unsigned char* pchar=chars[charnum];
	int x,y;
	int c;

	unsigned long int* pcpal=(unsigned long int*)&palette[0x200*4];
	color*=16;

	if ((!flipx) && (!flipy))
	{
		for (y=0;y<8;y++)
		{
			if ((yp+y)>=224) return;
			for (x=0;x<8;x++)
			{
				c=*pchar;

				if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
				else if (c!=15)
				{
#ifdef DEBUG
					*pbmp=(c*16)|((c*16)<<8)|((c*16)<<16);
#else
					*pbmp=pcpal[c+color];
#endif
				}
				pbmp++;
				pchar++;
			}
			pbmp+=(384-8);
		}
	}
	else if ((!flipx) && (flipy))
	{
		pchar+=7*8;
		for (y=0;y<8;y++)
		{
			if ((yp+y)>=224) return;
			for (x=0;x<8;x++)
			{
				c=*pchar;

				if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
				else if (c!=15)
				{
					*pbmp=pcpal[c+color];
				}
				pbmp++;
				pchar++;
			}
			pbmp+=(384-8);
			pchar-=16;
		}
	}
	else if ((flipx) && (!flipy))
	{
		pchar+=7;
		for (y=0;y<8;y++)
		{
			if ((yp+y)>=224) return;
			for (x=0;x<8;x++)
			{
				c=*pchar;

				if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
				else if (c!=15)
				{
					*pbmp=pcpal[c+color];
				}
				pbmp++;
				pchar--;
			}
			pbmp+=(384-8);
			pchar+=16;
		}
	}
	else
	{
		pchar+=(7*8+7);
		for (y=0;y<8;y++)
		{
			if ((yp+y)>=224) return;
			for (x=0;x<8;x++)
			{
				c=*pchar;

				if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
				else if (c!=15)
				{
					*pbmp=pcpal[c+color];
				}
				pbmp++;
				pchar--;
			}
			pbmp+=(384-8);
		}
	}
}

void tCharset::printTile16(unsigned long int* vb,
							  int charnum,int color,
							  int xp,int yp,
							  int flipx,int flipy)
{
	if ( (xp<-15) || (xp>=384) || (yp<-15) || (yp>=224) ) return;

	unsigned long int* pbmp=&vb[yp*384+xp];
	unsigned char* pchar=tiles16[charnum];
	int x,y;
	int c;

	unsigned long int* ptpal=(unsigned long int*)&palette[0x400*4];
	color*=16;

	// have we to do clipping?
	if ( (xp>=0) && (xp<(384-16)) && (yp>=0) && (yp<(224-16)) )
	{
		if ((!flipx) && (!flipy))
		{
			for (y=0;y<16;y++)
			{
				for (x=0;x<16;x++)
				{
					c=*pchar;

					if (c!=15)
					{
						*pbmp=ptpal[c+color];
#ifdef DEBUG
					*pbmp=(c*16)|((c*16)<<8)|((c*16)<<16);
#endif
					}
					pbmp++;
					pchar++;
				}
				pbmp+=(384-16);
			}
		}
		else if ((!flipx) && (flipy))
		{
			pchar+=(15*16);
			for (y=0;y<16;y++)
			{
				for (x=0;x<16;x++)
				{
					c=*pchar;

					if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					pchar++;
				}
				pchar-=32;
				pbmp+=(384-16);
			}
		}
		else if ((flipx) && (!flipy))
		{
			pchar+=15;
			for (y=0;y<16;y++)
			{
				for (x=0;x<16;x++)
				{
					c=*pchar;

					if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					pchar--;
				}
				pbmp+=(384-16);
				pchar+=32;
			}
		}
		else
		{
			pchar+=(15*16)+15;
			for (y=0;y<16;y++)
			{
				for (x=0;x<16;x++)
				{
					c=*pchar;

					if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					pchar--;
				}
				pbmp+=(384-16);
			}
		}
	}
	else
	// clip blit
	{
		if ((!flipx) && (!flipy))
		{
			for (y=0;y<16;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<16;x++)
				{
					c=*pchar;

					if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
					else if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					pchar++;
				}
				pbmp+=(384-16);
			}
		}
		else if ((!flipx) && (flipy))
		{
			pchar+=(15*16);
			for (y=0;y<16;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<16;x++)
				{
					c=*pchar;

					if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
					else if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					pchar++;
				}
				pchar-=32;
				pbmp+=(384-16);
			}
		}
		else if ((flipx) && (!flipy))
		{
			pchar+=15;
			for (y=0;y<16;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<16;x++)
				{
					c=*pchar;

					if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
					else if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					pchar--;
				}
				pbmp+=(384-16);
				pchar+=32;
			}
		}
		else
		{
			pchar+=(15*16)+15;
			for (y=0;y<16;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<16;x++)
				{
					c=*pchar;

					if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
					else if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					pchar--;
				}
				pbmp+=(384-16);
			}
		}
	}
}

//

void tCharset::printTile16rs(unsigned long int* vb,
							  int charnum,int color,
							  int xp,int yp,
							  int flipx,int flipy,int* rs,int ofz)
{
	if ( (yp<-15) || (yp>223) ) return;

	unsigned char* pchar=tiles16[charnum];
	unsigned long int* vbp;
	int x,y;
	int c;
	int rso;
	unsigned long int* ptpal=(unsigned long int*)&palette[0x400*4];

	color*=16;

	if ((!flipx) && (!flipy))
	{
		for (y=0;y<16;y++)
		{
			int ypp=yp+y;
			if (ypp>=224) return;
			rso=rs[ypp+ofz];
			int xpp=xp+rso;

			if (ypp>=0)
			{
				vbp=&vb[xpp+ypp*384];
				for (x=0;x<16;x++)
				{
					c=*pchar;

					if ( (xpp<0) || (xpp>=384) ) ;
					else if (c!=15)
					{
						*vbp=ptpal[c+color];
					}
					pchar++;
					xpp++;
					vbp++;
				}
			}
			else
			{
				pchar+=16;
			}
		}
	}
	else if ((!flipx) && (flipy))
	{
		pchar+=(15*16);
		for (y=0;y<16;y++)
		{
			int ypp=yp+y;
			if (ypp>=224) return;
			rso=rs[ypp+ofz];
			int xpp=xp+rso;

			if (ypp>=0)
			{
				vbp=&vb[xpp+ypp*384];
				for (x=0;x<16;x++)
				{
					c=*pchar;

					if ( (xpp<0) || (xpp>=384) ) ;
					else if (c!=15)
					{
						*vbp=ptpal[c+color];
					}
					pchar++;
					xpp++;
					vbp++;
				}
				pchar-=32;
			}
			else
			{
				pchar-=16;
			}
		}
	}
	else if ((flipx) && (!flipy))
	{
		pchar+=15;
		for (y=0;y<16;y++)
		{
			int ypp=yp+y;
			if (ypp>=224) return;
			rso=rs[ypp+ofz];
			int xpp=xp+rso;

			if (ypp>=0)
			{
				vbp=&vb[xpp+ypp*384];
				for (x=0;x<16;x++)
				{
					c=*pchar;

					if ( (xpp<0) || (xpp>=384) ) ;
					else if (c!=15)
					{
						*vbp=ptpal[c+color];
					}
					pchar--;
					xpp++;
					vbp++;
				}
			}
			else
			{
				pchar-=16;
			}
			pchar+=32;
		}
	}
	else
	{
		pchar+=(15*16)+15;
		for (y=0;y<16;y++)
		{
			int ypp=yp+y;
			if (ypp>=224) return;
			rso=rs[ypp+ofz];
			int xpp=xp+rso;

			if (ypp>=0)
			{
				vbp=&vb[xpp+ypp*384];
				for (x=0;x<16;x++)
				{
					c=*pchar;

					if ( (xpp<0) || (xpp>=384) ) ;
					else if (c!=15)
					{
						*vbp=ptpal[c+color];
					}
					pchar--;
					xpp++;
					vbp++;
				}
			}
			else
			{
				pchar-=16;
			}
		}
	}
}



//

void tCharset::printTile32(unsigned long int* vb,
							  int charnum,int color,
							  int xp,int yp,
							  int flipx,int flipy)
{
	if ( (xp<-31) || (xp>=384) || (yp<-31) || (yp>=224) ) return;

	unsigned long int* pbmp=&vb[yp*384+xp];
	unsigned char* pchar=tiles32[charnum];
	unsigned long int* ptpal=(unsigned long int*)&palette[0x600*4];
	int x,y;
	int c;
	color*=16;

	// have we to do clipping?
	if ( (xp>=0) && (xp<(384-32)) && (yp>=0) && (yp<(224-32)) )
	{
		//noclip
		if ((!flipx) && (!flipy))
		{
			for (y=0;y<32;y++)
			{
				for (x=0;x<32;x++)
				{
					c=*pchar;

					if (c!=15)
					{
						*pbmp=ptpal[c+color];
#ifdef DEBUG
						*pbmp=(c*16)|((c*16)<<8)|((c*16)<<16);
#endif
					}
					pbmp++;
					pchar++;
				}
				pbmp+=(384-32);
			}
		}
		else if ((!flipx) && (flipy))
		{
			pchar+=(31*32);
			for (y=0;y<32;y++)
			{
				for (x=0;x<32;x++)
				{
					c=*pchar;

					if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					pchar++;
				}
				pbmp+=(384-32);
				pchar-=64;
			}
		}
		else if ((flipx) && (!flipy))
		{
			pchar+=31;
			for (y=0;y<32;y++)
			{
				for (x=0;x<32;x++)
				{
					c=*pchar;

					if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					pchar--;
				}
				pbmp+=(384-32);
				pchar+=64;
			}
		}
		else
		{
			pchar+=(31*32)+31;
			for (y=0;y<32;y++)
			{
				for (x=0;x<32;x++)
				{
					c=*pchar;

					if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					pchar--;
				}
				pbmp+=(384-32);
			}
		}
	// clip
	}
	else
	{
		if ((!flipx) && (!flipy))
		{
			for (y=0;y<32;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<32;x++)
				{
					c=*pchar;

					if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
					else if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					pchar++;
				}
				pbmp+=(384-32);
			}
		}
		else if ((!flipx) && (flipy))
		{
			pchar+=(31*32);
			for (y=0;y<32;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<32;x++)
				{
					c=*pchar;

					if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
					else if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					pchar++;
				}
				pbmp+=(384-32);
				pchar-=64;
			}
		}
		else if ((flipx) && (!flipy))
		{
			pchar+=31;
			for (y=0;y<32;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<32;x++)
				{
					c=*pchar;

					if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
					else if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					pchar--;
				}
				pbmp+=(384-32);
				pchar+=64;
			}
		}
		else
		{
			pchar+=(31*32)+31;
			for (y=0;y<32;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<32;x++)
				{
					c=*pchar;

					if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
					else if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					pchar--;
				}
				pbmp+=(384-32);
			}
		}
	}
}

///////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////

void tCharset::cps2printChar(unsigned long int* vb,
							  int charnum,int color,
							  int xp,int yp,
							  int flipx,int flipy)
{
	unsigned long int* pbmp=&vb[yp*384+xp];
	unsigned char* pchar=&gfxrom[charnum*8*8+charoffs];
	int x,y;
	int c,cp;

	unsigned long int* pcpal=(unsigned long int*)&palette[0x200*4];
	color*=16;

	if ( (xp>=0) && (xp<(384-16)) && (yp>=0) && (yp<(224-16)) )
	{
		if ((!flipx) && (!flipy))
		{
			for (y=0;y<8;y++)
			{
				for (x=0;x<8;x+=2)
				{
					cp=*pchar;
					c=cp&0x0f;

					if (c!=15)
					{
	#ifdef DEBUG
						*pbmp=(c*16)|((c*16)<<8)|((c*16)<<16);
	#else
						*pbmp=pcpal[c+color];
	#endif
					}
					pbmp++;

					c=cp>>4;

					if (c!=15)
					{
	#ifdef DEBUG
						*pbmp=(c*16)|((c*16)<<8)|((c*16)<<16);
	#else
						*pbmp=pcpal[c+color];
	#endif
					}
					pbmp++;

					pchar++;
				}
				pchar+=4;
				pbmp+=(384-8);
			}
		}
		else if ((!flipx) && (flipy))
		{
			pchar+=(7*8);
			for (y=0;y<8;y++)
			{
				for (x=0;x<8;x+=2)
				{
					cp=*pchar;
					c=cp&0x0f;

					if (c!=15)
					{
						*pbmp=pcpal[c+color];
					}
					pbmp++;

					c=cp>>4;

					if (c!=15)
					{
						*pbmp=pcpal[c+color];
					}
					pbmp++;

					pchar++;
				}
				pbmp+=(384-8);
	//			pchar-=16;
				pchar-=12;
			}
		}
		else if ((flipx) && (!flipy))
		{
			pchar+=3;
			for (y=0;y<8;y++)
			{
				for (x=0;x<8;x+=2)
				{
					cp=*pchar;
					c=cp>>4;

					if (c!=15)
					{
						*pbmp=pcpal[c+color];
					}
					pbmp++;

					c=cp&0x0f;

					if (c!=15)
					{
						*pbmp=pcpal[c+color];
					}
					pbmp++;

					pchar--;
				}
				pbmp+=(384-8);
	//			pchar+=16;
				pchar+=12;
			}
		}
		else
		{
	//		pchar+=(7*8+7);
			pchar+=(7*8+3);
			for (y=0;y<8;y++)
			{
				for (x=0;x<8;x+=2)
				{
					cp=*pchar;
					c=cp>>4;

					if (c!=15)
					{
						*pbmp=pcpal[c+color];
					}
					pbmp++;

					c=cp&0x0f;

					if (c!=15)
					{
						*pbmp=pcpal[c+color];
					}
					pbmp++;

					pchar--;
				}
				pbmp+=(384-8);
				pchar-=4;
			}
		}
	}
	// clip
	else
	{
		if ((!flipx) && (!flipy))
		{
			for (y=0;y<8;y++)
			{
				if ((yp+y)>=224) return;
	//			for (x=0;x<8;x++)
				for (x=0;x<8;x+=2)
				{
					cp=*pchar;
					c=cp&0x0f;

					if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
					else if (c!=15)
					{
	#ifdef DEBUG
						*pbmp=(c*16)|((c*16)<<8)|((c*16)<<16);
	#else
						*pbmp=pcpal[c+color];
	#endif
					}
					pbmp++;

					c=cp>>4;

					if ( ((xp+x+1)<0) || ((xp+x+1)>=384) || ((yp+y)<0) ) ;
					else if (c!=15)
					{
	#ifdef DEBUG
						*pbmp=(c*16)|((c*16)<<8)|((c*16)<<16);
	#else
						*pbmp=pcpal[c+color];
	#endif
					}
					pbmp++;

					pchar++;
				}
				pchar+=4;
				pbmp+=(384-8);
			}
		}
		else if ((!flipx) && (flipy))
		{
	//		pchar+=7*8;
			pchar+=(7*8);
			for (y=0;y<8;y++)
			{
				if ((yp+y)>=224) return;
	//			for (x=0;x<8;x++)
				for (x=0;x<8;x+=2)
				{
					cp=*pchar;
					c=cp&0x0f;

					if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
					else if (c!=15)
					{
						*pbmp=pcpal[c+color];
					}
					pbmp++;

					c=cp>>4;

					if ( ((xp+x+1)<0) || ((xp+x+1)>=384) || ((yp+y)<0) ) ;
					else if (c!=15)
					{
						*pbmp=pcpal[c+color];
					}
					pbmp++;

					pchar++;
				}
				pbmp+=(384-8);
	//			pchar-=16;
				pchar-=12;
			}
		}
		else if ((flipx) && (!flipy))
		{
	//		pchar+=7;
			pchar+=3;
			for (y=0;y<8;y++)
			{
				if ((yp+y)>=224) return;
	//			for (x=0;x<8;x++)
				for (x=0;x<8;x+=2)
				{
					cp=*pchar;
					c=cp>>4;

					if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
					else if (c!=15)
					{
						*pbmp=pcpal[c+color];
					}
					pbmp++;

					c=cp&0x0f;

					if ( ((xp+x+1)<0) || ((xp+x+1)>=384) || ((yp+y)<0) ) ;
					else if (c!=15)
					{
						*pbmp=pcpal[c+color];
					}
					pbmp++;

					pchar--;
				}
				pbmp+=(384-8);
	//			pchar+=16;
				pchar+=12;
			}
		}
		else
		{
	//		pchar+=(7*8+7);
			pchar+=(7*8+3);
			for (y=0;y<8;y++)
			{
				if ((yp+y)>=224) return;
	//			for (x=0;x<8;x++)
				for (x=0;x<8;x+=2)
				{
					cp=*pchar;
					c=cp>>4;

					if ( ((xp+x)<0) || ((xp+x)>=384) || ((yp+y)<0) ) ;
					else if (c!=15)
					{
						*pbmp=pcpal[c+color];
					}
					pbmp++;

					c=cp&0x0f;

					if ( ((xp+x+1)<0) || ((xp+x+1)>=384) || ((yp+y)<0) ) ;
					else if (c!=15)
					{
						*pbmp=pcpal[c+color];
					}
					pbmp++;

					pchar--;
				}
				pbmp+=(384-8);
				pchar-=4;
			}
		}
	}
}

//

void tCharset::cps2printSprite(unsigned long int* vb,
							  int charnum,int color,
							  int xp,int yp,
							  int flipx,int flipy)
{
	if ( (xp<-15) || (xp>=384) || (yp<-15) || (yp>=224) ) return;

	unsigned long int* pbmp=&vb[yp*384+xp];
	unsigned char* pchar=&gfxrom[charnum*16*8+sproffs];
	unsigned long int* ppal=(unsigned long int*)palette;
	int x,y;
	int c,cp;

	color*=16;

	// have we to do clipping?
	if ( (xp>=0) && (xp<(384-16)) && (yp>=0) && (yp<(224-16)) )
	{
		if ((!flipx) && (!flipy))
		{
			for (y=0;y<16;y++)
			{
//				for (x=0;x<16;x++)
				for (x=0;x<16;x+=2)
				{
					cp=(*pchar);
					c=cp&0x0f;

					if (c!=15)
					{
#ifdef DEBUG
					*pbmp=(c*16)|((c*16)<<8)|((c*16)<<16);
#else
						*pbmp=ppal[c+color];
#endif
					}
					pbmp++;

					c=cp>>4;

					if (c!=15)
					{
#ifdef DEBUG
					*pbmp=(c*16)|((c*16)<<8)|((c*16)<<16);
#else
						*pbmp=ppal[c+color];
#endif
					}
					pbmp++;

					pchar++;
				}
				pbmp+=(384-16);
			}
		}
		else if ((!flipx) && (flipy))
		{
//			pchar+=(15*16);
			pchar+=(15*8);
			for (y=0;y<16;y++)
			{
//				for (x=0;x<16;x++)
				for (x=0;x<16;x+=2)
				{
					cp=(*pchar);
					c=cp&0x0f;

					if (c!=15)
					{
						*pbmp=ppal[c+color];
					}
					pbmp++;

					c=cp>>4;

					if (c!=15)
					{
						*pbmp=ppal[c+color];
					}
					pbmp++;

					pchar++;
				}
				pbmp+=(384-16);
				pchar-=16;
			}
		}
		else if ((flipx) && (!flipy))
		{
			pchar+=7;
			for (y=0;y<16;y++)
			{
//				for (x=0;x<16;x++)
				for (x=0;x<16;x+=2)
				{
					cp=(*pchar);
					c=cp>>4;

					if (c!=15)
					{
						*pbmp=ppal[c+color];
					}
					pbmp++;

					c=cp&0x0f;

					if (c!=15)
					{
						*pbmp=ppal[c+color];
					}
					pbmp++;
					
					pchar--;
				}
				pbmp+=(384-16);
				pchar+=16;
			}
		}
		else
		{
//			pchar+=(15*16)+15;
			pchar+=(15*8)+7;
			for (y=0;y<16;y++)
			{
//				for (x=0;x<16;x++)
				for (x=0;x<16;x+=2)
				{
					cp=(*pchar);
					c=cp>>4;

					if (c!=15)
					{
						*pbmp=ppal[c+color];
					}
					pbmp++;

					c=cp&0x0f;

					if (c!=15)
					{
						*pbmp=ppal[c+color];
					}
					pbmp++;

					pchar--;
				}
				pbmp+=(384-16);
			}
		}
	}
	else
	// clip
	{
		if ((!flipx) && (!flipy))
		{
			for (y=0;y<16;y++)
			{
				if ((yp+y)>=224) return;
//				for (x=0;x<16;x++)
				for (x=0;x<16;x+=2)
				{
					cp=(*pchar);
					c=cp&0x0f;

					if ( ((xp+x)>=0) && ((xp+x)<384) && ((yp+y)>=0) && (c!=15) )
					{
#ifdef DEBUG
					*pbmp=(c*16)|((c*16)<<8)|((c*16)<<16);
#else
						*pbmp=ppal[c+color];
#endif
					}
					pbmp++;

					c=cp>>4;

					if ( ((xp+x)>=-1) && ((xp+x)<383) && ((yp+y)>=0) && (c!=15) )
					{
#ifdef DEBUG
					*pbmp=(c*16)|((c*16)<<8)|((c*16)<<16);
#else
						*pbmp=ppal[c+color];
#endif
					}
					pbmp++;


					pchar++;
				}
				pbmp+=(384-16);
			}
		}
		else if ((!flipx) && (flipy))
		{
//			pchar+=15*16;
			pchar+=15*8;
			for (y=0;y<16;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<16;x+=2)
				{
					cp=(*pchar);
					c=cp&0x0f;

					if ( ((xp+x)>=0) && ((xp+x)<384) && ((yp+y)>=0) && (c!=15) )
					{
						*pbmp=ppal[c+color];
					}
					pbmp++;

					c=cp>>4;

					if ( ((xp+x)>=-1) && ((xp+x)<383) && ((yp+y)>=0) && (c!=15) )
					{
						*pbmp=ppal[c+color];
					}
					pbmp++;
					
					pchar++;
				}
				pbmp+=(384-16);
				pchar-=16;
			}
		}
		else if ((flipx) && (!flipy))
		{
			pchar+=7;
			for (y=0;y<16;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<16;x+=2)
				{
					cp=(*pchar);
					c=cp>>4;

					if ( ((xp+x)>=0) && ((xp+x)<384) && ((yp+y)>=0) && (c!=15) )
					{
						*pbmp=ppal[c+color];
					}
					pbmp++;

					c=cp&0x0f;

					if ( ((xp+x)>=-1) && ((xp+x)<383) && ((yp+y)>=0) && (c!=15) )
					{
						*pbmp=ppal[c+color];
					}
					pbmp++;
					
					pchar--;
				}
				pbmp+=(384-16);
				pchar+=16;
			}
		}
		else
		{
			pchar+=(15*8+7);
			for (y=0;y<16;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<16;x+=2)
				{
					cp=(*pchar);
					c=cp>>4;

					if ( ((xp+x)>=0) && ((xp+x)<384) && ((yp+y)>=0) && (c!=15) )
					{
						*pbmp=ppal[c+color];
					}
					pbmp++;

					c=cp&0x0f;

					if ( ((xp+x)>=-1) && ((xp+x)<383) && ((yp+y)>=0) && (c!=15) )
					{
						*pbmp=ppal[c+color];
					}
					pbmp++;

					pchar--;
				}
				pbmp+=(384-16);
			}
		}
	}
}

//

void tCharset::cps2printTile16(unsigned long int* vb,
							  int charnum,int color,
							  int xp,int yp,
							  int flipx,int flipy)
{
	if ( (xp<-14) || (xp>=384) || (yp<-14) || (yp>=224) ) return;

	unsigned long int* pbmp=&vb[yp*384+xp];
	unsigned char* pchar=&gfxrom[charnum*16*8+tile16offs];
	int x,y;
	int c,cp;

	unsigned long int* ptpal=(unsigned long int*)&palette[0x400*4];
	color*=16;

	// have we to do clipping?
	if ( (xp>=0) && (xp<(384-16)) && (yp>=0) && (yp<(224-16)) )
	{
		if ((!flipx) && (!flipy))
		{
			for (y=0;y<16;y++)
			{
				for (x=0;x<16;x+=2)
				{
					cp=*pchar;
					c=cp&0x0f;

					if (c!=15)
					{
						*pbmp=ptpal[c+color];
#ifdef DEBUG
					*pbmp=(c*16)|((c*16)<<8)|((c*16)<<16);
#endif
					}
					pbmp++;

					c=cp>>4;

					if (c!=15)
					{
						*pbmp=ptpal[c+color];
#ifdef DEBUG
					*pbmp=(c*16)|((c*16)<<8)|((c*16)<<16);
#endif
					}
					pbmp++;

					pchar++;
				}
				pbmp+=(384-16);
			}
		}
		else if ((!flipx) && (flipy))
		{
			pchar+=(15*8);
			for (y=0;y<16;y++)
			{
				for (x=0;x<16;x+=2)
				{
					cp=*pchar;
					c=cp&0x0f;

					if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;

					c=cp>>4;

					if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;

					pchar++;
				}
				pchar-=16;
				pbmp+=(384-16);
			}
		}
		else if ((flipx) && (!flipy))
		{
			pchar+=7;
			for (y=0;y<16;y++)
			{
				for (x=0;x<16;x+=2)
				{
					cp=*pchar;
					c=cp>>4;

					if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					
					c=cp&0x0f;

					if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;

					pchar--;
				}
				pbmp+=(384-16);
				pchar+=16;
			}
		}
		else
		{
			pchar+=(15*8)+7;
			for (y=0;y<16;y++)
			{
				for (x=0;x<16;x+=2)
				{
					cp=*pchar;
					c=cp>>4;

					if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;

					c=cp&0x0f;

					if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;

					pchar--;
				}
				pbmp+=(384-16);
			}
		}
	}
	else
	// clip blit
	{
		if ((!flipx) && (!flipy))
		{
			for (y=0;y<16;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<16;x+=2)
				{
					cp=*pchar;
					c=cp&0x0f;

					if ( ((xp+x)>=0) && ((xp+x)<384) && ((yp+y)>=0) && (c!=15) )
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					
					c=cp>>4;

					if ( ((xp+x)>=-1) && ((xp+x)<383) && ((yp+y)>=0) && (c!=15) )
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;

					pchar++;
				}
				pbmp+=(384-16);
			}
		}
		else if ((!flipx) && (flipy))
		{
			pchar+=(15*8);
			for (y=0;y<16;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<16;x+=2)
				{
					cp=*pchar;
					c=cp&0x0f;

					if ( ((xp+x)>=0) && ((xp+x)<384) && ((yp+y)>=0) && (c!=15) )
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					
					c=cp>>4;

					if ( ((xp+x)>=-1) && ((xp+x)<383) && ((yp+y)>=0) && (c!=15) )
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;

					pchar++;
				}
				pchar-=16;
				pbmp+=(384-16);
			}
		}
		else if ((flipx) && (!flipy))
		{
			pchar+=7;
			for (y=0;y<16;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<16;x+=2)
				{
					cp=*pchar;
					c=cp>>4;

					if ( ((xp+x)>=0) && ((xp+x)<384) && ((yp+y)>=0) && (c!=15) )
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					
					c=cp&0x0f;

					if ( ((xp+x)>=-1) && ((xp+x)<383) && ((yp+y)>=0) && (c!=15) )
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;

					pchar--;
				}
				pbmp+=(384-16);
				pchar+=16;
			}
		}
		else
		{
			pchar+=(15*8)+7;
			for (y=0;y<16;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<16;x+=2)
				{
					cp=*pchar;
					c=cp>>4;

					if ( ((xp+x)>=0) && ((xp+x)<384) && ((yp+y)>=0) && (c!=15) )
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					
					c=cp&0x0f;

					if ( ((xp+x)>=-1) && ((xp+x)<383) && ((yp+y)>=0) && (c!=15) )
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;

					pchar--;
				}
				pbmp+=(384-16);
			}
		}
	}
}

//

void tCharset::cps2printTile32(unsigned long int* vb,
							  int charnum,int color,
							  int xp,int yp,
							  int flipx,int flipy)
{
	if ( (xp<-32) || (xp>=384) || (yp<-32) || (yp>=224) ) return;

	unsigned long int* pbmp=&vb[yp*384+xp];
	unsigned char* pchar=&gfxrom[charnum*64*8+tile32offs];
	unsigned long int* ptpal=(unsigned long int*)&palette[0x600*4];
	int x,y;
	int c,cp;
	color*=16;

	// have we to do clipping?
	if ( (xp>=0) && (xp<(384-32)) && (yp>=0) && (yp<(224-32)) )
	{
		//noclip
		if ((!flipx) && (!flipy))
		{
			for (y=0;y<32;y++)
			{
				for (x=0;x<32;x+=2)
				{
//					c=(*pchar)&0x0f;
					cp=(*pchar);
					c=cp&0x0f;

					if (c!=15)
					{
						*pbmp=ptpal[c+color];
#ifdef DEBUG
						*pbmp=(c*16)|((c*16)<<8)|((c*16)<<16);
#endif
					}
					pbmp++;
					
//					c=(*pchar)>>4;
					c=cp>>4;

					if (c!=15)
					{
						*pbmp=ptpal[c+color];
#ifdef DEBUG
						*pbmp=(c*16)|((c*16)<<8)|((c*16)<<16);
#endif
					}
					pbmp++;

					pchar++;
				}
				pbmp+=(384-32);
			}
		}
		else if ((!flipx) && (flipy))
		{
			pchar+=(31*16);
			for (y=0;y<32;y++)
			{
				for (x=0;x<32;x+=2)
				{
					cp=(*pchar);
					c=cp&0x0f;

					if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					
					c=cp>>4;

					if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;

					pchar++;
				}
				pbmp+=(384-32);
				pchar-=32;
			}
		}
		else if ((flipx) && (!flipy))
		{
			pchar+=15;
			for (y=0;y<32;y++)
			{
				for (x=0;x<32;x+=2)
				{
					cp=(*pchar);
					c=cp>>4;

					if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					
					c=cp&0x0f;

					if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;

					pchar--;
				}
				pbmp+=(384-32);
				pchar+=32;
			}
		}
		else
		{
			pchar+=(31*16)+15;
			for (y=0;y<32;y++)
			{
				for (x=0;x<32;x+=2)
				{
//					c=(*pchar)>>4;
					cp=(*pchar);
					c=cp>>4;

					if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					
					c=cp&0x0f;

					if (c!=15)
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;

					pchar--;
				}
				pbmp+=(384-32);
			}
		}
	// clip
	}
	else
	{
		if ((!flipx) && (!flipy))
		{
			for (y=0;y<32;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<32;x+=2)
				{
					cp=(*pchar);
					c=cp&0x0f;

					if ( ((xp+x)>=0) && ((xp+x)<384) && ((yp+y)>=0) && (c!=15) )
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					
					c=cp>>4;

					if ( ((xp+x)>=-1) && ((xp+x)<383) && ((yp+y)>=0) && (c!=15) )
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;

					pchar++;
				}
				pbmp+=(384-32);
			}
		}
		else if ((!flipx) && (flipy))
		{
			pchar+=(31*16);
			for (y=0;y<32;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<32;x+=2)
				{
					cp=(*pchar);
					c=cp&0x0f;

					if ( ((xp+x)>=0) && ((xp+x)<384) && ((yp+y)>=0) && (c!=15) )
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					
					c=cp>>4;

					if ( ((xp+x)>=-1) && ((xp+x)<383) && ((yp+y)>=0) && (c!=15) )
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;

					pchar++;
				}
				pbmp+=(384-32);
				pchar-=32;
			}
		}
		else if ((flipx) && (!flipy))
		{
			pchar+=15;
			for (y=0;y<32;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<32;x+=2)
				{
//					c=(*pchar)>>4;
					cp=(*pchar);
					c=cp>>4;

					if ( ((xp+x)>=0) && ((xp+x)<384) && ((yp+y)>=0) && (c!=15) )
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					
					c=cp&0x0f;

					if ( ((xp+x)>=-1) && ((xp+x)<383) && ((yp+y)>=0) && (c!=15) )
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;

					pchar--;
				}
				pbmp+=(384-32);
				pchar+=32;
			}
		}
		else
		{
			pchar+=(31*16)+15;
			for (y=0;y<32;y++)
			{
				if ((yp+y)>=224) return;
				for (x=0;x<32;x+=2)
				{
//					c=(*pchar)>>4;
					cp=(*pchar);
					c=cp>>4;

					if ( ((xp+x)>=0) && ((xp+x)<384) && ((yp+y)>=0) && (c!=15) )
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;
					
					c=cp&0x0f;

					if ( ((xp+x)>=-1) && ((xp+x)<383) && ((yp+y)>=0) && (c!=15) )
					{
						*pbmp=ptpal[c+color];
					}
					pbmp++;

					pchar--;
				}
				pbmp+=(384-32);
			}
		}
	}
}

//

void tCharset::cps2printTile16rs(unsigned long int* vb,
							  int charnum,int color,
							  int xp,int yp,
							  int flipx,int flipy,int ofz)
{
	if ( (yp<-14) || (yp>223) ) return;

	unsigned char* pchar=&gfxrom[charnum*16*8+tile16offs];
	unsigned long int* vbp;
	int x,y;
	int c;
	int rso;
	int cp;
	unsigned long int* ptpal=(unsigned long int*)&palette[0x400*4];

	color*=16;

	if ((!flipx) && (!flipy))
	{
		for (y=0;y<16;y++)
		{
			int ypp=yp+y;
			if (ypp>=224) return;
			rso=rowscroll[(ypp+ofz)&0x3ff];
			int xpp=xp+rso;

			if (ypp>=0)
			{
				vbp=&vb[xpp+ypp*384];
				if ( (xpp>=0) && (xpp<(384-16)) )
				{
					for (x=0;x<16;x+=2)
					{
						cp=(*pchar);
						c=cp&0x0f;

						if (c!=15)
						{
							*vbp=ptpal[c+color];
						}
						vbp++;
						
						c=cp>>4;

						if (c!=15)
						{
							*vbp=ptpal[c+color];
						}
						vbp++;

						pchar++;
					}
				}
				else
				{
					for (x=0;x<16;x+=2)
					{
						cp=(*pchar);
						c=cp&0x0f;

						if ( (xpp>=0) && (xpp<384) && (c!=15) )
						{
							*vbp=ptpal[c+color];
						}
						xpp++;
						vbp++;
						
						c=cp>>4;

						if ( (xpp>=0) && (xpp<384) && (c!=15) )
						{
							*vbp=ptpal[c+color];
						}
						xpp++;
						vbp++;

						pchar++;
					}
				}
			}
			else
			{
				pchar+=8;
			}
		}
	}
	else if ((!flipx) && (flipy))
	{
		pchar+=(15*8);
		for (y=0;y<16;y++)
		{
			int ypp=yp+y;
			if (ypp>=224) return;
			rso=rowscroll[(ypp+ofz)&0x3ff];
			int xpp=xp+rso;

			if (ypp>=0)
			{
				vbp=&vb[xpp+ypp*384];
				if ( (xpp>=0) && (xpp<(384-16)) )
				{
					for (x=0;x<16;x+=2)
					{
						cp=(*pchar);
						c=cp&0x0f;

						if (c!=15)
						{
							*vbp=ptpal[c+color];
						}
						vbp++;
						
						c=cp>>4;

						if (c!=15)
						{
							*vbp=ptpal[c+color];
						}
						vbp++;

						pchar++;
					}
					pchar-=16;
				}
				else
				{
					for (x=0;x<16;x+=2)
					{
						cp=(*pchar);
						c=cp&0x0f;

						if ( (xpp>=0) && (xpp<384) && (c!=15) )
						{
							*vbp=ptpal[c+color];
						}
						xpp++;
						vbp++;
						
						c=cp>>4;

						if ( (xpp>=0) && (xpp<384) && (c!=15) )
						{
							*vbp=ptpal[c+color];
						}
						xpp++;
						vbp++;

						pchar++;
					}
					pchar-=16;
				}
			}
			else
			{
				pchar-=8;
			}
		}
	}
	else if ((flipx) && (!flipy))
	{
		pchar+=7;
		for (y=0;y<16;y++)
		{
			int ypp=yp+y;
			if (ypp>=224) return;
			rso=rowscroll[(ypp+ofz)&0x3ff];
			int xpp=xp+rso;

			if (ypp>=0)
			{
				vbp=&vb[xpp+ypp*384];
				if ( (xpp>=0) && (xpp<(384-16)) )
				{
					for (x=0;x<16;x+=2)
					{
						cp=(*pchar);
						c=cp>>4;

						if (c!=15)
						{
							*vbp=ptpal[c+color];
						}
						vbp++;
						
						c=cp&0x0f;

						if (c!=15)
						{
							*vbp=ptpal[c+color];
						}
						vbp++;

						pchar--;
					}
				}
				else
				{
					for (x=0;x<16;x+=2)
					{
						cp=(*pchar);
						c=cp>>4;

						if ( (xpp>=0) && (xpp<384) && (c!=15) )
						{
							*vbp=ptpal[c+color];
						}
						xpp++;
						vbp++;
						
						c=cp&0x0f;

						if ( (xpp>=0) && (xpp<384) && (c!=15) )
						{
							*vbp=ptpal[c+color];
						}
						xpp++;
						vbp++;

						pchar--;
					}
				}
			}
			else
			{
				pchar-=8;
			}
			pchar+=16;
		}
	}
	else
	{
		pchar+=(15*8)+7;
		for (y=0;y<16;y++)
		{
			int ypp=yp+y;
			if (ypp>=224) return;
			rso=rowscroll[(ypp+ofz)&0x3ff];
			int xpp=xp+rso;

			if (ypp>=0)
			{
				vbp=&vb[xpp+ypp*384];
				if ( (xpp>=0) && (xpp<(384-16)) )
				{
					for (x=0;x<16;x+=2)
					{
						cp=(*pchar);
						c=cp>>4;

						if (c!=15)
						{
							*vbp=ptpal[c+color];
						}
						vbp++;
						
						c=cp&0x0f;

						if (c!=15)
						{
							*vbp=ptpal[c+color];
						}
						vbp++;

						pchar--;
					}
				}
				else
				{
					for (x=0;x<16;x+=2)
					{
						cp=(*pchar);
						c=cp>>4;

						if ( (xpp>=0) && (xpp<384) && (c!=15) )
						{
							*vbp=ptpal[c+color];
						}
						xpp++;
						vbp++;
						
						c=cp&0x0f;

						if ( (xpp>=0) && (xpp<384) && (c!=15) )
						{
							*vbp=ptpal[c+color];
						}
						xpp++;
						vbp++;

						pchar--;
					}
				}
			}
			else
			{
				pchar-=8;
			}
		}
	}
}
