
/*

  a.c.e.
  Another Capcom (Play System) Emulator

  project started 12/5/2001

*/
 
//#define DEBUG


#include <stdio.h>
#include <stdlib.h>
#include <windows.h>

#include "component.h"
#include "background.h"
#include "sysbar.h"
#include "menu.h"
#include "video32.h"
#include "sound.h"
#include "romspace.h"
#include "charset.h"
#include "cpucore.h"
#include "timing.h"
#include "gui.h"

// the gui

tGui* gui;


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	HWND wh;
	MSG msg;
 
	int secs;
	char strsecs[256];
	int exit=0;
 
	// init gui
  
	gui=new tGui(0,SCRX,SCRY,"ace.ini");
	if (gui->initfailed) 
	{
		MessageBox(NULL,"can't start ace","error",MB_OK);
		return 1;
	}
	gui->printMessage(". Ace started");

	// init display

	if (initDisplay(hInstance,nCmdShow,&wh,&gui->videoBuf,gui->guiinfo.videomode)) 
	{
		MessageBox(NULL,"can't do video init","error",MB_OK);
		return 1;
	}
	gui->printMessage(". Init display ok");
	gui->printMessage(". Press right mouse button");
 
	// init timer

	init_timer();

	// init sound

	if (gui->guiinfo.soundon)
	{
		if (initSound(wh,gui->guiinfo.soundfreq))
		{
			MessageBox(NULL,"can't do sound init","error",MB_OK);
			freeDisplay(&gui->videoBuf);
			shutdownSound();
			delete gui;
			return 1;
		}
	}

	// main loop

	while (!exit)
	{
		if (gui->emulating!=0)
		{
#ifndef DEBUG
			if (gui->romsettype==ROMSET_CPS1)
			{
				if (!gui->ginput.paused) cps1cpucore_emulate(gui->guiinfo.soundon);
				if ((gui->frames%gui->guiinfo.frameskip)==0) cps1cpucore_drawScreen(gui->videoBuf,gui->chset);
			}
			else if (gui->romsettype==ROMSET_CPS2)
			{
				if (!gui->ginput.paused) cps2cpucore_emulate(gui->guiinfo.soundon);
				if ((gui->frames%gui->guiinfo.frameskip)==0) cps2cpucore_drawScreen(gui->videoBuf,gui->chset);
			}

			gui->update();
			gui->redraw(gui->videoBuf);
			if ((gui->frames%gui->guiinfo.frameskip)==0) videoput32(gui->videoBuf);

			gui->frames++;
			if (gui->guiinfo.trimspeed!=0) trim_speed();
#else 
			extern int goshift;
			int r,c;

			int chr=0;
			int nx=16;
			int ny=16;
			int dim=8;

			chr=0;
			for (r=0;r<ny;r++)
			{	
				for (c=0;c<nx;c++)
				{
					gui->chset->cps2printChar(
									gui->videoBuf,
									chr+goshift*nx,0,
									c*dim,r*dim,0,0);
					chr++;
				}
			}
			gui->update();
			gui->redraw(gui->videoBuf);
			videoput32(gui->videoBuf);
#endif
		}
		else
		{
			gui->update();
			gui->redraw(gui->videoBuf);
			videoput32(gui->videoBuf);
			if (gui->guiinfo.trimspeed!=0) trim_speed();
		}

		if (PeekMessage(&msg, NULL,0,0,PM_NOREMOVE))
		{
			GetMessage (&msg, NULL, 0, 0);
			TranslateMessage (&msg);
			DispatchMessage (&msg);
		}

		if (gui->exitprogram())
		{
			while (PeekMessage(&msg, NULL,0,0,PM_NOREMOVE))
			{
				GetMessage (&msg, NULL, 0, 0);
				TranslateMessage (&msg);
				DispatchMessage (&msg);
			}
			exit=1;
		}
	}

	if (gui->guiinfo.showfps)
	{
		// show fps
		secs=secs_elapsed();
		sprintf(strsecs,"[fps: %f]",(float)((float)gui->frames/(float)secs));
		MessageBox(0,strsecs,"info",MB_OK);
	}

	// dispose soundz

	if (gui->guiinfo.soundon)
	{
		shutdownSound();
	}

	// dispose video

	freeDisplay(&gui->videoBuf);

	// dispose cpu

	if (gui->romsettype==ROMSET_CPS1)
	{
		cps1cpucore_free();
	}
	else if (gui->romsettype==ROMSET_CPS2)
	{
		cps2cpucore_free();
	}
 
	// dispose gui

	delete gui;

	// exit

	return (msg.wParam);
}
