
/*

  gui implementation

*/

#include "gui.h"
#include "cps1romspace.h"
#include "cps2romspace.h"
#include "s16romspace.h"
#include "neogeoromspace.h"
#include "cps1cpucore.h"
#include "cps2cpucore.h"
#include "s16cpucore.h"
#include "s18cpucore.h"
#include "neogeocpucore.h"

extern tRomSet romsetList[];


tGui::tGui(int s,int gxs,int gys,char* inifile)
{
	int i,err;

	machine=0;
	romset=0;
	z80init=0;
	initfailed=0;
	frames=0;
	orientation=VIDEO_NORMAL;
	changingmode=0;
	chset=0;
	exitp=0;
	state=s;
	emulating=0;
	guiXsize=gxs;
	guiYsize=gys;

	// reset controls

	ginput.mouseXpos=0;
	ginput.mouseYpos=0;
	ginput.mousebutton1=0;
	ginput.mousebutton2=0;

	for (i=0;i<6;i++)
	{
		ginput.p1button[i]=0;
		ginput.p2button[i]=0;
	}

	ginput.coin1=ginput.coin2=0;
	ginput.start1=ginput.start2=0;
	ginput.p1up=ginput.p2up=0;
	ginput.p1down=ginput.p2down=0;
	ginput.p1left=ginput.p2left=0;
	ginput.p1right=ginput.p2right=0;

	for (i=0;i<3;i++)
	{
		ginput.layeron[i]=0;
	}

	ginput.paused=0;
	ginput.quicksavestate=0;
	ginput.quickloadstate=0;

	// inifile

	err=readIniFile(inifile);
	if (err!=0)
	{
		initfailed=1;
		return ;
	}

	// store font

	font8x8=new tFont(64,8,8,0xcecece);

	// add components

	compList.resize(5);
 
	compList[0]=new tBackground(0x151535,guiXsize,guiYsize);
	compList[1]=new tSysbar(0x202050,0x9292c2,12,guiXsize,guiYsize);
	compList[2]=new tMenu(0x9292c2,guiXsize,guiYsize);
	compList[3]=new tSelwindow(0x202050,0x9292c2,guiXsize,guiYsize);
	compList[4]=new tGawin(0x202050,0x9292c2,guiXsize,guiYsize);
 
	// create first menu items

	tMenuItem mi1("File",	1,0x9292c2,-1);
	tMenuItem mi3("Cpu",	2,0x9292c2,-1);
	tMenuItem mi4("About",	-1,0x9292c2,1);
	tMenuItem mi5("Exit",	-1,0x9292c2,2);

	tMenuItem mi6("Open",-1,0x9292c2,3);
	tMenuItem mi7("Free",-1,0x9292c2,4);
	tMenuItem mi6bis("CmPro dat",-1,0x9292c2,5);
	tMenuItem mi7bis("RomCenter dat",-1,0x9292c2,6);

	tMenuItem mi11("Reset",-1,0x9292c2,8);
	tMenuItem mi12("Pause",-1,0x9292c2,9);
	tMenuItem mi13("Save state",-1,0x9292c2,10);
	tMenuItem mi14("Load state",-1,0x9292c2,11);

	tSubMenu sm1(2,210,0x202050,guiXsize,guiYsize,0);
	tSubMenu sm2(59,174,0x202050,guiXsize,guiYsize,0);
	tSubMenu sm4(59,210,0x202050,guiXsize,guiYsize,0);

	sm1.addItem(mi1);
	sm1.addItem(mi3);
	sm1.addItem(mi4);
	sm1.addItem(mi5);

	sm2.addItem(mi6);
	sm2.addItem(mi7);
	sm2.addItem(mi6bis);
	sm2.addItem(mi7bis);

	sm4.addItem(mi11);
	sm4.addItem(mi12);
	sm4.addItem(mi13);
	sm4.addItem(mi14);

	tMenu* mp=(tMenu*)compList[2];
	mp->addItem(sm1);
	mp->addItem(sm2);
	mp->addItem(sm4);

	// init rotobuf
	rotobuf=new unsigned long int [SCRX*SCRY];

	// add messages
	msgs=new tMessage();
}

//
// open a rom and start emulation
// (returns 1 if any problem occurs)
//

int tGui::openRom(char* romdesc)
{
	// find romsettype

	int found;
	int pos;

	found=0;
	pos=0;

	while ( (found==0) && (strlen(romsetList[pos].name)!=0) )
	{
		if (!strcmp(romdesc,romsetList[pos].name))
		{
			found=1;
		}
		else
		{
			pos++;
		}
	}

	if (found==0)
	{
		printMessage(". Romset not supported");
		return 1;
	}

	romsettype=romsetList[pos].romsettype;

	// init romset

	if (romset!=0) delete romset;

	if (romsettype==ROMSET_CPS1)
	{
		romset=new cps1romspace();
	}
	else if (romsettype==ROMSET_CPS2)
	{
		romset=new cps2romspace();
	}
	else if ( (romsettype==ROMSET_S16) || (romsettype==ROMSET_S18) )
	{
		romset=new s16romspace();
	}
	else if (romsettype==ROMSET_NEOGEO)
	{
		romset=new neogeoromspace();
	}

	if (romset->readRoms(guiinfo.rompath,romdesc,&chset))
	{
		printMessage(". Unable to read roms");
		return 1;
	}

	// decode charset
	romset->decodeCharset(chset);

	// choose which cpucore to use

	if (machine!=0) delete machine;

	if (romsettype==ROMSET_CPS1)
	{
		machine=new cps1cpucore(romset);
	}
	else if (romsettype==ROMSET_CPS2)
	{
		machine=new cps2cpucore(romset);
	}
	else if (romsettype==ROMSET_S16)
	{
		machine=new s16cpucore(romset);
	}
	else if (romsettype==ROMSET_S18)
	{
		machine=new s18cpucore(romset);
	}
	else if (romsettype==ROMSET_NEOGEO)
	{
		machine=new neogeocpucore(romset);
	}

	// which is the rom orientation ?
	if (romset->orientation!=orientation)
	{
		// re-init display
		changingmode=1;
		orientation=romset->orientation;
		reinitDisplay(romset->orientation);
		changingmode=0;

		// reset gui dimensions
		int tmp=guiXsize;
		guiXsize=guiYsize;
		guiYsize=tmp;

		// reset also all component's sizes
		for (int i=0;i<compList.size();i++)
		{
			compList[i]->resetguisize(guiXsize,guiYsize);
		}

		// reset menu positions
	}

	emulating=1;
	frames=0;
	start_count();	//start counting seconds of emulation
	printMessage(". Emulation started");

	return 0;
}

void tGui::redraw(unsigned long int* vbuf)
{
	// which is the gui's orientation?
	if (orientation==VIDEO_NORMAL)
	{
		if (state==1)
		{
			for (int i=0;i<compList.size();i++)
			{
				if (compList[i]->isActive()!=0)
				{
					compList[i]->draw(vbuf,font8x8,&ginput);
				}
			}
		}
		else 
		{
			// if gui is paused, draw "paused" somewhere
			if (ginput.paused)
			{
				font8x8->drawString("Paused",0xcecece,
									guiXsize/2-30,guiYsize/2,guiXsize,guiYsize,
									vbuf);
			}
		}

		msgs->draw(vbuf,font8x8,guiXsize,guiYsize);
	}
	else if (orientation==VIDEO_ROT90)
	{
		// draw the gui onto rotobuf
		if (state==1)
		{
			// copy the vbuf into rotobuf

			int i,l;
			unsigned long int* pSrc=vbuf;
			unsigned long int* pRb;
			unsigned long int* pVb;

			for (i=0;i<SCRY;i++)
			{
				pRb=&rotobuf[i+(SCRX-1)*SCRY];
				for (l=0;l<SCRX;l++)
				{
					*pRb=*pSrc;
					pSrc++;
					pRb-=SCRY;
				}
			}

			for (i=0;i<compList.size();i++)
			{
				if (compList[i]->isActive()!=0)
				{
					compList[i]->draw(rotobuf,font8x8,&ginput);
				}
			}

			// reput rotobuf onto vbuf
			pSrc=rotobuf;

			for (i=0;i<SCRX;i++)
			{
				pVb=&vbuf[(SCRX-1-i)];
				for (l=0;l<SCRY;l++)
				{
					*pVb=*pSrc;
					pSrc++;
					pVb+=SCRX;
				}
			}
		}
		else 
		{
			// if gui is paused, draw "paused" somewhere

			if (ginput.paused)
			{
				font8x8->drawString("Paused",0xcecece,
									guiXsize/2-30,guiYsize/2,guiXsize,guiYsize,
									vbuf);
			}
		}

		msgs->draw(vbuf,font8x8,guiXsize,guiYsize);
	}
}

void tGui::update()
{
	if (state==1)
	{
		// update components
		// basing on user input and mouse movements

		for (int i=0;i<compList.size();i++)
		{
			if (compList[i]->isActive())
			{
				compList[i]->update(&ginput);
			}

			if (i==1)		// sysbar...
			{
				tSysbar* sbp=(tSysbar*)compList[1];

				if (sbp->menuclicked==1)
				{
					compList[2]->active=!compList[2]->active;
				}
			}
			else if (i==2)		// sysmenu
			{
				tMenu* mp=(tMenu*)compList[2];
				int act;

				act=mp->getAction();

				// exit
				if (act==2)
				{
					exitp=1;
					mp->deactivate();
				}
				else if (act==3)
				{
					//openRom();
					// activate selwindow
					compList[3]->active=1;
					compList[4]->active=0;
					mp->deactivate();
				}
				// about
				else if (act==1)
				{
					compList[3]->active=0;
					compList[4]->active=1;
					mp->deactivate();
				}
				// pause
				else if (act==9)
				{
					ginput.paused=1;
				}
				// romcenter datz
				else if (act==5)
				{
					// cmpro dat
					write_datfile(0);
					mp->deactivate();
					printMessage(". CmPro datfile saved");
				}
				else if (act==6)
				{
					// romcenter dat
					write_datfile(1);
					mp->deactivate();
					printMessage(". RomCenter datfile saved");
				}

				mp->resetAction();
			}
			else if (i==3)		// filesel
			{
				tSelwindow* wp=(tSelwindow*)compList[3];

				if (wp->clicked)
				{
					compList[3]->active=0;
					wp->clicked=0;
					openRom(romsetList[wp->selected].name);
				}
			}
		}
	}

	msgs->serveQueue();
}

int tGui::exitprogram()
{
	return exitp;
}

void tGui::printMessage(char* m)
{
	msgs->addMessage(m);
}

tGui::~tGui()
{
	for (int i=0;i<compList.size();i++)
	{
		delete(compList[i]);
	}

	delete(font8x8);

	if (chset!=0) 
	{
		delete chset;
	}

	if (romset!=0) delete romset;

	if (machine!=0) delete machine;

	delete [] rotobuf;

	delete msgs;
}

/*

  functions to deal with .ini files

*/

int tGui::readIniFile(char* inifile)
{
	int err;
	char pval[1024];
	int ipval;

	//
	memset(guiinfo.rompath,0,sizeof(guiinfo.rompath));

	err=getParameter(inifile,"[general]","rompath",pval);
	if (err!=0)
	{
		MessageBox(0,"Error: can't read rompath var. from .ini file","error",MB_OK);
		return 1;
	}
	else
	{
		int sl=strlen(pval);

		if (pval[sl-1]!='\\')
		{
			pval[sl]='\\';
			pval[sl+1]='\0';
		}

		strcpy(guiinfo.rompath,pval);
	}

	//

	err=getParameter(inifile,"[general]","showfps",pval);
	if (err!=0)
	{
		MessageBox(0,"Error: can't read showfps var. from .ini file","error",MB_OK);
		return 1;
	}
	else
	{
		ipval=atoi(pval);
		if (ipval) guiinfo.showfps=1;
		else guiinfo.showfps=0;
	}

	//

	err=getParameter(inifile,"[sound]","soundon",pval);
	if (err!=0)
	{
		MessageBox(0,"Error: can't read soundon var. from .ini file","error",MB_OK);
		return 1;
	}
	else
	{
		ipval=atoi(pval);
		if (ipval) guiinfo.soundon=1;
		else guiinfo.soundon=0;
	}

	//

	err=getParameter(inifile,"[sound]","soundfreq",pval);
	if (err!=0)
	{
		MessageBox(0,"Error: can't read soundfreq var. from .ini file","error",MB_OK);
		return 1;
	}
	else
	{
		ipval=atoi(pval);
		guiinfo.soundfreq=ipval;
	}

	//
	err=getParameter(inifile,"[sound]","interpolate",pval);
	if (err!=0)
	{
		MessageBox(0,"Error: can't read interpolate var. from .ini file","error",MB_OK);
		return 1;
	}
	else
	{
		ipval=atoi(pval);
		guiinfo.interpolate=ipval;
	}

	//
	err=getParameter(inifile,"[sound]","usedirectsound",pval);
	if (err!=0)
	{
		MessageBox(0,"Error: can't read usedirectsound var. from .ini file","error",MB_OK);
		return 1;
	}
	else
	{
		ipval=atoi(pval);
		guiinfo.usedsound=ipval;
	}


	//
	err=getParameter(inifile,"[video]","frameskip",pval);
	if (err!=0)
	{
		MessageBox(0,"Error: can't read frameskip var. from .ini file","error",MB_OK);
		return 1;
	}
	else
	{
		ipval=atoi(pval);
		guiinfo.frameskip=ipval;
		if ((ipval<=0) || (ipval>20)) guiinfo.frameskip=1;
	}

	//
	err=getParameter(inifile,"[video]","videomode",pval);
	if (err!=0)
	{
		MessageBox(0,"Error: can't read videomode var. from .ini file","error",MB_OK);
		return 1;
	}
	else
	{
		ipval=atoi(pval);
		guiinfo.videomode=ipval;
	}

	//

	err=getParameter(inifile,"[video]","trimspeed",pval);
	if (err!=0)
	{
		MessageBox(0,"Error: can't read trimspeed var. from .ini file","error",MB_OK);
		return 1;
	}
	else
	{
		ipval=atoi(pval);
		guiinfo.trimspeed=ipval;
	}

	err=getParameter(inifile,"[video]","usesysmem",pval);
	if (err!=0)
	{
		MessageBox(0,"Error: can't read usesysmem var. from .ini file","error",MB_OK);
		return 1;
	}
	else
	{
		ipval=atoi(pval);
		guiinfo.usesysmemory=ipval;
	}

	return 0;
}

int tGui::ltrim(char* str)
{
    int found=0;
    char* pstr=str;

    while (found==0)
    {
        if (*pstr==' ')
        {
            pstr++;
        }
        else found=1;
    }

    strcpy(str,pstr);

    return 0;
}

int tGui::rtrim(char* str)
{
    unsigned int sl;
    unsigned int pos;
    int found=0;

    sl=strlen(str);
    if (sl==0) return 0;

    pos=sl-1;

    while ( (found==0) && (pos>0) )
    {
        if ( (str[pos]==' ') || (str[pos]==0x0a) || (str[pos]==0x0d) )
        {
            str[pos]='\0';
            pos--;
        }
        else
        {
            found=1;
        }
    }

    return 0;
}

int tGui::getParameter(char* filename,char* sectionname,
                       char* paramname,char* paramvalue)
{
    FILE* inifile;
    int sectionFound=0;
    int paramFound=0;
    int goOut=0;
    
    if ( (filename==NULL) || (sectionname==NULL) || (paramname==NULL) || (paramvalue==NULL) ) return 1;

    inifile=fopen(filename,"rb");
    if (inifile==NULL) return 1;

    /* scan the file until you don't find a matching section or until it ends */
    while (!goOut)
    {
        char line[256];
        char* bufptr;
        
        bufptr=(char*)fgets((char*)line,256,inifile);
        if (bufptr==NULL)
        {
            goOut=1;
        }
        else
        {
            /* rtrim and ltrim the line found */
            ltrim(line);
            rtrim(line);
                
            /* check if it's a comment */
            if (line[0]==';') ;
            else
            {
                /* check if it's a section and matches */
                if (!strcmp(line,sectionname))
                {
                    sectionFound=1;
                    /* printf("trovata la sezione %s\n",line); */
                    goOut=1;
                }
            }
        }
    }

    if (sectionFound!=0)
    {
        goOut=0;
        
        while (!goOut)
        {
            /*
               scan the inifile until:
               1) it ends
               2) you find "[" (another section)
               3) you find the parameter
            */
        
            char line[256];
            char* bufptr;
        
            bufptr=(char*)fgets(line,256,inifile);
            if (bufptr==NULL)
            {
                goOut=1;
                /* printf("eof?\n"); */
            }
            else
            {
                /* rtrim and ltrim the line found */
                ltrim(line);
                rtrim(line);
                
                /* printf("line trimmed\n"); */
                
                /* check if it's a comment */
                if (line[0]==';') ;
                if (line[0]=='[')
                {
                    goOut=1;
                }
                else
                {
                    /* check if there's a parameter and if it matches */
                    char lvalue[256];
                    char rvalue[256];
                    char* eqptr;
                
                    eqptr=(char*)strstr(line,"=");
                    if (eqptr==NULL)
                    {
                        /* skip the line (it would be a syntax error really) */
                    }
                    else
                    {
                        strcpy(rvalue,eqptr+1);
                        *eqptr='\0';
                        strcpy(lvalue,line);
                        
                        ltrim(lvalue); rtrim(lvalue);
                        ltrim(rvalue); rtrim(rvalue);
                        
                        /* printf("rtrimmed\n"); */
                        /* printf("paramname:%s lvalue:%s\n",paramname,lvalue); */
                        
                        /* is it our parameter? */
                        if (!strcmp(paramname,lvalue))
                        {
                            /* allocate paramvalue */
                        
                            strcpy(paramvalue,rvalue);
                            goOut=1;
                            paramFound=1;
                        }
                    }
                }
            }
        
        }
    }

    fclose(inifile);

	if ( (sectionFound==0) || (paramFound==0) ) return 1;
    return 0;
}

//
// user input
//

void tGui::mouseMove(int x,int y)
{
	ginput.mouseXpos=x;
	ginput.mouseYpos=y;
}

void tGui::lButtonDown()
{
	ginput.mousebutton1=1;
}

void tGui::lButtonUp()
{
	ginput.mousebutton1=0;
}

void tGui::rButtonDown()
{
	ginput.mousebutton2=1;

	// the right button should activate/deactivate the gui

	state=!state;
}

void tGui::rButtonUp()
{
	ginput.mousebutton2=0;
}

