
/*

  Neo-Geo cpu core implementation
  todo:
	z80

*/

#include <time.h>

#include "neogeocpucore.h"
#include "neogeoromspace.h"


// forward declarations

#define WRITE_WORD(a,d)       (*(unsigned short *)(a) = (d))

unsigned char neogeo_m68kr8(int address);
unsigned short int neogeo_m68kr16(int address);
unsigned long int neogeo_m68kr32(int address);
void neogeo_m68kw8(int address, unsigned char value);
void neogeo_m68kw16(int address, unsigned short int value);
void neogeo_m68kw32(int address, unsigned long int value);

// variables

extern tGui* gui;
extern int goshift;

static unsigned char* rom68k;
static unsigned char* ngbios;
static unsigned char* ngram;
static unsigned char* bank4;
static unsigned char* cpubkp;
static unsigned char sram[0x10000];

static unsigned char vidram[0x20000];

static unsigned char* paletteram;
static unsigned char palbank1[0x2000];
static unsigned char palbank2[0x2000];

static int in0=0xff;
static int in1=0xff;
static int in2=0xff;
static int in3=0x01|0x04|0x38|0x40|0x80;
static int in4=0xff;

static int vidoffs=0;
static int vidmodulo=1;
static int fix_bank=0;
static int sram_lock=0;
static int curr_scanline=0;
static int framecount=0;

static int nsprites;
static int nchars;

// a68k

static __inline int irqcall(int i)
{
	return -1;
}
 
void neogeocpucore::a68kReset()
{
	memset(&M68000_regs,0,sizeof(M68000_regs));

	unsigned int sp,pc;

	sp=READ32(rom68k,0,0);
	pc=READ32(rom68k,4,0);

	M68000_regs.pc=pc;
	M68000_regs.a[7]=sp; 
	M68000_regs.srh=0x27;

	M68000_regs.IrqCallback=irqcall;

    M68000_RESET();
}

static __inline void changepc(int pc)
{
}

neogeocpucore::neogeocpucore(tRomSpace* roms)
{
	neogeoromspace* rsp=(neogeoromspace*)roms;

//	rom68k=new unsigned char [0x1000000];
//	memset(rom68k,0,0x1000000);
	rom68k=new unsigned char [0xc20000];
	memset(rom68k,0,0xc20000);

	memcpy(rom68k,rsp->romCPU1,0x100000);
	memcpy(rom68k+0xc00000,rsp->bios,0x20000);		//bios

	// set zone and console/arcade on the bios rom

	ngbios=&rom68k[0xc00000];
	ngram= &rom68k[0x100000];
	bank4= &rom68k[0x200000];

	ngbios[0x401]=0x80;	//machine (0x80 arcade - 0x00 console)
	ngbios[0x400]=0x02;	//country

	cpubkp=new unsigned char [0x200000];
	memcpy(cpubkp,rsp->romCPU1+0x100000,0x200000);

	memcpy(bank4,cpubkp,0x100000);

	memset(ngram,0,0x10000);
	memset(vidram,0,0x20000);
	memset(sram,0x0,0x10000);

	paletteram=palbank2;
	memset(paletteram,0,0x2000);
	paletteram=palbank1;
	memset(paletteram,0,0x2000);

	// patch some bios code

    WRITE_WORD(&ngbios[0x11b00],0x4e71);
    WRITE_WORD(&ngbios[0x11b02],0x4e71);
    WRITE_WORD(&ngbios[0x11b16],0x4ef9);
    WRITE_WORD(&ngbios[0x11b18],0x00c1);
    WRITE_WORD(&ngbios[0x11b1a],0x1b6a);
//    WRITE_WORD(&ngbios[0x11b1a],0x500a);		//memory card

    WRITE_WORD(&ngbios[0x11c14],0x4e71);
    WRITE_WORD(&ngbios[0x11c16],0x4e71);
    WRITE_WORD(&ngbios[0x11c1c],0x4e71);
    WRITE_WORD(&ngbios[0x11c1e],0x4e71);

    WRITE_WORD(&ngbios[0x11c62],0x4e71);
    WRITE_WORD(&ngbios[0x11c64],0x4e71);

	// init pd4990a

	time_t ltime;
	struct tm *today;

	time(&ltime);
	today = localtime(&ltime);

	pd4990a.seconds = ((today->tm_sec/10)<<4) + (today->tm_sec%10);
	pd4990a.minutes = ((today->tm_min/10)<<4) + (today->tm_min%10);
	pd4990a.hours = ((today->tm_hour/10)<<4) + (today->tm_hour%10);
	pd4990a.days = ((today->tm_mday/10)<<4) + (today->tm_mday%10);
	pd4990a.month = (today->tm_mon + 1);
	pd4990a.year = (((today->tm_year%100)/10)<<4) + (today->tm_year%10);
	pd4990a.weekday = today->tm_wday;

	// memory functions

 	a68k_memory_intf.opcode_xor = 0;
 	a68k_memory_intf.read8 = neogeo_m68kr8;
 	a68k_memory_intf.read16 = neogeo_m68kr16;
 	a68k_memory_intf.read32 = neogeo_m68kr32;
 	a68k_memory_intf.write8 = neogeo_m68kw8;
 	a68k_memory_intf.write16 = neogeo_m68kw16;
 	a68k_memory_intf.write32 = neogeo_m68kw32;
 	a68k_memory_intf.changepc = changepc;

 	a68k_memory_intf.read8pc = neogeo_m68kr8;
 	a68k_memory_intf.read16pc = neogeo_m68kr16;
 	a68k_memory_intf.read32pc = neogeo_m68kr32;
 	a68k_memory_intf.read16d = neogeo_m68kr16;
 	a68k_memory_intf.read32d = neogeo_m68kr32;

	OP_ROM=rom68k;
	OP_RAM=rom68k;
	
	a68kReset();

	//

	strcpy(romname,roms->romname);
	nsprites=rsp->nspr;
	nchars=rsp->nchars;
}

neogeocpucore::~neogeocpucore()
{
	delete [] rom68k;
	delete [] cpubkp;

/*	FILE* bk=fopen("vidram.bin","wb");
	fwrite(vidram,0x20000,1,bk);
	fclose(bk);*/
}
 
void neogeocpucore::emulate(int snd)
{
	m68k_ICount=200000;
	M68000_RUN();
	M68000_regs.irq=1;
	pd4990a_addretrace();
	framecount++;
}


// read-write handlers

static unsigned char  neogeo_m68kr8(int address)
{
	address&=0xffffff;

	if (address<=0x0fffff)
	{
		return READ8(rom68k,address,0);
	}
	else if ( (address>=0x100000) && (address<=0x10ffff) )
	{
		return READ8(ngram,address,0x100000);
	}
	else if ( (address>=0x200000) && (address<=0x2fffff) )
	{
		return READ8(bank4,address,0x200000);
	}
	else if ( (address>=0xc00000) && (address<=0xc1ffff) )
	{
		return READ8(ngbios,address,0xc00000);
	}
	else if ( (address>=0xd00000) && (address<=0xd0ffff) )
	{
		return READ8(sram,address,0xd00000);
	}
	else if (address==0x300000)
	{
		// in0 high
		return in0;
	}
	else if (address==0x300001)
	{
		// test, comm setting, free play
		return in3;
	}
	else if (address==0x320000)
	{
		return 0x01;
	}
	else if (address==0x320001)
	{
		// in0 low
		int res;
		int coinflip = pd4990a_testbit_r(0);
		int databit = pd4990a_databit_r(0);

		res=in4^(coinflip<<6)^(databit<<7);
		return (res);
	}
	else if (address==0x300080)
	{
		// test switch high
		return 0x00;
	}
	else if (address==0x300081)
	{
		// test switch
		return 0x80;
	}
	else if (address==0x340000) 
	{
		// controller 2
		return in1;
	}
	else if (address==0x340001)
	{
		//return in1 low (0)
		return 0;
	}
	else if (address==0x380000)
	{
		// in2 high byte
		return in2;
	}
	else if (address==0x380001)
	{
		//low
		return 0;
	}
	else if (address==0x3c0006)
	{
		return (0x80);
	}
	else if ( (address>=0x400000) && (address<=0x401fff) )
	{
		return READ8(paletteram,address,0x400000);
	}

	return 0x0;
}

static unsigned short int neogeo_m68kr16(int address)
{
	address&=0xffffff;

	if (address<=0x0fffff)
	{
		return READ16(rom68k,address,0);
	}
	else if ( (address>=0x100000) && (address<=0x10ffff) )
	{
		return READ16(ngram,address,0x100000);
	}
	else if ( (address>=0x200000) && (address<=0x2fffff) )
	{
		return READ16(bank4,address,0x200000);
	}
	else if ( (address>=0xc00000) && (address<=0xc1ffff) )
	{
		return READ16(ngbios,address,0xc00000);
	}
	else if ( (address>=0xd00000) && (address<=0xd0ffff) )
	{
		return READ16(sram,address,0xd00000);
	}
	else if ( (address>=0x3c0000) && (address<=0x3c0001) )
	{
		return READ16(vidram,vidoffs&0x1ffff,0);
	}
	else if ( (address>=0x3c000a) && (address<=0x3c000b) )
	{
		return READ16(vidram,vidoffs&0x1ffff,0);
	}
	else if ( (address>=0x3c0004) && (address<=0x3c0005) )
	{
		return (vidmodulo>>1);
	}
	else if (address==0x3c0006)
	{
		return 0x8000;
	}
	else if ( (address>=0x400000) && (address<=0x401fff) )
	{
		return READ16(paletteram,address,0x400000);
	}

	return 0x0;
}

static unsigned long int neogeo_m68kr32(int address)
{
	address&=0xffffff;

	if (address<=0x0fffff)
	{
		return READ32(rom68k,address,0);
	}
	else if ( (address>=0x100000) && (address<=0x10ffff) )
	{
		return READ32(ngram,address,0x100000);
	}
	else if ( (address>=0x200000) && (address<=0x2fffff) )
	{
		return READ32(bank4,address,0x200000);
	}
	else if ( (address>=0xc00000) && (address<=0xc1ffff) )
	{
		return READ32(ngbios,address,0xc00000);
	}
	else if ( (address>=0xd00000) && (address<=0xd0ffff) )
	{
		return READ32(sram,address,0xd00000);
	}
	else if ( (address>=0x3c0000) && (address<=0x3c0001) )
	{
		// vidram data read
		return READ32(vidram,vidoffs&0x1ffff,0);
	}
	else if ( (address>=0x3c0004) && (address<=0x3c0005) )
	{
		// vidram modulo read
		return (vidmodulo>>1);
	}
	else if ( (address>=0x400000) && (address<=0x401fff) )
	{
		return READ32(paletteram,address,0x400000);
	}

	return 0x0;

//	return (((neogeo_m68kr16(address)) << 16) + 
//		    ((neogeo_m68kr16(address + 2))&0xffff));
}

static void neogeo_m68kw8(int address, unsigned char value)
{
	address&=0xffffff;

	if ( (address>=0x100000) && (address<=0x10ffff) )
	{
		WRITE8(ngram,address,0x100000,value);
		return;
	}
	else if (address==0x3a001f)
	{
		paletteram=palbank1;
		return ;
	}
	else if (address==0x3a000f)
	{
		paletteram=palbank2;
		return ;
	}
	else if ( (address>=0xd00000) && (address<=0xd0ffff) )
	{
		if (!sram_lock)
		{
			WRITE8(sram,address,0xd00000,value);
		}
		return;
	}
	else if (address==0x300001)
	{
		// reset?
		return;
	}
	else if (address==0x3a000b)
	{
		fix_bank=1;
		return ;
	}
	else if (address==0x3a001b)
	{
		fix_bank=0;
		return ;
	}
	else if (address==0x3a001d)
	{
		// sram unlock
		sram_lock=0;
		return ;
	}
	else if (address==0x3a000d)
	{
		// sram lock
		sram_lock=1;
		return ;
	}
	else if (address==0x380051)
	{
		pd4990a_control_16_w(value);
		return;
	}
	else if (address==0x320000)
	{
		/* z80 command write */
		return;
	}
	else if ( (address>=0x2ffff0) && (address<=0x2fffff) )
	{
		value=value&0x01;
		int bankaddress = (value)*0x100000;
		memcpy(bank4,cpubkp+bankaddress,0x100000);
		return;
	}
	else if ( 
		    (address==0x380031) || (address==0x380041) || 
			(address==0x3a0003) || (address==0x380001) ||
			(address==0x3a0013) || (address==0x3a0001) ||
			(address==0x3a0007) || (address==0x3a0015) ||
			(address==0x380021) || (address==0x3800e1) ||
			(address==0x380061)
			)
	{
		// bios calls
		return ;
	}
}

static void neogeo_m68kw16(int address, unsigned short int value)
{
	address&=0xffffff;

	if ( (address>=0x100000) && (address<=0x10ffff) )
	{
		WRITE16(ngram,address,0x100000,value);
		return;
	}
	else if ( (address>=0x3c0000) && (address<=0x3c0001) )
	{
		vidoffs=value*2;
		return;
	}
	else if ( (address>=0x3c0002) && (address<=0x3c0003) )
	{
		WRITE16(vidram,vidoffs&0x1ffff,0,value);
		vidoffs+=vidmodulo;
		return;
	}
	else if ( (address>=0x3c0004) && (address<=0x3c0005) )
	{
		int v=value;
		if (value&0x8000) 
		{
			/* Sign extend it. */
			vidmodulo=((value-0x10000)<<1);
		}
		else
		{
			vidmodulo=(v<<1);
		}
		return;
	}
	else if ( (address>=0x400000) && (address<=0x401fff) )
	{
		// paletteram write
		WRITE16(paletteram,address,0x400000,value);
		return;
	}
	else if ( (address>=0xd00000) && (address<=0xd0ffff) )
	{
		if (!sram_lock)
		{
			WRITE16(sram,address,0xd00000,value);
		}
		return ;
	}
	// bank switch
	else if ( (address>=0x2ffff0) && (address<=0x2fffff) )
	{
		value=value&0x01;
		int bankaddress = (value)*0x100000;
		memcpy(bank4,cpubkp+bankaddress,0x100000);
		return;
	}
	else if (address==0x3c000c)
	{
		// irq ack
		return ;
	}
}

static void neogeo_m68kw32(int address, unsigned long int value)
{
	address&=0xffffff;

	if ( (address>=0x100000) && (address<=0x10ffff) )
	{
		WRITE32(ngram,address,0x100000,value);
		return;
	}
	else if ( (address>=0x400000) && (address<=0x401fff) )
	{
		// paletteram write
		WRITE32(paletteram,address,0x400000,value);
		return;
	}
	else if ( (address>=0xd00000) && (address<=0xd0ffff) )
	{
		if (!sram_lock)
		{
			WRITE32(sram,address,0xd00000,value);
		}
		return ;
	}
	else if ( (address>=0x3c0000) && (address<=0x3c0001) )
	{
		vidoffs=(value>>16)*2;
		WRITE16(vidram,vidoffs&0x1ffff,0,value&0xffff);
		vidoffs+=vidmodulo;
		return;
	}

//	neogeo_m68kw16(address, (value >> 16) &0xffff);
//	neogeo_m68kw16(address + 2, value & 0xffff);
}

// player controls

void neogeocpucore::update_controls()
{
	if (gui->ginput.coin1==1)
	{
		in4=in4&(~1);
	}
	else
	{
		in4=in4|1;
	}

	if (gui->ginput.start1==1)
	{
		in2=in2&(~0x1);
	}
	else
	{
		in2=in2|0x1;
	}

	if (gui->ginput.p1button[0]==1)
	{
		in0=in0&(~0x10);
	}
	else
	{
		in0=in0|0x10;
	}

	if (gui->ginput.p1button[1]==1)
	{
		in0=in0&(~0x20);
	}
	else
	{
		in0=in0|0x20;
	}

	if (gui->ginput.p1button[2]==1)
	{
		in0=in0&(~0x40);
	}
	else
	{
		in0=in0|0x40;
	}

	if (gui->ginput.p1left==1)
	{
		in0=in0&(~0x04);
	}
	else
	{
		in0=in0|0x04;
	}

	if (gui->ginput.p1right==1)
	{
		in0=in0&(~0x08);
	}
	else
	{
		in0=in0|0x08;
	}

	if (gui->ginput.p1up==1)
	{
		in0=in0&(~0x01);
	}
	else
	{
		in0=in0|0x01;
	}

	if (gui->ginput.p1down==1)
	{
		in0=in0&(~0x02);
	}
	else
	{
		in0=in0|0x02;
	}
}

// palette

void neogeocpucore::build_palette(tCharset* chset)
{
    for (int index=0;index<0x1000;index++)
    {
		int temp;

        temp=READ_WORD(&paletteram[index*2]);

		int red=  (temp>>8)&0x0f;
		int green=(temp>>4)&0x0f;
		int blue=  temp    &0x0f;

        chset->palette[index*4+2] = (red*0x11)&0xff;
        chset->palette[index*4+1] = (green*0x11)&0xff;
        chset->palette[index*4+0] = (blue*0x11)&0xff;
        chset->palette[index*4+3] = 0;
    }
}

/* quick load-save handler */

static const int ssendmarker=		0;
static const int ssneogeoram=		1;
static const int ssbank4=			2;
static const int sssram=			3;
static const int sspalram1=			5;
static const int sspalram2=			6;
static const int ssa68kstate=		7;
static const int ssvidram=			8;
static const int ssvidoffs=			9;
static const int ssvidmodulo=		10;
static const int ssfix_bank=		11;
static const int sssram_lock=		12;

/*  					   */

void neogeocpucore::checkquickstate()
{
	int tag;
	FILE* savestate;
	char ssname[128];

	strcpy(ssname,"save//");
	strcat(ssname,romname);
	strcat(ssname,"qs.sav");

	if (gui->ginput.quicksavestate==1)
	{
		gui->ginput.quicksavestate=0;

		savestate=fopen(ssname,"wb");
		if (savestate==0)
		{
			gui->printMessage(". unable to quicksave state");
			return ;
		}

		tag=ssneogeoram;
		fwrite(&tag,4,1,savestate);
		fwrite(ngram,0x10000,1,savestate);

		tag=ssbank4;
		fwrite(&tag,4,1,savestate);
		fwrite(bank4,0x100000,1,savestate);

		tag=sssram;
		fwrite(&tag,4,1,savestate);
		fwrite(sram,0x10000,1,savestate);

		tag=sspalram1;
		fwrite(&tag,4,1,savestate);
		fwrite(palbank1,0x2000,1,savestate);

		tag=sspalram2;
		fwrite(&tag,4,1,savestate);
		fwrite(palbank2,0x2000,1,savestate);

		tag=ssa68kstate;
		fwrite(&tag,4,1,savestate);
		fwrite(&M68000_regs,sizeof(M68000_regs),1,savestate);

		tag=ssvidram;
		fwrite(&tag,4,1,savestate);
		fwrite(vidram,0x20000,1,savestate);

		tag=ssvidoffs;
		fwrite(&tag,4,1,savestate);
		fwrite(&vidoffs,0x4,1,savestate);

		tag=ssvidmodulo;
		fwrite(&tag,4,1,savestate);
		fwrite(&vidmodulo,0x4,1,savestate);

		tag=ssfix_bank;
		fwrite(&tag,4,1,savestate);
		fwrite(&fix_bank,0x4,1,savestate);

		tag=sssram_lock;
		fwrite(&tag,4,1,savestate);
		fwrite(&sram_lock,0x4,1,savestate);

		// end of savestate
		tag=ssendmarker;
		fwrite(&tag,4,1,savestate);

		fclose(savestate);

		gui->printMessage(". quickstate saved");
	}
	if (gui->ginput.quickloadstate==1)
	{
		int exitflag=0;

		gui->ginput.quickloadstate=0;

		savestate=fopen(ssname,"rb");
		if (savestate==0)
		{
			gui->printMessage(". unable to open quicksave");
			return ;
		}

		while(!exitflag)
		{
			int tag;

			fread(&tag,4,1,savestate);

			if (tag==ssendmarker)
			{
				exitflag=1;
			}
			else if (tag==ssneogeoram)
			{
				fread(ngram,0x10000,1,savestate);
			}
			else if (tag==ssbank4)
			{
				fread(bank4,0x100000,1,savestate);
			}
			else if (tag==sssram)
			{
				fread(sram,0x10000,1,savestate);
			}
			else if (tag==sspalram1)
			{
				fread(palbank1,0x2000,1,savestate);
			}
			else if (tag==sspalram2)
			{
				fread(palbank2,0x2000,1,savestate);
			}
			else if (tag==ssvidram)
			{
				fread(vidram,0x20000,1,savestate);
			}
			else if (tag==ssa68kstate)
			{
				fread(&M68000_regs,sizeof(M68000_regs),1,savestate);
				M68000_regs.IrqCallback=irqcall;
			}
			else if (tag==ssvidoffs)
			{
				fread(&vidoffs,4,1,savestate);
			}
			else if (tag==ssvidmodulo)
			{
				fread(&vidmodulo,4,1,savestate);
			}
			else if (tag==ssfix_bank)
			{
				fread(&fix_bank,4,1,savestate);
			}
			else if (tag==sssram_lock)
			{
				fread(&sram_lock,4,1,savestate);
			}
			else
			{
				exitflag=1;
			}
		}

		fclose(savestate);

		gui->printMessage(". quickstate loaded");
	}
}


/* main draw function */

void neogeocpucore::drawScreen(unsigned long int* video,tCharset* chset)
{
	update_controls();
	build_palette(chset);
	checkquickstate();


	// background palette
#define xpad ((384-304)/2)

	unsigned long int* colt=(unsigned long int*)&chset->palette[0xfff*4+goshift];
	unsigned long int fillcol=*colt;
	for (int ln=0;ln<224;ln++)
	{
		memset(video+ln*384+xpad,fillcol,(384-2*xpad)*4);
	}

	//
	// draw sprites
	// (this is taken from mame source)
	//

	int count,t1,t2,t3,y;
	int offs,tileno,tileatr;
	int sx =0,sy =0,oy =0,my =0,zx = 1, rzy = 1;
    char fullmode=0;
    int ddax=16,dday=256,rzx=15,yskip=0;

    for (count=0;count<0x300;count+=2) 
	{
		t3 = READ_WORD( &vidram[0x10000 + count] );
		t1 = READ_WORD( &vidram[0x10400 + count] );
		t2 = READ_WORD( &vidram[0x10800 + count] );

		if (t1 & 0x40) 
		{
			sx += rzx;
			if ( sx >= 0x1F0 ) sx -= 0x200;

            zx = (t3 >> 8) & 0x0f;
			sy = oy;
		} 
		else 
		{	
			zx = (t3 >> 8) & 0x0f;
			rzy = t3 & 0xff;

			sx = (t2 >> 7);
			if ( sx >= 0x1F0 ) sx -= 0x200;

            my = t1 & 0x3f;
			if (my == 0x20) fullmode = 1;
			else if (my >= 0x21) fullmode = 2;
											
			else fullmode = 0;

			if (sy > 0x100) sy -= 0x200;

			if (fullmode == 2 || (fullmode == 1 && rzy == 0xff))
			{
				while (sy < -16) sy += 2 * (rzy + 1);
			}
			oy = sy;

			sy = 0x200 - (t1 >> 7);
			if (sy > 0x110) sy -= 0x200;
			if (fullmode == 2 || (fullmode == 1 && rzy == 0xff))
			{
				while (sy < 0) sy += 2 * (rzy + 1);
			}
			oy = sy;


		  	if(my==0x21) my=0x20;
			else if(rzy!=0xff && my!=0)
				my=((my*16*256)/(rzy+1) + 15)/16;

            if(my>0x20) my=0x20;

            ddax=16;
		}

		if (my==0) continue;

		if(zx!=15) {
			rzx=0;
			for(int i=0;i<16;i++) {
				ddax-=zx+1;
				if(ddax<=0) {
					ddax+=15+1;
//					dda_x_skip[i]=1;
					rzx++;
				}
//				else dda_x_skip[i]=0;
			}
		}
		else rzx=16;

		if (sx>320) continue;

		if(rzy==255) yskip=16;
		else dday=256;

		offs = count<<6;

        for (y=0; y < my ;y++) 
		{
			tileno  = READ_WORD(&vidram[offs]);
            offs+=2;
			tileatr = READ_WORD(&vidram[offs]);
            offs+=2;

			if (nsprites>0x10000)
			{
				if (tileatr&0x10) tileno+=0x10000;
			}
			if (nsprites>0x20000)
			{
				if (tileatr&0x20) tileno+=0x20000;
			}

            if (tileatr&0x8) tileno=(tileno&~7)+((tileno+framecount)&7);
            else if (tileatr&0x4) tileno=(tileno&~3)+((tileno+framecount)&3);

			if (fullmode == 2 || (fullmode == 1 && rzy == 0xff))
			{
				if (sy >= 248) sy -= 2 * (rzy + 1);
			}
			else if (fullmode == 1)
			{
				if (y == 0x10) sy -= 2 * (rzy + 1);
			}
			else if (sy > 0x110) sy -= 0x200;


            if(rzy!=255)
            {
            	yskip=0;
                for(int i=0;i<16;i++)
                {
                    dday-=rzy+1;
                    if(dday<=0)
                    {
                    	dday+=256;
                    	yskip++;
                    }
                }
            }

			if ((tileatr>>8) != 0)
			{
				chset->printSprite(video,tileno%nsprites,(tileatr>>8)&0xff,
					               32+sx,sy-16,tileatr&0x1,tileatr&0x2);
			}

			sy +=yskip;
		}
	}

	//
	// draw fixed charset
	//

	for (y=2;y<30;y++) 
	{
 		for (int x=1;x<39;x++) 
		{
  			int byte1 = (READ_WORD( &vidram[(0xe000+(y*2)+(x*64))&0x1ffff] ));
  			int byte2 = byte1 >> 12;
           	byte1&=0xfff;
			if (nchars>0x1000)
			{
				chset->printChar(video,byte1+fix_bank*0x1000,byte2,24+(x*8),((y-2)*8),0,0);
			}
			else
			{
				chset->printChar(video,byte1,byte2,24+(x*8),((y-2)*8),0,0);
			}
		}
	}
}
