
/*

  NeoGeo romspace

*/

#include "neogeoromspace.h"
#include "neogeocharset.h"
#include "gui.h"


extern tGui* gui;
extern tRomSet romsetList[];


// constructor

neogeoromspace::neogeoromspace()
{
	romCPU1=0;
	romCPU2=0;
	bios=0;
}

// destructor

neogeoromspace::~neogeoromspace()
{
	if (romCPU1!=0) delete [] romCPU1;
	if (romCPU2!=0) delete [] romCPU2;
	if (bios!=0) delete [] bios;
}

// read roms

int neogeoromspace::readRoms(char* path,char* romsetname,tCharset** chs)
{
	// load in graphic data

	int err;
	int rs;
	long int i,j;
	unsigned char* temprom;
	char fullname[1024];
	tRomSet* rsp;

	if (romsetByName(romsetname,&rs)) return 1;
	rsp=&romsetList[rs];

	temprom=new unsigned char [0x800000];

	// load code roms & bios

	bios=new unsigned char [0x20000];
	strcpy(fullname,path);
	strcat(fullname,"neogeo.zip");

	err=loadZippedRom(fullname,
					  "neo-geo.rom",
					  0x20000,
					  bios);
	if (err!=0)
	{
		gui->printMessage(". Unable to load neogeo bios");
		return 1;
	}

	//

	romCPU1=new unsigned char [0x300000];
	memset(romCPU1,0,0x300000);

	for (i=0;i<rsp->ncoderoms;i++)
	{
		strcpy(fullname,path);
		strcat(fullname,rsp->romlist[i].romset);
		strcat(fullname,".zip");

		err=loadZippedRom(fullname,
						  rsp->romlist[i].name,
						  rsp->romlist[i].size,
						  temprom);

		if (rsp->romlist[i].loadtype==ROM_LOAD_NORMAL)
		{
			memcpy(romCPU1+rsp->romlist[i].position,temprom,rsp->romlist[i].size);
		}
		else if (rsp->romlist[i].loadtype==ROM_LOAD_HALFSPLIT)
		{
			int rsize2=rsp->romlist[i].size/2;

			memcpy(romCPU1+rsp->romlist[i].position+rsize2,temprom,rsize2);
			memcpy(romCPU1+rsp->romlist[i].position,temprom+rsize2,rsize2);
		}

		if (err!=0)
		{
			gui->printMessage(". Unable to load code rom");
			return 1;
		}
	}

	//
	//

	if (*chs!=0) delete *chs;

	if (rsp->romsettype==ROMSET_NEOGEO)
	{
		*chs=new neogeocharset(rsp->nspr,rsp->nchars,0,0);
	}
	neogeocharset* pchs=(neogeocharset*)(*chs);

	// read char gfx roms into gfx1
	// (char gfx roms should be always 2)

	pchs->gfx1=new unsigned char [0x100000];

	for (i=0;i<2;i++)
	{
		strcpy(fullname,path);
		strcat(fullname,rsp->romlist[i+rsp->ncoderoms].romset);
		strcat(fullname,".zip");

		err=loadZippedRom(fullname,
						  rsp->romlist[i+rsp->ncoderoms].name,
						  rsp->romlist[i+rsp->ncoderoms].size,
						  pchs->gfx1+rsp->romlist[i+rsp->ncoderoms].position);
		if (err!=0)
		{
			gui->printMessage(". Unable to load tile rom");
			return 1;
		}
	}

	// read tile16 gfx roms into gfx2

	int gromsize=rsp->romlist[rsp->ncoderoms+2].size*rsp->ngfxroms;
	pchs->gfx2=new unsigned char [gromsize];

	for (i=(rsp->ncoderoms+2);i<(rsp->ncoderoms+2+rsp->ngfxroms);i++)
	{
		strcpy(fullname,path);
		strcat(fullname,rsp->romlist[i].romset);
		strcat(fullname,".zip");

		err=loadZippedRom(fullname,
						  rsp->romlist[i].name,
						  rsp->romlist[i].size,
						  temprom);
		if (err!=0)
		{
			gui->printMessage(". Unable to load gfx2 rom");
			return 1;
		}

		if (rsp->romlist[i].loadtype==ROM_LOAD_ODD)
		{
			for (j=0;j<rsp->romlist[i].size;j++)
			{
				pchs->gfx2[rsp->romlist[i].position+j*2+1]=temprom[j];
			}
		}
		else if (rsp->romlist[i].loadtype==ROM_LOAD_EVEN)
		{
			for (j=0;j<rsp->romlist[i].size;j++)
			{
				pchs->gfx2[rsp->romlist[i].position+j*2]=temprom[j];
			}
		}
		else if (rsp->romlist[i].loadtype==ROM_LOAD_HSEVEN)
		{
			int rsize2=rsp->romlist[i].size/2;

			for (j=0;j<rsp->romlist[i].size;j++)
			{
				pchs->gfx2[rsp->romlist[i].position+j*2]=temprom[j];
			}

			// halph swap

			for (j=0;j<rsize2*2;j+=2)
			{
				unsigned char tmp=pchs->gfx2[j+rsp->romlist[i].position];
				pchs->gfx2[j+rsp->romlist[i].position]=pchs->gfx2[j+rsp->romlist[i].position+rsize2*2];
				pchs->gfx2[j+rsp->romlist[i].position+rsize2*2]=tmp;
			}
		}
		else if (rsp->romlist[i].loadtype==ROM_LOAD_HSODD)
		{
			int rsize2=rsp->romlist[i].size/2;

			for (j=0;j<rsp->romlist[i].size;j++)
			{
				pchs->gfx2[rsp->romlist[i].position+j*2+1]=temprom[j];
			}

			// halph swap

			for (j=0;j<rsize2*2;j+=2)
			{
				unsigned char tmp=pchs->gfx2[j+1+rsp->romlist[i].position];
				pchs->gfx2[j+1+rsp->romlist[i].position]=pchs->gfx2[j+1+rsp->romlist[i].position+rsize2*2];
				pchs->gfx2[j+1+rsp->romlist[i].position+rsize2*2]=tmp;
			}
		}
	}

	// store important info
	strcpy(romname,rsp->name);
	orientation=rsp->orientation;
	romsettype=rsp->romsettype;
	nspr=rsp->nspr;
	nchars=rsp->nchars;

	delete [] temprom;

	return 0;
}

/*

  functions to decode the charset

*/

void neogeoromspace::decodechars(int num,unsigned char *src,unsigned char** dest)
{
	int nplanes=4;
	int plane,x,y;
	unsigned char* dp;
	unsigned char* tmpdp;
	int charincrement=32*8;

	long int planeoffset[4]={0,1,2,3};
	long int xoffset[8]={ 33*4, 32*4, 49*4, 48*4, 1*4, 0*4, 17*4, 16*4 };
	long int yoffset[8]={ 0*8, 1*8, 2*8, 3*8, 4*8, 5*8, 6*8, 7*8 };

	unsigned char* psrc=src;

	for (plane = 0;plane < nplanes;plane++)
	{
		int offs;

		offs = num*charincrement+planeoffset[plane];

		for (y = 0;y < 8;y++)
		{
			tmpdp=dest[num];
			dp=tmpdp+y*8;

			for (x = 0;x < 8;x++)
			{
				int xoffs,yoffs;

				if (plane == 0) dp[x] = 0;
				else dp[x] <<= 1;

				xoffs = x;
				yoffs = y;

				dp[x]+=readbit(psrc,offs+yoffset[yoffs]+xoffset[xoffs]);
			}
		}
	}
}

void neogeoromspace::decodesprites(int num,unsigned char *src,unsigned char* dest)
{
	int x,y;
	long int xo,yo;
	long int offz0,offz1,offz2,offz3;

	long int xoffset[16]={ 64*8+7, 64*8+6, 64*8+5, 64*8+4, 64*8+3, 64*8+2, 64*8+1, 64*8+0, 7, 6, 5, 4, 3, 2, 1, 0 };
	long int yoffset[16]={ 0*32, 1*32, 2*32, 3*32, 4*32, 5*32, 6*32, 7*32, 8*32, 9*32, 10*32, 11*32, 12*32, 13*32, 14*32, 15*32 };

	long int* pxo;
	long int* pyo=yoffset;

	unsigned char src2[128];

	offz0=3*8;
	offz1=1*8;
	offz2=2*8;
	offz3=0;

	memcpy(src2,src+num*128,128);

	// rearrange the tiles in order to be sequential

	for (y = 0;y < 16;y++)
	{
		yo=*pyo;
		pxo=xoffset;
		for (x = 0;x < 16;x+=2)
		{
			long int bn;
			unsigned char n1,n2;		// two nibbles
			xo=*pxo;

			n1=0;
			bn=offz0+yo+xo;
			n1+=((src2[bn>>3]>>(7-(bn&7)))&1)<<3;
			bn=offz1+yo+xo;
			n1+=((src2[bn>>3]>>(7-(bn&7)))&1)<<2;
			bn=offz2+yo+xo;
			n1+=((src2[bn>>3]>>(7-(bn&7)))&1)<<1;
			bn=offz3+yo+xo;
			n1+=(src2[bn>>3]>>(7-(bn&7)))&1;

			pxo++;
			xo=*pxo;

			n2=0;
			bn=offz0+yo+xo;
			n2+=((src2[bn>>3]>>(7-(bn&7)))&1)<<3;
			bn=offz1+yo+xo;
			n2+=((src2[bn>>3]>>(7-(bn&7)))&1)<<2;
			bn=offz2+yo+xo;
			n2+=((src2[bn>>3]>>(7-(bn&7)))&1)<<1;
			bn=offz3+yo+xo;
			n2+=(src2[bn>>3]>>(7-(bn&7)))&1;

			pxo++;

			dest[num*128+(x/2)+(y*8)]=n1|(n2<<4);
		}
		pyo++;
	}
}

void neogeoromspace::decodeCharset(tCharset* chset)
{
	int i;
	neogeocharset* pchs=(neogeocharset*)chset;

	// decode 8x8 chars
	for (i=0;i<nchars;i++)
	{
		decodechars(i,pchs->gfx1,chset->chars);
	}

	// interleave sprites

	for (i=0;i<nspr;i++)
	{
		decodesprites(i,pchs->gfx2,pchs->gfx2);
	}
}