
/*

  system 16 romspace

*/


#include "s16romspace.h"
#include "s16charset.h"
#include "gui.h"


extern tGui* gui;
extern tRomSet romsetList[];


// constructor

s16romspace::s16romspace()
{
	romCPU1=0;
	romCPU2=0;
}

// destructor

s16romspace::~s16romspace()
{
	if (romCPU1!=0) delete [] romCPU1;
	if (romCPU2!=0) delete [] romCPU2;
}

// read roms

int s16romspace::readRoms(char* path,char* romsetname,tCharset** chs)
{
	int err;
	int rs;
	long int i,j;
	unsigned char* temprom;
	char fullname[1024];
	tRomSet* rsp;

	if (romsetByName(romsetname,&rs)) return 1;
	rsp=&romsetList[rs];

	// load in system 16 cpu1 romz
	temprom=new unsigned char [0x40000];
	romCPU1=new unsigned char [0x80000];
	romCPU2=new unsigned char [0x8000];

	memset(romCPU1,0,0x80000);
	memset(romCPU2,0,0x8000);

	for (i=0;i<rsp->ncoderoms;i++)
	{
		// load rom into temprom

		strcpy(fullname,path);
		strcat(fullname,rsp->romlist[i].romset);
		strcat(fullname,".zip");

		err=loadZippedRom(fullname,
						  rsp->romlist[i].name,
						  rsp->romlist[i].size,
						  temprom);
		if (err!=0)
		{
			gui->printMessage(". Unable to load code rom");
			return 1;
		}

		// type of rom?

		if (rsp->romlist[i].loadtype==ROM_LOAD_ODD)
		{
			for (j=0;j<rsp->romlist[i].size;j++)
			{
				romCPU1[rsp->romlist[i].position+j*2+1]=temprom[j];
			}
		}
		else if (rsp->romlist[i].loadtype==ROM_LOAD_EVEN)
		{
			for (j=0;j<rsp->romlist[i].size;j++)
			{
				romCPU1[rsp->romlist[i].position+j*2]=temprom[j];
			}
		}
		else
		{
			gui->printMessage(".Invalid rom load format");
			return 1;
		}
	}

	// allocate a system 1x charset
	// how big is the objram?


	if (*chs!=0) delete *chs;

	if (rsp->romsettype==ROMSET_S16)
	{
		*chs=new s16charset(rsp->nspr,rsp->nchars,0,0,0x100000);
	}
	else if (rsp->romsettype==ROMSET_S18)
	{
		*chs=new s16charset(rsp->nspr,rsp->nchars,0,0,0x400000);
	}
	s16charset* pchs=(s16charset*)(*chs);

	pchs->gfx1=new unsigned char [0xc0000];
	pchs->gfx2=new unsigned char [0x200000];

	// read char gfx roms into gfx1

	for (i=0;i<3;i++)
	{
		strcpy(fullname,path);
		strcat(fullname,rsp->romlist[i+rsp->ncoderoms].romset);
		strcat(fullname,".zip");

		if (rsp->romlist[i+rsp->ncoderoms].loadtype==ROM_LOAD_NORMAL)
		{
			err=loadZippedRom(fullname,
							  rsp->romlist[i+rsp->ncoderoms].name,
							  rsp->romlist[i+rsp->ncoderoms].size,
							  pchs->gfx1+rsp->romlist[i+rsp->ncoderoms].position);
			if (err!=0)
			{
				gui->printMessage(".Unable to load graphic rom");
				return 1;
			}
		}
	}

	// store average rom size
	gromsize=rsp->romlist[rsp->ncoderoms].size;

	// read tile gfx rom into gfx2

	for (i=0;i<8;i++)
	{
		strcpy(fullname,path);
		strcat(fullname,rsp->romlist[i+rsp->ncoderoms+3].romset);
		strcat(fullname,".zip");

		err=loadZippedRom(fullname,
						  rsp->romlist[i+rsp->ncoderoms+3].name,
						  rsp->romlist[i+rsp->ncoderoms+3].size,
						  temprom);
		if (err!=0)
		{
			gui->printMessage(". Unable to load tile rom");
			return 1;
		}

		// type of rom?

		if (rsp->romlist[i+rsp->ncoderoms+3].loadtype==ROM_LOAD_ODD)
		{
			for (j=0;j<rsp->romlist[i+rsp->ncoderoms+3].size;j++)
			{
				pchs->gfx2[rsp->romlist[i+rsp->ncoderoms+3].position+j*2+1]=temprom[j+7];
			}
		}
		else if (rsp->romlist[i+rsp->ncoderoms+3].loadtype==ROM_LOAD_EVEN)
		{
			for (j=0;j<rsp->romlist[i+rsp->ncoderoms+3].size;j++)
			{
				pchs->gfx2[rsp->romlist[i+rsp->ncoderoms+3].position+j*2]=temprom[j+7];
			}
		}
		else
		{
			gui->printMessage(".Invalid rom load format");
			return 1;
		}
	}

	// read z80 sound rom

	for (i=0;i<rsp->nsndcoderoms;i++)
	{
		strcpy(fullname,path);
		strcat(fullname,rsp->romlist[i+15].romset);
		strcat(fullname,".zip");

		err=loadZippedRom(fullname,
						  rsp->romlist[i+15].name,
						  rsp->romlist[i+15].size,
						  romCPU2+rsp->romlist[i+15].position);
		if (err!=0)
		{
			gui->printMessage(". Unable to load tile rom");
			return 1;
		}
	}

	// store important info
	strcpy(romname,rsp->name);
	orientation=rsp->orientation;
	romsettype=rsp->romsettype;
	nspr=rsp->nspr;
	nchars=rsp->nchars;

	// deallocate temporary space
	delete [] temprom;

	return 0;
}

/*

  functions to decode the charset

*/

void s16romspace::decodechars(int num,unsigned char *src,unsigned char** dest)
{
    int x, y, c, i;

    for (y=0;y<8;y++)
    {
        i=(num<<3)|y;

        for (x=0;x<8;x++)
        {
            c  = ((src[i] >> (7 - x)) & 1) << 0;
            c |= ((src[gromsize+i] >> (7 - x)) & 1) << 1;
            c |= ((src[gromsize*2+i] >> (7 - x)) & 1) << 2;
            c &= 0x07;

            dest[num][y*8+x]=c;
        }
    }

/*
        if(count == 0x40)
            bg_tile_info[name] |= TILE_TRANSPARENT;
        else if(count == 0x00)
            bg_tile_info[name] |= TILE_OPAQUE;
        else if(count != 0x00)
            bg_tile_info[name] |= TILE_SEMITRANSPARENT;
*/
}

void s16romspace::decodeCharset(tCharset* chset)
{
	int i,c;

	s16charset* pchs=(s16charset*)chset;

	for (i=0;i<nchars;i++)
	{
		decodechars(i,pchs->gfx1,chset->chars);
	}

	// decode 4bitperpixel into 8bpp

    for (i=0;i<nspr;i++)
    {
        c=(pchs->gfx2[i]>>4)&0x0f;
        if (c==0x0f) c=0x00;
        pchs->obj[(i<<1)]=c;

        c=pchs->gfx2[i]&0x0f;
        if (c==0x0f) c=0x00;
        pchs->obj[(i<<1)|1]=c;
    }
}