
/*

  System 18 machine inner-workings
  (some help from David Raingeard, some help from Mame)

*/


#include "s18cpucore.h"
#include "s16charset.h"

// profiling

//#define PROFILE

#ifdef PROFILE
int srr;
int txtrr;
int tilerr;
int objrr;
int palrr;
#endif

// forward declarations

unsigned char s18_m68kr8(int address);
unsigned short int s18_m68kr16(int address);
unsigned long int s18_m68kr32(int address);
void s18_m68kw8(int address, unsigned char value);
void s18_m68kw16(int address, unsigned short int value);
void s18_m68kw32(int address, unsigned long int value);

// control variables
 
extern tGui* gui;

static unsigned char* rom68k;
static unsigned char s18ram[0x4000];
static unsigned char tileram[0x10000];
static unsigned char textram[0x1000]; 
static unsigned char objram[0x1000];  
static unsigned char palram[0x1000];  
static unsigned char extram[8];

static unsigned short int fg_page[4];
static unsigned short int bg_page[4];
static unsigned short int fg2_page[4];
static unsigned short int bg2_page[4];
static unsigned short int fg_scrollx, fg_scrolly;
static unsigned short int bg_scrollx, bg_scrolly;
static unsigned short int fg2_scrollx, fg2_scrolly;
static unsigned short int bg2_scrollx, bg2_scrolly;
static int sys16_tile_bank0;
static int sys16_tile_bank1;

static int gencontrols=0xff;
static int p1controls=0xff;
static int p2controls=0xff;
static int dsw1=0xff;
static int dsw2=0xff;

extern int goshift;

// a68k

static __inline int irqcall(int i)
{
	return -1;
}
 
void s18cpucore::a68kReset()
{
	memset(&M68000_regs,0,sizeof(M68000_regs));

	unsigned int sp,pc;

	sp=READ32(rom68k,0,0);
	pc=READ32(rom68k,4,0);

	M68000_regs.pc=pc;
	M68000_regs.a[7]=sp; 
	M68000_regs.srh=0x27;

	M68000_regs.IrqCallback=irqcall;

    M68000_RESET();
}

static __inline void changepc(int pc)
{
}

s18cpucore::s18cpucore(tRomSpace* roms)
{
	rom68k=new unsigned char [0x80000];

	memcpy(rom68k,roms->romCPU1,0x80000);
	memset(s18ram,0,0x4000);
	memset(tileram,0,0x10000);
	memset(textram,0,0x1000);
	memset(palram,0,0x1000);
	memset(objram,0,0x1000);
	memset(extram,0,8);

	// memory functions

 	a68k_memory_intf.opcode_xor = 0;
 	a68k_memory_intf.read8 = s18_m68kr8;
 	a68k_memory_intf.read16 = s18_m68kr16;
 	a68k_memory_intf.read32 = s18_m68kr32;
 	a68k_memory_intf.write8 = s18_m68kw8;
 	a68k_memory_intf.write16 = s18_m68kw16;
 	a68k_memory_intf.write32 = s18_m68kw32;
 	a68k_memory_intf.changepc = changepc;

 	a68k_memory_intf.read8pc = s18_m68kr8;
 	a68k_memory_intf.read16pc = s18_m68kr16;
 	a68k_memory_intf.read32pc = s18_m68kr32;
 	a68k_memory_intf.read16d = s18_m68kr16;
 	a68k_memory_intf.read32d = s18_m68kr32;

	OP_ROM=rom68k;
	OP_RAM=rom68k;
	
	a68kReset();

	strcpy(romname,roms->romname);

	// profiling

#ifdef PROFILE
	srr=0;
	txtrr=0;
	tilerr=0;
	objrr=0;
	palrr=0;
#endif
}

s18cpucore::~s18cpucore()
{
	delete [] rom68k;

#ifdef PROFILE
	char str[1024];

	sprintf(str,"srr [%d] txtrr [%d] objrr [%d] palrr [%d] tilerr [%d]\n",
		    srr,txtrr,objrr,palrr,tilerr);
	MessageBox(0,str,"info",MB_OK);
#endif
}
 
void s18cpucore::emulate(int snd)
{
//	if (!snd)
	{
		m68k_ICount=166666;
		M68000_RUN();
		M68000_regs.irq=4;
	}
}

// player controls

void s18cpucore::update_controls()
{
	if (gui->ginput.coin1==1)
	{
		gencontrols=gencontrols&(~1);
	}
	else
	{
		gencontrols=gencontrols|1;
	}

	if (gui->ginput.start1==1)
	{
		gencontrols=gencontrols&(~0x10);
	}
	else
	{
		gencontrols=gencontrols|0x10;
	}

	if (gui->ginput.p1left==1)
	{
		p1controls=p1controls&(~0x80);
	}
	else
	{
		p1controls=p1controls|0x80;
	}

	if (gui->ginput.p1right==1)
	{
		p1controls=p1controls&(~0x40);
	}
	else
	{
		p1controls=p1controls|0x40;
	}

	if (gui->ginput.p1up==1)
	{
		p1controls=p1controls&(~0x20);
	}
	else
	{
		p1controls=p1controls|0x20;
	}

	if (gui->ginput.p1down==1)
	{
		p1controls=p1controls&(~0x10);
	}
	else
	{
		p1controls=p1controls|0x10;
	}

	if (gui->ginput.p1button[0]==1)
	{
		p1controls=p1controls&(~0x01);
	}
	else
	{
		p1controls=p1controls|0x01;
	}

	if (gui->ginput.p1button[1]==1)
	{
		p1controls=p1controls&(~0x02);
	}
	else
	{
		p1controls=p1controls|0x02;
	}

	if (gui->ginput.p1button[2]==1)
	{
		p1controls=p1controls&(~0x04);
	}
	else
	{
		p1controls=p1controls|0x04;
	}
}

// read-write handlers

static unsigned char  s18_m68kr8(int address)
{
	address&=0xffffff;

	if (address>=0xffc000)
	{
		return READ8(s18ram,address,0xffc000);
	}
	else if (address<=0x7ffff)
	{
		return READ8(rom68k,address,0);
	}
	else if ( (address>=0x400000) && (address<=0x40ffff) )
	{
		return READ8(tileram,address,0x400000);
	}
	else if (address==0xe40009)
	{
		//general controlz
		return gencontrols;
	}
	else if (address==0xe40001)
	{
		//p1
		return p1controls;
	}
	else if (address==0xe40003)
	{
		//p2
		return p2controls;
	}
	else if (address==0xe4000b)
	{
		// dip switch 1
		return dsw1;
	}
	else if (address==0xe4000d)
	{
		// dip switch 2
		return dsw2;
	}

	return 0x0;
}

static unsigned short int s18_m68kr16(int address)
{
	address&=0xffffff;

	if (address>=0xffc000)
	{
		return READ16(s18ram,address,0xffc000);
	}
	else if (address<=0x7ffff)
	{
		return READ16(rom68k,address,0);
	}
	else if ( (address>=0x400000) && (address<=0x40ffff) )
	{
		return READ16(tileram,address,0x400000);
	}
	else if ( (address>=0x410000) && (address<=0x410fff) )
	{
		return READ16(textram,address,0x410000);
	}
	else if ( (address>=0xc00000) && (address<=0xc00007) )
	{
		return READ16(extram,address,0xc00000);
	}

	return 0;
}

static unsigned long int s18_m68kr32(int address)
{
	address&=0xffffff;

	if (address>=0xffc000)
	{
		return READ32(s18ram,address,0xffc000);
	}
	else if (address<=0x7ffff)
	{
		return READ32(rom68k,address,0);
	}
	else if ( (address>=0x410000) && (address<=0x410fff) )
	{
		return READ32(textram,address,0x410000);
	}
	else if ( (address>=0x400000) && (address<=0x40ffff) )
	{
		return READ32(tileram,address,0x400000);
	}
	else if ( (address>=0xc00000) && (address<=0xc00007) )
	{
		return READ32(extram,address,0xc00000);
	}

	return 0;
}

static void s18_m68kw8(int address, unsigned char value)
{
	address&=0xffffff;

	if (address>=0xffc000)
	{
		WRITE8(s18ram,address,0xffc000,value);
		return;
	}
	else if ( (address>=0x410000) && (address<=0x410fff) )
	{
		WRITE8(textram,address,0x410000,value);
		return;
	}
	else if ( (address>=0x400000) && (address<=0x40ffff) )
	{
		WRITE8(tileram,address,0x400000,value);
		return;
	}
	else if ( (address>=0xe40000) && (address<=0xe4001f) )
	{
		return;
	}
	else if (address==0xe43035)
	{
		return;
	}
	else if (address==0xfe0007)
	{
		return;
	}

	address=address;
}

static void s18_m68kw16(int address, unsigned short int value)
{
	address&=0xffffff;

	if (address>=0xffc000)
	{
		WRITE16(s18ram,address,0xffc000,value);
		return;
	}
	else if ( (address>=0x410000) && (address<=0x410fff) )
	{
		WRITE16(textram,address,0x410000,value);
		return;
	}
	else if ( (address>=0x400000) && (address<=0x40ffff) )
	{
		WRITE16(tileram,address,0x400000,value);
		return;
	}
	else if ( (address>=0x440000) && (address<=0x440fff) )
	{
		WRITE16(objram,address,0x440000,value);
		return;
	}
	else if ( (address>=0xc00000) && (address<=0xc00007) )
	{
		WRITE16(extram,address,0xc00000,value);
		return;
	}
	else if ( (address>=0x840000) && (address<=0x840fff) )
	{
		WRITE16(palram,address,0x840000,value);
		return;
	}
	else if ( (address>=0xfe0000) && (address<=0xfe0040) )
	{
		return;
	}

	address=address;
}

static void s18_m68kw32(int address, unsigned long int value)
{
	address&=0xffffff;

	if (address>=0xffc000)
	{
		WRITE32(s18ram,address,0xffc000,value);
		return;
	}
	else if ( (address>=0x400000) && (address<=0x40ffff) )
	{
		WRITE32(tileram,address,0x400000,value);
		return;
	}
	else if ( (address>=0x840000) && (address<=0x840fff) )
	{
		WRITE32(palram,address,0x840000,value);
		return;
	}
	else if ( (address>=0x440000) && (address<=0x440fff) )
	{
		WRITE32(objram,address,0x440000,value);
		return;
	}
	else if ( (address>=0x410000) && (address<=0x410fff) )
	{
		WRITE32(textram,address,0x410000,value);
		return;
	}
	else if ( (address>=0xc00000) && (address<=0xc00007) )
	{
		WRITE32(extram,address,0xc00000,value);
		return;
	}

	address=address;
}

/* draw helpers */

void s18cpucore::draw_sprites(unsigned long int* video,tCharset* chset,int pri)
{
/*
   440000-4403FF: The sprites registers. There are 16 control bytes for each
                  sprite (64 sprites maximum). This is the function of the 16
                  bytes.

   00 : last sprite line on the screen
   01 : first sprites line on the screen (sprites height = [00]-[01])
   02 : high value of the sprite X position
   03 : low value of the sprite X position (X = [02]*256+[03])
   04 : horizontal sprite flipping flag (flip if value = 01)
   05 : sprite width (real width = [05]*2) (!! signed byte)
        if the bit7 is set to 1, the sprite is flipped vertically
   06 : high value of sprite rom position 
   07 : low value of sprite rom position
   08 : sprite bank selector
        bit7-bit4 are always at 1 if the sprite is active
        bit3-bit0 is the sprite bank number (0-F)
   09 : sprites priority and base color register selector
        bit7-bit6: 10 - foreground sprite
                   01 - background sprite
                   11 - sprite over everything
        bit5-bit0: base color registers (is 1024+16*value)
   0A : zoom function \
   0B : zoom function - unknown format (never used in Shinobi)
   0C to 0F are not used in current emulated games
*/

	int name;
    int count;
	int bank[16] = { 0,2,4,6,8,0x0A,0x0C,0x0E,0x10,0x12,0x14,0x16,0x18,0x1A,0x1C,0x1E };
	s16charset* pchs=(s16charset*)chset;

    for (count=0;count<0x80;count++)
    {
        unsigned char* p=&objram[count<<3];
        int start = p[0];		// first sprite line on screen
        int end = p[1];			// last sprite line on screen
//		int priority=(p[8]&0xc0);

		int s2=(p[5]<<8)|p[4];
		int s3=(p[7]<<8)|p[6];

        if ( (s2==0x8000) && (end==0xff) ) return ;

//		if (priority==pri)
		{
/*
   09 : sprites priority and base color register selector
        bit7-bit6: 10 - foreground sprite
                   01 - background sprite
                   11 - sprite over everything
        bit5-bit0: base color registers (is 1024+16*value)
*/

			int pal=(p[8]&0x3f)<<4;
     
			// xposition of the sprite
			// p[2] -> low byte
			// p[3] -> high byte

			int xpos=(p[3]<<8|p[2])-(160+24);
			int width=p[4]&0x7f;
			int height=(end-start);
			int flipx=p[5]&1;
			int flipy=p[4]&0x80;

			// p[4] -> vertical flip

			if (flipy)
			{
				width=0x80-width;
			}
			width*=4;

			// sprite code
			// p[6] -> low byte
			// p[7] -> high byte

			name = 4*(s3|((bank[(p[9])&0xf])*0x8000));

//			if (p[9]&0xf0)		// sprite active
			{
				pchs->printSprite(video,name,pal,
					              xpos,start,width,height,
								  flipx,flipy,0,0);
			}
		}
    }
}

void s18cpucore::draw_bg(unsigned long int* video,tCharset* chset,int scrollx,int scrolly,unsigned short int *page,int bg)
{
	int x,y;
	int pag;
	unsigned short int attr;
	int pal,name;
	s16charset* pch=(s16charset*)chset;

	if (bg)
	{
		for (y=0;y<29;y++)
		{
			int ry=((scrolly/8)+y);
			for (x=4;x<46;x++)
			{
				int rx=((scrollx/8)+x);

				pag=(((rx&0x40)>>6)|(((ry&0x20)>>5)<<1))&3;
				attr = *(unsigned short int *)&tileram[((page[pag]*64*32)+(((rx+goshift)&0x3f)+(ry&0x1f)*64))*2];

				int bk;
				if (attr&0x1000) bk=sys16_tile_bank1;
				else bk=sys16_tile_bank0;
				name = (attr&0xfff)|(bk*0x1000);

				// palette
				pal=((attr>>6)&0x7F)<<3;

				pch->printbgChar(video,name,pal,(x*8)-(scrollx&7),(y*8)-(scrolly&7),0,0);
			}
		}
	}
	else
	{
		for (y=0;y<29;y++)
		{
			int ry=((scrolly/8)+y);
			for (x=4;x<46;x++)
			{
				int rx=((scrollx/8)+x);

				pag=(((rx&0x40)>>6)|(((ry&0x20)>>5)<<1))&3;
				attr = *(unsigned short int *)&tileram[((page[pag]*64*32)+((rx&0x3f)+(ry&0x1f)*64))*2];

				int bk;
				if (attr&0x1000) bk=sys16_tile_bank1;
				else bk=sys16_tile_bank0;
				name = (attr&0xfff)|(bk*0x1000);

				// palette
				pal=((attr>>6)&0x7F)<<3;

				pch->printfgChar(video,name,pal,x*8-(scrollx&7),y*8-(scrolly&7),0,0);
			}
		}
	}

}

void s18cpucore::draw_text(unsigned long int* video,tCharset* chset)
{
    int xx,yy,name,cc;
	s16charset* pch=(s16charset*)chset;

    for (yy=0;yy<28;yy++)
    {
        for (xx=0;xx<40;xx++)
        {
            name = *(unsigned short int*)&textram[((yy << 6) + (xx + 24)) << 1];
            cc = (name >> 6) & 0x38;
            name = (name & 0x1FF)+sys16_tile_bank0*0x1000;

			int sx=xx*8+((384-320)/2);
			int sy=(yy*8);
			pch->printfgChar(video,name,cc,sx,sy,0,0);
        }
    }
}

/* palette build */

void s18cpucore::build_palette(tCharset* chset)
{
    int index;
    unsigned short int temp;

    for (index = 0; index < 2048; index++)
    {
        int r, g, b;
        temp = palram[(index << 1) | (1)] << 8 | palram[(index << 1)];

        r = (temp & 0x000F) << 1;
        g = (temp & 0x00F0) >> 2;
        b = (temp & 0x0F00) >> 7;

        if (temp & 0x1000) r |= 1;
        if (temp & 0x2000) g |= 2;
        if (temp & 0x8000) g |= 1;
        if (temp & 0x4000) b |= 1;

        chset->palette[index*4+2] = (r<<3)&0xff;
        chset->palette[index*4+1] = (g<<2)&0xff;
        chset->palette[index*4+0] = (b<<3)&0xff;
        chset->palette[index*4+3] = 0;
    }
}

/* quick load-save handler */

static const int ssendmarker=  0;
static const int sss18ram=     1;
static const int sstileram=    2;
static const int sstextram=    3;
static const int ssobjram=     4;
static const int sspalram=     5;
static const int ssa68kstate=  6;
static const int sssoundlatch= 7;
static const int sssoundcomm=  8;
static const int ssz80prg=     9;
static const int ssz80context=10;
static const int ssym2151=    11;

/*  					   */

void s18cpucore::checkquickstate()
{
	int tag;
	FILE* savestate;
	char ssname[128];

	strcpy(ssname,"save//");
	strcat(ssname,romname);
	strcat(ssname,"qs.sav");

	if (gui->ginput.quicksavestate==1)
	{
		gui->ginput.quicksavestate=0;

		savestate=fopen(ssname,"wb");
		if (savestate==0)
		{
			gui->printMessage(". unable to quicksave state");
			return ;
		}

		tag=sss18ram;
		fwrite(&tag,4,1,savestate);
		fwrite(s18ram,0x4000,1,savestate);

		tag=sstileram;
		fwrite(&tag,4,1,savestate);
		fwrite(tileram,0x10000,1,savestate);

		tag=sstextram;
		fwrite(&tag,4,1,savestate);
		fwrite(textram,0x1000,1,savestate);

		tag=ssobjram;
		fwrite(&tag,4,1,savestate);
		fwrite(objram,0x1000,1,savestate);

		tag=sspalram;
		fwrite(&tag,4,1,savestate);
		fwrite(palram,0x1000,1,savestate);

		tag=ssa68kstate;
		fwrite(&tag,4,1,savestate);
		fwrite(&M68000_regs,sizeof(M68000_regs),1,savestate);

		// end of savestate
		tag=ssendmarker;
		fwrite(&tag,4,1,savestate);

		fclose(savestate);

		gui->printMessage(". quickstate saved");
	}
	if (gui->ginput.quickloadstate==1)
	{
		int exitflag=0;

		gui->ginput.quickloadstate=0;

		savestate=fopen(ssname,"rb");
		if (savestate==0)
		{
			gui->printMessage(". unable to open quicksave");
			return ;
		}

		while(!exitflag)
		{
			int tag;

			fread(&tag,4,1,savestate);

			if (tag==ssendmarker)
			{
				exitflag=1;
			}
			else if (tag==sss18ram)
			{
				fread(s18ram,0x4000,1,savestate);
			}
			else if (tag==sstileram)
			{
				fread(tileram,0x10000,1,savestate);
			}
			else if (tag==sstextram)
			{
				fread(textram,0x1000,1,savestate);
			}
			else if (tag==ssobjram)
			{
				fread(objram,0x1000,1,savestate);
			}
			else if (tag==sspalram)
			{
				fread(palram,0x1000,1,savestate);
			}
			else if (tag==ssa68kstate)
			{
				fread(&M68000_regs,sizeof(M68000_regs),1,savestate);
				M68000_regs.IrqCallback=irqcall;
			}
			else
			{
				exitflag=1;
			}
		}

		fclose(savestate);

		gui->printMessage(". quickstate loaded");
	}
}

/* helper */

void s18cpucore::do_videohw(unsigned long int* video,tCharset* chset)
{
	unsigned short int data=*(unsigned short int*)extram;
	sys16_tile_bank0 = data&0xf;
	sys16_tile_bank1 = (data>>4)&0xf;

	int p1,p2,p3,p4;
	int vreg=0xE80;

	int draw_bg2;
	int draw_fg2;

	int act1,act2;

	act1=textram[0xe85]|(textram[0xe84]<<8);
	act2=textram[0xe87]|(textram[0xe86]<<8);

	p1=textram[vreg+1]|(textram[vreg+0]<<8);
	p2=textram[vreg+3]|(textram[vreg+2]<<8);
	p3=textram[vreg+5]|(textram[vreg+4]<<8);
	p4=textram[vreg+7]|(textram[vreg+6]<<8);

    fg_page[0] = (p1>>12)&0x0f;
    fg_page[1] = (p1>>8)&0x0f;
    fg_page[2] = (p1>>4)&0x0f;
    fg_page[3] = p1&0x0f;

    bg_page[0] = (p2>>12)&0x0f;
    bg_page[1] = (p2>>8)&0x0f;
    bg_page[2] = (p2>>4)&0x0f;
    bg_page[3] = p2&0x0f;

    fg2_page[0] =(p3>>12)&0x0f;
    fg2_page[1] =(p3>>8)&0x0f;
    fg2_page[2] =(p3>>4)&0x0f;
    fg2_page[3] =p3&0x0f;

    bg2_page[0] =(p4>>12)&0x0f;
    bg2_page[1] =(p4>>8)&0x0f;
    bg2_page[2] =(p4>>4)&0x0f;
    bg2_page[3] = p4&0x0f;

    fg_scrollx=((textram[0xE98] | textram[0xE99] << 8));
    bg_scrollx=((textram[0xE9A] | textram[0xE9B] << 8));
    fg2_scrollx=((textram[0xE9c] | textram[0xE9d] << 8));
    bg2_scrollx=((textram[0xE9e] | textram[0xE9f] << 8));

    fg_scrolly=( textram[0xE90] | textram[0xE91] << 8);
    bg_scrolly=( textram[0xE92] | textram[0xE93] << 8);
    fg2_scrolly=((textram[0xE94] | textram[0xE95] << 8));
    bg2_scrolly=((textram[0xE96] | textram[0xE97] << 8));

	if (textram[0x0e85]) draw_fg2=1;
	else draw_fg2=0;
	if (textram[0x0e86]|textram[0x0e87]) draw_bg2=1;
	else draw_bg2=0;

	fg_scrollx= 0x2a0-fg_scrollx&0x3ff;
	bg_scrollx= 0x2a0-bg_scrollx&0x3ff;
	fg2_scrollx=0x2a0-fg2_scrollx&0x3ff;
	bg2_scrollx=0x2a0-bg2_scrollx&0x3ff;

	fg_scrolly =fg_scrolly&0x1ff;
	bg_scrolly =bg_scrolly&0x1ff;
	fg2_scrolly=fg2_scrolly&0x1ff;
	bg2_scrolly=bg2_scrolly&0x1ff;


	// draw layers

	if (draw_bg2)
	{
		draw_bg(video,chset,bg2_scrollx, bg2_scrolly, bg2_page,1);
	}
	else draw_bg(video,chset,bg_scrollx, bg_scrolly, bg_page,1);

	if (draw_fg2)
	{
		draw_bg(video,chset,fg2_scrollx, fg2_scrolly, fg2_page,0);
	}
	else draw_bg(video,chset,fg_scrollx, fg_scrolly, fg_page,0);

	draw_sprites(video,chset,0xc0);
	draw_text(video,chset);


	//////////////
/*
	char scrx[256];
	char scry[256];
	char bg2pg0[256];
	char bg2pg1[256];
	char bg2pg2[256];
	char bg2pg3[256];

	sprintf(scrx,"scrollx %d",bg2_scrollx);
	sprintf(scry,"scrolly %d",bg2_scrolly);

	sprintf(bg2pg0,"page0 %d",bg2_page[0]);
	sprintf(bg2pg1,"page1 %d",bg2_page[1]);
	sprintf(bg2pg2,"page2 %d",bg2_page[2]);
	sprintf(bg2pg3,"page3 %d",bg2_page[3]);

	gui->font8x8->drawString(scrx,0xfefefe,240,120,384,224,video);
	gui->font8x8->drawString(scry,0xfefefe,240,130,384,224,video);

	gui->font8x8->drawString(bg2pg0,0xfefefe,240,140,384,224,video);
	gui->font8x8->drawString(bg2pg1,0xfefefe,240,150,384,224,video);
	gui->font8x8->drawString(bg2pg2,0xfefefe,240,160,384,224,video);
	gui->font8x8->drawString(bg2pg3,0xfefefe,240,170,384,224,video);
*/
}


/* main draw function */

void s18cpucore::drawScreen(unsigned long int* video,tCharset* chset)
{
	// build pal
	build_palette(chset);

	// update controls
	update_controls();

	// video hardware(z)
	do_videohw(video,chset);

	// quick save/load
	checkquickstate();
}
