
/*

  Neo-Geo cpu core implementation
  todo:
	better ym2610 timing
	raster effects

*/

#include <time.h>

#include "neogeocpucore.h"
#include "neogeoromspace.h"
#include "neogeocharset.h"

#include "4990/pd4990a.h"
#include "ym2610/fm.h"
#include "ym2610/ym2610.h"
#include "ym2610/ay8910.h"

// forward declarations

unsigned char neogeo_m68kr8(int address);
unsigned short int neogeo_m68kr16(int address);
unsigned long int neogeo_m68kr32(int address);
void neogeo_m68kw8(int address, unsigned char value);
void neogeo_m68kw16(int address, unsigned short int value);
void neogeo_m68kw32(int address, unsigned long int value);

// variables

extern tGui* gui;

static unsigned char* rom68k;
static unsigned char* ngbios;
static unsigned char* ngram;
static unsigned char* bank4;
static unsigned char* cpubkp;
static unsigned char sram[0x10000];

static unsigned char* z80mem;
static unsigned char z80rom[0x20000];
static unsigned char* bank5;
static unsigned char* bank6;
static unsigned char* bank7;
static unsigned char* bank8;
static char* samples;

static int noint=0;
static int mincyc[2];
static short int* soundbuf[2];
static short int* realsoundbuf;

static unsigned char vidram[0x20000];

static unsigned char* paletteram;
static unsigned char palbank1[0x2000];
static unsigned char palbank2[0x2000];

static int in0=0xff;
static int in1=0xff;
static int in2=0xff;
// test switch - freeze...
static int in3=0x01|0x04|0x38|0x40|0x80;
static int in4=0xff;
 
static int vidoffs=0;
static int vidmodulo=1;
static int fix_bank=0;
static int sram_lock=0;
static int curr_scanline=0;
static int framecount=0;
static int curbank=0;
static int ncbanks;
  
static int pending_command=0;
static int soundlatch=0;
static int result_code=0;
static int ymromsize;

static int nsprites;
static int nchars;

// Zilog 80

static unsigned char ngz80rd(unsigned long int A,struct MemoryReadByte* mwb)
{
	return 0;
}

static void ngz80wr(unsigned long int A,unsigned char V,struct MemoryWriteByte* mwb)
{
}

static UINT16 z80rdp(UINT16 port, struct z80PortRead * s)
{
	int offz;

	switch (port&0xff)
	{
	case 0x00:
		pending_command = 0;
		return soundlatch;
		break;

	case 0x04:
		return YM2610Read(0,0);
		break;

	case 0x05:
		return YM2610Read(0,1);
		break;

	case 0x06:
		return YM2610Read(0,2);
		break;

	case 0x08:
		{
			offz=0x0800*((port>>8)&0x7f);
			memcpy(bank8,&z80rom[offz],0x800);
			return 0;
		}

	case 0x09:
		{
			offz=0x1000*((port>>8)&0x3f);
			memcpy(bank7,&z80rom[offz],0x1000);
			return 0;
		}

	case 0x0a:
		{
			offz=0x2000*((port>>8)&0x1f);
			memcpy(bank6,&z80rom[offz],0x2000);
			return 0;
		}

	case 0x0b:
		{
			offz=0x4000*((port>>8)&0xf);
			memcpy(bank5,&z80rom[offz],0x4000);
			return 0;
		}

	default:
		return 0;
	}

	return 0;
}

static void z80wrp(UINT16 port, UINT8 data, struct z80PortWrite *)
{
	switch (port&0xff)
	{
	case 0x04:
		YM2610Write(0,0,data);
		break;

	case 0x05:
		YM2610Write(0,1,data);
		break;

	case 0x06:
		YM2610Write(0,2,data);
		break;

	case 0x07:
		YM2610Write(0,3,data);
		break;

	case 0x0c:
		result_code = data;
		break;

	default:
		break;
	}
}

static struct z80PortRead ngz80ReadPorts[] =
{
	{0x0,0xffff,z80rdp,0},
	{-1,-1,0}
};

static struct z80PortWrite ngz80WritePorts[] =
{
	{0x0,0xffff,z80wrp,0},
	{-1,-1,0}
};

static struct MemoryReadByte ngz80readmem[] =
{
	{-1,-1,0,0}
};
  
static struct MemoryWriteByte ngz80writemem[] =
{
	{-1,-1,0,0}
};

static void fmirq(int n,int irq)
{
	if (irq!=0)
	{
		mz80int(0xff);
	}
	else
	{
		mz80ClearPendingInterrupt();
	}
}

static void fmtimer(int n,int c,int cnt,double stepTime)
{
	if (cnt!=0)
	{
		if (mincyc[c]==0)
		{
			double cycles=(((double)cnt)*stepTime)*6000000.0f;
			mincyc[c]=(int)cycles;
		}
//		char str[1024];
//		sprintf(str,"resting %d z80 cycles\n",mincyc);
//		MessageBox(0,str,"info",MB_OK);
//		char str[1024];
//		sprintf(str,"timer called with channel [%d], count [%d] and step [%f]\n",c,cnt,stepTime);
	}
	else
	{
		mincyc[c]=0;
	}
}

// a68k

static __inline int irqcall(int i)
{
	return -1;
}

static __inline void changepc(int pc)
{
	if (pc<=0xfffff)
	{
		OP_ROM=rom68k;
		OP_RAM=rom68k;
	}
	else if ( (pc>=0x200000) && (pc<=0x2fffff) )
	{
		OP_ROM=bank4-0x200000;
		OP_RAM=bank4-0x200000;
	}
	else if ( (pc>=0xc00000) && (pc<0xc20000) )
	{
		OP_ROM=ngbios-0xc00000;
		OP_RAM=ngbios-0xc00000;
	}
}

void neogeocpucore::a68kReset()
{
	memset(&M68000_regs,0,sizeof(M68000_regs));

	unsigned int sp,pc;

	sp=READ32(rom68k,0,0);
	pc=READ32(rom68k,4,0);

	M68000_regs.pc=pc;
	changepc(pc);
	M68000_regs.a[7]=sp; 
	M68000_regs.srh=0x27;

	M68000_regs.IrqCallback=irqcall;

    M68000_RESET();
}

neogeocpucore::neogeocpucore(tRomSpace* roms)
{
	neogeoromspace* rsp=(neogeoromspace*)roms;

	// patch mslugx

	if (!strcmp(roms->romname,"mslugx"))
	{
		int i;
		unsigned short int* mem16=(unsigned short int*)rsp->romCPU1;

		for (i = 0;i < (0x100000 >> 1) - 4;i++)
		{
			if (mem16[i+0] == 0x0243 &&
				mem16[i+1] == 0x0001 && 	/* andi.w  #$1, D3 */
				mem16[i+2] == 0x6600)		/* bne xxxx */
			{
				mem16[i+2] = 0x4e71;
				mem16[i+3] = 0x4e71;
			}
		}

		mem16[0x3bdc >> 1] = 0x4e71;
		mem16[0x3bde >> 1] = 0x4e71;
		mem16[0x3be0 >> 1] = 0x4e71;
		mem16[0x3c0c >> 1] = 0x4e71;
		mem16[0x3c0e >> 1] = 0x4e71;
		mem16[0x3c10 >> 1] = 0x4e71;

		mem16[0x3c36 >> 1] = 0x4e71;
		mem16[0x3c38 >> 1] = 0x4e71;
	}

	rom68k=new unsigned char [0x100000];
	memset(rom68k,0,0x100000);
	memcpy(rom68k,rsp->romCPU1,0x100000);

	ngbios=new unsigned char [0x20000];
	memcpy(ngbios,rsp->bios,0x20000);
	ngram=new unsigned char [0x10000];
	memset(ngram,0,0x10000);

	// set zone and console/arcade on the bios rom
	ngbios[0x401]=0x00;	//machine (0x80 arcade - 0x00 console)
	ngbios[0x400]=0x02;	//country

	cpubkp=new unsigned char [0x400000];
	memcpy(cpubkp,rsp->romCPU1+0x100000,0x400000);

	bank4=cpubkp;

	memset(vidram,0,0x20000);
	memset(sram,0x0,0x10000);

	paletteram=palbank2;
	memset(paletteram,0,0x2000);
	paletteram=palbank1;
	memset(paletteram,0,0x2000);

	ncbanks=rsp->ncodebanks;

	// patch some bios code

	unsigned short int* pbios=(unsigned short int*)ngbios;

    pbios[0x11b00>>1]=0x4e71;
    pbios[0x11b02>>1]=0x4e71;
    pbios[0x11b16>>1]=0x4ef9;
    pbios[0x11b18>>1]=0x00c1;
    pbios[0x11b1a>>1]=0x1b6a;

    pbios[0x11c14>>1]=0x4e71;
    pbios[0x11c16>>1]=0x4e71;
    pbios[0x11c1c>>1]=0x4e71;
    pbios[0x11c1e>>1]=0x4e71;

    pbios[0x11c62>>1]=0x4e71;
    pbios[0x11c64>>1]=0x4e71;

	// init pd4990a

	time_t ltime;
	struct tm *today;

	time(&ltime);
	today = localtime(&ltime);

	pd4990a.seconds = ((today->tm_sec/10)<<4) + (today->tm_sec%10);
	pd4990a.minutes = ((today->tm_min/10)<<4) + (today->tm_min%10);
	pd4990a.hours = ((today->tm_hour/10)<<4) + (today->tm_hour%10);
	pd4990a.days = ((today->tm_mday/10)<<4) + (today->tm_mday%10);
	pd4990a.month = (today->tm_mon + 1);
	pd4990a.year = (((today->tm_year%100)/10)<<4) + (today->tm_year%10);
	pd4990a.weekday = today->tm_wday;

	// memory functions

 	a68k_memory_intf.opcode_xor = 0;
 	a68k_memory_intf.read8 = neogeo_m68kr8;
 	a68k_memory_intf.read16 = neogeo_m68kr16;
 	a68k_memory_intf.read32 = neogeo_m68kr32;
 	a68k_memory_intf.write8 = neogeo_m68kw8;
 	a68k_memory_intf.write16 = neogeo_m68kw16;
 	a68k_memory_intf.write32 = neogeo_m68kw32;
 	a68k_memory_intf.changepc = changepc;

 	a68k_memory_intf.read8pc = neogeo_m68kr8;
 	a68k_memory_intf.read16pc = neogeo_m68kr16;
 	a68k_memory_intf.read32pc = neogeo_m68kr32;
 	a68k_memory_intf.read16d = neogeo_m68kr16;
 	a68k_memory_intf.read32d = neogeo_m68kr32;

	a68kReset();

	// init z80

	z80mem=new unsigned char [0x10000];
	memset(z80mem,0,0x10000);

	memcpy(z80rom,rsp->romCPU2,0x20000);
	memcpy(z80mem,rsp->romCPU2,0xf800);

	bank5=&z80mem[0x8000];
	bank6=&z80mem[0xc000];
	bank7=&z80mem[0xe000];
	bank8=&z80mem[0xf000];

	if (!gui->z80init)
	{
		gui->z80init=1;
		mz80init();
	}

	ctx1.z80Base=z80mem;
	ctx1.z80IoRead=ngz80ReadPorts;
	ctx1.z80IoWrite=ngz80WritePorts;
	ctx1.z80MemRead=ngz80readmem;
	ctx1.z80MemWrite=ngz80writemem;

	mz80SetContext(&ctx1);
	mz80reset();

	// ym2610

	samples=new char [rsp->ymromsize];
	memcpy(samples,rsp->ymsamples,rsp->ymromsize);

	soundbuf[0]=new short int [gui->guiinfo.soundfreq/60];
	soundbuf[1]=new short int [gui->guiinfo.soundfreq/60];
	realsoundbuf=new short int [gui->guiinfo.soundfreq*2/60];

	int err;
	void *pcmbufa[1],*pcmbufb[1];
	pcmbufa[0]=0;pcmbufb[0]=0;

	err=YM2610Init(1,8000000,gui->guiinfo.soundfreq,16,pcmbufa,pcmbufb,fmtimer,fmirq);
	if (err!=0)
	{
		MessageBox(0,"error","error ym2610 init",MB_OK);
	}

	Set_YM2610_ADPCM_Buffers(0,0,samples,0,rsp->ymromsize);
	mincyc[0]=mincyc[1]=0;
	ymromsize=rsp->ymromsize;

	// init local variables

	strcpy(romname,roms->romname);
	nsprites=rsp->nspr;
	nchars=rsp->nchars;

	noint=0;
	if ( (!strcmp(romname,"mslug")) ||
		 (!strcmp(romname,"mslug2")) )
	{
		noint=1;
	}
}

neogeocpucore::~neogeocpucore()
{
	delete [] rom68k;
	delete [] cpubkp;
	delete [] z80mem;
	delete [] samples;

	delete [] ngbios;
	delete [] ngram;

	delete [] soundbuf[0];
	delete [] soundbuf[1];
	delete [] realsoundbuf;

	// snd
	YM2610Shutdown();
}
 
void neogeocpucore::emulate(int snd)
{
	if (!snd)
	{
		m68k_ICount=200000;
		M68000_RUN();
		M68000_regs.irq=1;
		pd4990a_addretrace();
		framecount++;
	}
	else
	{
		m68k_ICount=200000;
		M68000_RUN();

		int cyctodo=100000;
		int step=3000;

		while (cyctodo>0)
		{
			if ( (mincyc[0]==0) && (mincyc[1]==0) )
			{
				mz80exec(step);
				cyctodo-=step;
			}
			else if (mincyc[0]==0)
			{
				if (mincyc[1]<=step)
				{
					mz80exec(mincyc[1]);
					cyctodo-=mincyc[1];
					YM2610TimerOver(0,1);
					mincyc[1]=0;
				}
				else 
				{
					mz80exec(step);
					cyctodo-=step;
					mincyc[1]-=step;
				}
			}
			else if (mincyc[1]==0)
			{
				if (mincyc[0]<=step)
				{
					mz80exec(mincyc[0]);
					cyctodo-=mincyc[0];
					YM2610TimerOver(0,0);
					mincyc[0]=0;
				}
				else
				{
					mz80exec(step);
					cyctodo-=step;
					mincyc[0]-=step;
				}
			}
			else
			{
				// both>0
				if (mincyc[0]<mincyc[1])
				{
					if (mincyc[0]<=step)
					{
						mz80exec(mincyc[0]);
						cyctodo-=mincyc[0];
						YM2610TimerOver(0,0);
						mincyc[1]-=mincyc[0];
						mincyc[0]=0;
					}
					else
					{
						mz80exec(step);
						cyctodo-=step;
						mincyc[0]-=step;
						mincyc[1]-=step;
					}
				}
				else
				{
					if (mincyc[1]<=step)
					{
						mz80exec(mincyc[1]);
						cyctodo-=mincyc[1];
						YM2610TimerOver(0,1);
						mincyc[0]-=mincyc[1];
						mincyc[1]=0;
					}
					else
					{
						mz80exec(step);
						cyctodo-=step;
						mincyc[0]-=step;
						mincyc[1]-=step;
					}
				}
			}
		}

		// update soundchip

		int sndbufsize=gui->guiinfo.soundfreq/60;

		YM2610UpdateOne(0,(FMSAMPLE **)soundbuf,sndbufsize);

		signed short int* prsb=realsoundbuf;
		signed short int* psb0=soundbuf[0];
		signed short int* psb1=soundbuf[1];

		for (int i=0;i<sndbufsize;i++)
		{
			*prsb=*psb0;
			prsb++;psb0++;
			*prsb=*psb1;
			prsb++;psb1++;
		}

		feedBuffer(realsoundbuf,sndbufsize*2);

		// irqs

		M68000_regs.irq=1;
		pd4990a_addretrace();
		framecount++;
		if (!noint) mz80int(0);
	}
}

// read-write handlers

static unsigned char  neogeo_m68kr8(int address)
{
	address&=0xffffff;

	if (address<=0x0fffff)
	{
		return READ8(rom68k,address,0);
	}
	else if ( (address>=0x100000) && (address<=0x10ffff) )
	{
		return READ8(ngram,address,0x100000);
	}
	else if ( (address>=0x200000) && (address<=0x2fffff) )
	{
		return READ8(bank4,address,0x200000);
	}
	else if ( (address>=0xc00000) && (address<=0xc1ffff) )
	{
		return READ8(ngbios,address,0xc00000);
	}
	else if ( (address>=0xd00000) && (address<=0xd0ffff) )
	{
		return READ8(sram,address,0xd00000);
	}
	else if (address==0x300000)
	{
		// in0 high
		return in0;
	}
	else if (address==0x300001)
	{
		// test, comm setting, free play
		return in3;
	}
	else if (address==0x320000)
	{
		if (gui->guiinfo.soundon==0)
		{
			return 1;
		}
		else
		{
			int res=0;
			res=result_code;
			if (pending_command) res &= 0x7f;
			return res;
		}
	}
	else if (address==0x320001)
	{
		// in0 low
		int res;
		int coinflip = pd4990a_testbit_r(0);
		int databit = pd4990a_databit_r(0);

		res=in4^(coinflip<<6)^(databit<<7);
		return (res);
	}
	else if (address==0x300080)
	{
		// test switch high
		return 0x00;
	}
	else if (address==0x300081)
	{
		// test switch
		return 0x80;
	}
	else if (address==0x340000) 
	{
		// controller 2
		return in1;
	}
	else if (address==0x340001)
	{
		//return in1 low (0)
		return 0;
	}
	else if (address==0x380000)
	{
		// in2 high byte
		return in2;
	}
	else if (address==0x380001)
	{
		//low
		return 0;
	}
	else if (address==0x3c0006)
	{
		return (0x80);
	}
	else if ( (address>=0x400000) && (address<=0x401fff) )
	{
		return READ8(paletteram,address,0x400000);
	}

	return 0x0;
}
 
static unsigned short int neogeo_m68kr16(int address)
{
	address&=0xffffff;

	if (address<=0x0fffff)
	{
		return READ16(rom68k,address,0);
	}
	else if ( (address>=0x100000) && (address<=0x10ffff) )
	{
		return READ16(ngram,address,0x100000);
	}
	else if ( (address>=0x200000) && (address<=0x2fffff) )
	{
		return READ16(bank4,address,0x200000);
	}
	else if ( (address>=0xc00000) && (address<=0xc1ffff) )
	{
		return READ16(ngbios,address,0xc00000);
	}
	else if ( (address>=0xd00000) && (address<=0xd0ffff) )
	{
		return READ16(sram,address,0xd00000);
	}
	else if ( (address>=0x3c0000) && (address<=0x3c0001) )
	{
		return READ16(vidram,vidoffs&0x1ffff,0);
	}
	else if ( (address>=0x3c000a) && (address<=0x3c000b) )
	{
		return READ16(vidram,vidoffs&0x1ffff,0);
	}
	else if ( (address>=0x3c0004) && (address<=0x3c0005) )
	{
		return (vidmodulo>>1);
	}
	else if (address==0x3c0006)
	{
		return 0x8000;
	}
	else if ( (address>=0x400000) && (address<=0x401fff) )
	{
		return READ16(paletteram,address,0x400000);
	}

	return 0x0;
}

static unsigned long int neogeo_m68kr32(int address)
{
	address&=0xffffff;

	if (address<=0x0fffff)
	{
		return READ32(rom68k,address,0);
	}
	else if ( (address>=0x100000) && (address<=0x10ffff) )
	{
		return READ32(ngram,address,0x100000);
	}
	else if ( (address>=0x200000) && (address<=0x2fffff) )
	{
		return READ32(bank4,address,0x200000);
	}
	else if ( (address>=0x3c0000) && (address<=0x3c0001) )
	{
		// vidram data read
		return READ32(vidram,vidoffs&0x1ffff,0);
	}
	else if ( (address>=0x3c0004) && (address<=0x3c0005) )
	{
		// vidram modulo read
		return (vidmodulo>>1);
	}
	else if ( (address>=0x400000) && (address<=0x401fff) )
	{
		return READ32(paletteram,address,0x400000);
	}
	else if ( (address>=0xc00000) && (address<=0xc1ffff) )
	{
		return READ32(ngbios,address,0xc00000);
	}
	else if ( (address>=0xd00000) && (address<=0xd0ffff) )
	{
		return READ32(sram,address,0xd00000);
	}

	return 0x0;

//	return (((neogeo_m68kr16(address)) << 16) + 
//		    ((neogeo_m68kr16(address + 2))&0xffff));
}

static void neogeo_m68kw8(int address,unsigned char value)
{
	address&=0xffffff;

	if ( (address>=0x100000) && (address<=0x10ffff) )
	{
		WRITE8(ngram,address,0x100000,value);
		return;
	}
	else if (address==0x3a001f)
	{
		paletteram=palbank1;
		return ;
	}
	else if (address==0x3a000f)
	{
		paletteram=palbank2;
		return ;
	}
	else if ( (address>=0xd00000) && (address<=0xd0ffff) )
	{
		if (!sram_lock)
		{
			WRITE8(sram,address,0xd00000,value);
		}
		return;
	}
	else if (address==0x300001)
	{
		// reset?
		return;
	}
	else if (address==0x3a000b)
	{
		fix_bank=1;
		return ;
	}
	else if (address==0x3a001b)
	{
		fix_bank=0;
		return ;
	}
	else if (address==0x3a001d)
	{
		// sram unlock
		sram_lock=0;
		return ;
	}
	else if (address==0x3a000d)
	{
		// sram lock
		sram_lock=1;
		return ;
	}
	else if (address==0x380051)
	{
		pd4990a_control_16_w(value);
		return;
	}
	else if (address==0x320000)
	{
		/* z80 command write */
		soundlatch=value&0xff;
		pending_command=1;
		mz80nmi();
		return;
	}
	else if ( 
		    (address==0x380031) || (address==0x380041) || 
			(address==0x3a0003) || (address==0x380001) ||
			(address==0x3a0013) || (address==0x3a0001) ||
			(address==0x3a0007) || (address==0x3a0015) ||
			(address==0x380021) || (address==0x3800e1) ||
			(address==0x380061) || (address==0x380065) ||
			(address==0x380067)
			)
	{
		// bios calls
		return ;
	}
}

static void neogeo_m68kw16(int address, unsigned short int value)
{
	address&=0xffffff;

	if ( (address>=0x100000) && (address<=0x10ffff) )
	{
		WRITE16(ngram,address,0x100000,value);
		return;
	}
	else if ( (address>=0x3c0000) && (address<=0x3c0001) )
	{
		vidoffs=value*2;
		return;
	}
	else if ( (address>=0x3c0002) && (address<=0x3c0003) )
	{
		WRITE16(vidram,vidoffs&0x1ffff,0,value);
		vidoffs+=vidmodulo;
		return;
	}
	else if ( (address>=0x3c0004) && (address<=0x3c0005) )
	{
		int v=value;
		if (value&0x8000) 
		{
			/* Sign extend it. */
			vidmodulo=((value-0x10000)<<1);
		}
		else
		{
			vidmodulo=(v<<1);
		}
		return;
	}
	else if ( (address>=0x400000) && (address<=0x401fff) )
	{
		// paletteram write
		WRITE16(paletteram,address,0x400000,value);
		return;
	}
	else if ( (address>=0xd00000) && (address<=0xd0ffff) )
	{
		if (!sram_lock)
		{
			WRITE16(sram,address,0xd00000,value);
		}
		return ;
	}
	// bank switch
	else if ( (address>=0x2ffff0) && (address<=0x2fffff) )
	{
		value=value%ncbanks;
		curbank=value;
		bank4=cpubkp+value*0x100000;
		return; 
	}
	else if (address==0x3c000c)
	{
		// irq ack
		return ;
	}
	else if (address==0x3c0006)
	{
		// video reg write
		return ;
	}
	else if ( (address>=0x3c0008) && (address<=0x3c000b) )
	{
		// irq2 pos
		return ;
	}

	address=address;
}

static void neogeo_m68kw32(int address, unsigned long int value)
{
	address&=0xffffff;

	if ( (address>=0x100000) && (address<=0x10ffff) )
	{
		WRITE32(ngram,address,0x100000,value);
		return;
	}
	else if ( (address>=0x400000) && (address<=0x401fff) )
	{
		// paletteram write
		WRITE32(paletteram,address,0x400000,value);
		return;
	}
	else if ( (address>=0xd00000) && (address<=0xd0ffff) )
	{
		if (!sram_lock)
		{
			WRITE32(sram,address,0xd00000,value);
		}
		return ;
	}
	else if ( (address>=0x3c0000) && (address<=0x3c0001) )
	{
		vidoffs=(value>>16)*2;
		WRITE16(vidram,vidoffs&0x1ffff,0,value&0xffff);
		vidoffs+=vidmodulo;
		return;
	}
//	neogeo_m68kw16(address, (value >> 16) &0xffff);
//	neogeo_m68kw16(address + 2, value & 0xffff);
}

// player controls

void neogeocpucore::update_controls()
{
	if (gui->ginput.coin1==1)
	{
		in4=in4&(~1);
	}
	else
	{
		in4=in4|1;
	}

	if (gui->ginput.start1==1)
	{
		in2=in2&(~0x1);
	}
	else
	{
		in2=in2|0x1;
	}

	if (gui->ginput.p1button[0]==1)
	{
		in0=in0&(~0x10);
	}
	else
	{
		in0=in0|0x10;
	}

	if (gui->ginput.p1button[1]==1)
	{
		in0=in0&(~0x20);
	}
	else
	{
		in0=in0|0x20;
	}

	if (gui->ginput.p1button[2]==1)
	{
		in0=in0&(~0x40);
	}
	else
	{
		in0=in0|0x40;
	}

	if (gui->ginput.p1left==1)
	{
		in0=in0&(~0x04);
	}
	else
	{
		in0=in0|0x04;
	}

	if (gui->ginput.p1right==1)
	{
		in0=in0&(~0x08);
	}
	else
	{
		in0=in0|0x08;
	}

	if (gui->ginput.p1up==1)
	{
		in0=in0&(~0x01);
	}
	else
	{
		in0=in0|0x01;
	}

	if (gui->ginput.p1down==1)
	{
		in0=in0&(~0x02);
	}
	else
	{
		in0=in0|0x02;
	}
}

// palette

void neogeocpucore::build_palette(tCharset* chset)
{
    for (int index=0;index<0x1000;index++)
    {
		int temp;

        temp=READ_WORD(&paletteram[index*2]);

		int red=  (temp>>8)&0x0f;
		int green=(temp>>4)&0x0f;
		int blue=  temp    &0x0f;

        chset->palette[index*4+2] = (red*0x11)&0xff;
        chset->palette[index*4+1] = (green*0x11)&0xff;
        chset->palette[index*4+0] = (blue*0x11)&0xff;
        chset->palette[index*4+3] = 0;
    }
}

/* quick load-save handler */

static const int ssendmarker=		0;
static const int ssneogeoram=		1;
static const int ssbank4=			2;
static const int sssram=			3;
static const int sspalram1=			5;
static const int sspalram2=			6;
static const int ssa68kstate=		7;
static const int ssvidram=			8;
static const int ssvidoffs=			9;
static const int ssvidmodulo=		10;
static const int ssfix_bank=		11;
static const int sssram_lock=		12;
static const int ssz80rom=			13;
static const int ssym2610=			14;
static const int ssz80state=		15;
static const int ssmincyc=			16;
static const int sspendingcommand=	17;
static const int sssoundlatch=		18;
static const int ssresultcode=		19;


/*  					   */

void neogeocpucore::checkquickstate()
{
	int tag;
	FILE* savestate;
	char ssname[128];

	strcpy(ssname,"save//");
	strcat(ssname,romname);
	strcat(ssname,"qs.sav");

	if (gui->ginput.quicksavestate==1)
	{
		gui->ginput.quicksavestate=0;

		savestate=fopen(ssname,"wb");
		if (savestate==0)
		{
			gui->printMessage(". unable to quicksave state");
			return ;
		}

		/* main cpu + ram */
		tag=ssneogeoram;
		fwrite(&tag,4,1,savestate);
		fwrite(ngram,0x10000,1,savestate);

		tag=ssbank4;
		fwrite(&tag,4,1,savestate);
		fwrite(&curbank,4,1,savestate);

		tag=sssram;
		fwrite(&tag,4,1,savestate);
		fwrite(sram,0x10000,1,savestate);

		tag=sspalram1;
		fwrite(&tag,4,1,savestate);
		fwrite(palbank1,0x2000,1,savestate);

		tag=sspalram2;
		fwrite(&tag,4,1,savestate);
		fwrite(palbank2,0x2000,1,savestate);

		tag=ssa68kstate;
		fwrite(&tag,4,1,savestate);
		fwrite(&M68000_regs,sizeof(M68000_regs),1,savestate);

		tag=ssvidram;
		fwrite(&tag,4,1,savestate);
		fwrite(vidram,0x20000,1,savestate);

		tag=ssvidoffs;
		fwrite(&tag,4,1,savestate);
		fwrite(&vidoffs,0x4,1,savestate);

		tag=ssvidmodulo;
		fwrite(&tag,4,1,savestate);
		fwrite(&vidmodulo,0x4,1,savestate);

		tag=ssfix_bank;
		fwrite(&tag,4,1,savestate);
		fwrite(&fix_bank,0x4,1,savestate);

		tag=sssram_lock;
		fwrite(&tag,4,1,savestate);
		fwrite(&sram_lock,0x4,1,savestate);

		/* z80 memory + banks + ym2610 */
		tag=ssz80rom;
		fwrite(&tag,4,1,savestate);
		fwrite(z80mem+0x8000,0x8000,1,savestate);

		tag=ssym2610;
		fwrite(&tag,4,1,savestate);
		YM2610savestate(savestate);

		tag=ssz80state;
		fwrite(&tag,4,1,savestate);
		fwrite(&ctx1,sizeof(CONTEXTMZ80),1,savestate);

		tag=ssmincyc;
		fwrite(&tag,4,1,savestate);
		fwrite(mincyc,8,1,savestate);

		tag=sspendingcommand;
		fwrite(&tag,4,1,savestate);
		fwrite(&pending_command,4,1,savestate);

		tag=sssoundlatch;
		fwrite(&tag,4,1,savestate);
		fwrite(&soundlatch,4,1,savestate);

		tag=ssresultcode;
		fwrite(&tag,4,1,savestate);
		fwrite(&result_code,4,1,savestate);

		// end of savestate
		tag=ssendmarker;
		fwrite(&tag,4,1,savestate);

		fclose(savestate);

		gui->printMessage(". quickstate saved");
	}
	if (gui->ginput.quickloadstate==1)
	{
		int exitflag=0;

		gui->ginput.quickloadstate=0;

		savestate=fopen(ssname,"rb");
		if (savestate==0)
		{
			gui->printMessage(". unable to open quicksave");
			return ;
		}

		while(!exitflag)
		{
			int tag;

			fread(&tag,4,1,savestate);

			if (tag==ssendmarker)
			{
				exitflag=1;
			}
			/* main cpu + ram */
			else if (tag==ssneogeoram)
			{
				fread(ngram,0x10000,1,savestate);
			}
			else if (tag==ssbank4)
			{
				fread(&curbank,4,1,savestate);
				bank4=cpubkp+curbank*0x100000;
			}
			else if (tag==sssram)
			{
				fread(sram,0x10000,1,savestate);
			}
			else if (tag==sspalram1)
			{
				fread(palbank1,0x2000,1,savestate);
			}
			else if (tag==sspalram2)
			{
				fread(palbank2,0x2000,1,savestate);
			}
			else if (tag==ssvidram)
			{
				fread(vidram,0x20000,1,savestate);
			}
			else if (tag==ssa68kstate)
			{
				fread(&M68000_regs,sizeof(M68000_regs),1,savestate);
				M68000_regs.IrqCallback=irqcall;
				changepc(M68000_regs.pc);
			}
			else if (tag==ssvidoffs)
			{
				fread(&vidoffs,4,1,savestate);
			}
			else if (tag==ssvidmodulo)
			{
				fread(&vidmodulo,4,1,savestate);
			}
			else if (tag==ssfix_bank)
			{
				fread(&fix_bank,4,1,savestate);
			}
			else if (tag==sssram_lock)
			{
				fread(&sram_lock,4,1,savestate);
			}
			/* z80 memory + banks + ym2610 */
			else if (tag==ssz80rom)
			{
				fread(z80mem+0x8000,0x8000,1,savestate);
			}
			else if (tag==ssym2610)
			{
				YM2610loadstate(savestate);
				Set_YM2610_ADPCM_Buffers(0,0,samples,0,ymromsize);
				YM2610sethandlers(fmtimer,fmirq);
			}
			else if (tag==ssz80state)
			{
				fread(&ctx1,sizeof(CONTEXTMZ80),1,savestate);
				ctx1.z80Base=z80mem;
				ctx1.z80IoRead=ngz80ReadPorts;
				ctx1.z80IoWrite=ngz80WritePorts;
				ctx1.z80MemRead=ngz80readmem;
				ctx1.z80MemWrite=ngz80writemem;
				mz80SetContext(&ctx1);
			}
			else if (tag==ssmincyc)
			{
				fread(mincyc,8,1,savestate);
			}
			else if (tag==sspendingcommand)
			{
				fread(&pending_command,4,1,savestate);
			}
			else if (tag==sssoundlatch)
			{
				fread(&soundlatch,4,1,savestate);
			}
			else if (tag==ssresultcode)
			{
				fread(&result_code,4,1,savestate);
			}
			else
			{
				exitflag=1;
			}
		}

		fclose(savestate);

		gui->printMessage(". quickstate loaded");
	}
}


/* main draw function */

void neogeocpucore::drawScreen(unsigned long int* video,tCharset* chset)
{
	update_controls();
	build_palette(chset);
	checkquickstate();


	// background palette
#define xpad ((384-304)/2)

	unsigned long int* colt=(unsigned long int*)&chset->palette[0xfff*4];
	unsigned long int fillcol=*colt;
	for (int ln=0;ln<224;ln++)
	{
		memset(video+ln*384+xpad,fillcol,(384-2*xpad)*4);
	}

	//
	// draw sprites
	// (this is taken from mame source)
	//

	int count,t1,t2,t3,y,i;
	int offs,tileno,tileatr;
	int sx =0,sy =0,oy =0,my =0,zx = 1, rzy = 1;
    char fullmode=0;
    int ddax=16,dday=256,rzx=15,yskip=0;

	//
	//

	neogeocharset* pchset=(neogeocharset*)chset;

    for (count=0;count<0x300;count+=2) 
	{
		int myxzoom;
		int myyzoom;

		t3 = READ_WORD( &vidram[0x10000 + count] );
		t1 = READ_WORD( &vidram[0x10400 + count] );
		t2 = READ_WORD( &vidram[0x10800 + count] );

		myxzoom=(t3>>8)&0x0f;
		myyzoom=t3&0xff;

		/* If this bit is set this new column is placed next to last one */
		if (t1 & 0x40)
		{
			sx += rzx;
			if ( sx >= 0x1F0 ) sx -= 0x200;

			/* Get new zoom for this column */
			zx = (t3 >> 8) & 0x0f;
			sy = oy;
		}
		else /* nope it is a new block */
		{
			/* Sprite scaling */
			zx = (t3 >> 8) & 0x0f;
			rzy = t3 & 0xff;

			sx = (t2 >> 7);
			if ( sx >= 0x1F0 ) sx -= 0x200;

			/* Number of tiles in this strip */
			my = t1 & 0x3f;
			if (my == 0x20) fullmode = 1;
			else if (my >= 0x21) fullmode = 2;	/* most games use 0x21, but */
												/* Alpha Mission II uses 0x3f */
			else fullmode = 0;

			sy = 0x200 - (t1 >> 7);
//			if (clip->max_y - clip->min_y > 8 ||	
//					clip->min_y == Machine->visible_area.min_y)
			{
				if (sy > 0x110) sy -= 0x200;
				if (fullmode == 2 || (fullmode == 1 && rzy == 0xff))
				{
					while (sy < 0) sy += 2 * (rzy + 1);
				}
			}
			oy = sy;

			if (rzy < 0xff && my < 0x10 && my)
			{
				my = ((my*16*256)/(rzy+1)+15)/16;
				if (my > 0x10) my = 0x10;
			}
			if (my > 0x20) my=0x20;

			ddax=0; /* setup x zoom */
		}

		/* No point doing anything if tile strip is 0 */
		if (my==0) continue;

		/* Process x zoom */
		if (zx != 15)
		{
			rzx = 0;
			for(i = 0; i < 16; i++)
			{
				ddax -= zx+1;
				if (ddax <= 0)
				{
					ddax += 15+1;
//					dda_x_skip[i] = 1;
					rzx++;
				}
//				else dda_x_skip[i] = 0;
			}
		}
		else rzx=16;

		if (sx >= 320) continue;

		/* Setup y zoom */
		if (rzy == 255) yskip=16;
		else dday=0; /* =256; NS990105 mslug fix */

		offs = count<<6;

		/* my holds the number of tiles in each vertical multisprite block */
		for (y=0; y < my ;y++)
		{
			tileno  = READ_WORD(&vidram[offs]);
            offs+=2;
			tileatr = READ_WORD(&vidram[offs]);
            offs+=2;

			if (nsprites>0x10000)
			{
				if (tileatr&0x10) tileno+=0x10000;
			}
			if (nsprites>0x20000)
			{
				if (tileatr&0x20) tileno+=0x20000;
			}
			if (nsprites>0x40000)
			{
				if (tileatr&0x40) tileno+=0x40000;
			}

			if (tileatr&0x8) tileno=(tileno&~7)+((tileno+framecount/4)&7);
			else if (tileatr&0x4) tileno=(tileno&~3)+((tileno+framecount/4)&3);

			if (fullmode == 2 || (fullmode == 1 && rzy == 0xff))
			{
				if (sy >= 248) sy -= 2 * (rzy + 1);
			}
			else if (fullmode == 1)
			{
				if (y == 0x10) sy -= 2 * (rzy + 1);
			}
			else if (sy > 0x110) sy -= 0x200;	/* NS990105 mslug2 fix */

			if (rzy != 255)
			{
				yskip=0;
//				dda_y_skip[0]=0;
				for(i = 0; i < 16; i++)
				{
//					dda_y_skip[i+1] = 0;
					dday -= rzy+1;
					if (dday <= 0)
					{
						dday += 256;
						yskip++;
//						dda_y_skip[yskip]++;
					}
//					else dda_y_skip[yskip]++;
				}
			}

//			if (sy+15 >= clip->min_y && sy <= clip->max_y)
			{
				pchset->printSprite(video,tileno%nsprites,(tileatr>>8)&0xff,
					               32+sx,sy-16,tileatr&0x1,tileatr&0x2,rzx,yskip);
			}

			sy +=yskip;
		}
	}

	//
	// draw fixed charset
	//

	for (y=2;y<30;y++) 
	{
 		for (int x=1;x<39;x++) 
		{
  			int byte1 = (READ_WORD( &vidram[(0xe000+(y*2)+(x*64))&0x1ffff] ));
  			int byte2 = byte1 >> 12;
           	byte1&=0xfff;
			if (nchars>0x1000)
			{
				chset->printChar(video,byte1+fix_bank*0x1000,byte2,32+(x*8),((y-2)*8),0,0);
			}
			else
			{
				chset->printChar(video,byte1,byte2,32+(x*8),((y-2)*8),0,0);
			}
		}
	}
}
