
/*

  System 16 machine inner-workings

*/


#include "s16cpucore.h"
#include "s16charset.h"

#include "ym2151/ym2151.h"

// profiling

//#define PROFILE

#ifdef PROFILE
int srr;
int txtrr;
int tilerr;
int objrr;
int palrr;
#endif

// forward declarations

unsigned char s16_m68kr8(int address);
unsigned short int s16_m68kr16(int address);
unsigned long int s16_m68kr32(int address);
void s16_m68kw8(int address, unsigned char value);
void s16_m68kw16(int address, unsigned short int value);
void s16_m68kw32(int address, unsigned long int value);

// control variables
 
extern tGui* gui;

static unsigned char* rom68k;
static unsigned char s16ram[0x4000];
static unsigned char tileram[0x10000];
static unsigned char textram[0x1000]; 
static unsigned char objram[0x1000];  
static unsigned char palram[0x1000];  

static unsigned short int fg_page[4];
static unsigned short int bg_page[4];
static unsigned short int fg_scrollx, fg_scrolly;
static unsigned short int bg_scrollx, bg_scrolly;

static int gencontrols=0xff;
static int p1controls=0xff;
static int p2controls=0xff;
static int dsw1=0xff;
static int dsw2=0xff;

static unsigned char* z80rom;
static int ym2151_latch=0;
static int sound_command=0;

// z80 

static UINT16 z80rdp(UINT16 port, struct z80PortRead *)
{
    switch (port & 0xFF)
    {
        case 0x01: /* YM2151 status */
			return YM2151ReadStatus(0);

        case 0xC0: /* Sound command latch */
            return (sound_command);
    }

	return 0xff;
}

static void z80wrp(UINT16 port, UINT8 data, struct z80PortWrite *)
{
    switch(port & 0xFF)
    {
        case 0x00: 
            ym2151_latch = (data & 0xFF);            
            return;

        case 0x01: 
			YM2151WriteReg(0,ym2151_latch,data);
            return;

        case 0x40: /* UPD7759 */
            return;

        case 0x80: /* UPD7759 */
            return;
    }
}

// z80 handlers

static struct z80PortRead z80ReadPorts[] =
{
	{0x00,0xffff,z80rdp,0},
	{-1,-1,0}
};

static struct z80PortWrite z80WritePorts[] =
{
	{0x00,0xffff,z80wrp,0},
	{-1,-1,0}
};

static struct MemoryReadByte z80readmem[] =
{
	{-1,-1,0,0}
};
  
static struct MemoryWriteByte z80writemem[] =
{
	{-1,-1,0,0}
};

// a68k

static __inline int irqcall(int i)
{
	return -1;
}
 
void s16cpucore::a68kReset()
{
	memset(&M68000_regs,0,sizeof(M68000_regs));

	unsigned int sp,pc;

	sp=READ32(rom68k,0,0);
	pc=READ32(rom68k,4,0);

	M68000_regs.pc=pc;
	M68000_regs.a[7]=sp; 
	M68000_regs.srh=0x27;

	M68000_regs.IrqCallback=irqcall;

    M68000_RESET();
}

static __inline void changepc(int pc)
{
}

s16cpucore::s16cpucore(tRomSpace* roms)
{
	rom68k=new unsigned char [0x40000];

	memcpy(rom68k,roms->romCPU1,0x40000);
	memset(s16ram,0,0x4000);
	memset(tileram,0,0x10000);
	memset(textram,0,0x1000);
	memset(palram,0,0x1000);
	memset(objram,0,0x1000);

	z80rom=new unsigned char [0x10000];
	memset(z80rom,0,0x10000);
	memcpy(z80rom,roms->romCPU2,0x8000);

	// memory functions

 	a68k_memory_intf.opcode_xor = 0;
 	a68k_memory_intf.read8 = s16_m68kr8;
 	a68k_memory_intf.read16 = s16_m68kr16;
 	a68k_memory_intf.read32 = s16_m68kr32;
 	a68k_memory_intf.write8 = s16_m68kw8;
 	a68k_memory_intf.write16 = s16_m68kw16;
 	a68k_memory_intf.write32 = s16_m68kw32;
 	a68k_memory_intf.changepc = changepc;

 	a68k_memory_intf.read8pc = s16_m68kr8;
 	a68k_memory_intf.read16pc = s16_m68kr16;
 	a68k_memory_intf.read32pc = s16_m68kr32;
 	a68k_memory_intf.read16d = s16_m68kr16;
 	a68k_memory_intf.read32d = s16_m68kr32;

	OP_ROM=rom68k;
	OP_RAM=rom68k;
	
	a68kReset();

	strcpy(romname,roms->romname);

	// profiling

#ifdef PROFILE
	srr=0;
	txtrr=0;
	tilerr=0;
	objrr=0;
	palrr=0;
#endif

	// init z80

	if (!gui->z80init)
	{
		mz80init();
	}

	ctx1.z80Base=z80rom;
	ctx1.z80IoRead=z80ReadPorts;
	ctx1.z80IoWrite=z80WritePorts;
	ctx1.z80MemRead=z80readmem;
	ctx1.z80MemWrite=z80writemem;

	mz80SetContext(&ctx1);
	mz80reset();

	// init sound
	sndbufsize=gui->guiinfo.soundfreq/60;

	sndbuf[0]=new short int [sndbufsize];
	sndbuf[1]=new short int [sndbufsize];

	realsoundbuf=new short int [sndbufsize*2];

	YM2151Init(1,2079580,gui->guiinfo.soundfreq,16);
}

s16cpucore::~s16cpucore()
{
	delete [] rom68k;
	delete [] z80rom;

	delete [] sndbuf[0];
	delete [] sndbuf[1];

	delete [] realsoundbuf;

	YM2151Shutdown();

#ifdef PROFILE
	char str[1024];

	sprintf(str,"srr [%d] txtrr [%d] objrr [%d] palrr [%d] tilerr [%d]\n",
		    srr,txtrr,objrr,palrr,tilerr);
	MessageBox(0,str,"info",MB_OK);
#endif
}
 
void s16cpucore::emulate(int snd)
{
	int dwres;

	if (!snd)
	{
		m68k_ICount=166666;
		M68000_RUN();
		M68000_regs.irq=4;
	}
	else
	{
		for (int j=0;j<2;j++)
		{
			m68k_ICount=166666/2;
			M68000_RUN();

			dwres=mz80exec(68283/2);	
			if (dwres!=0x80000000)
			{
				MessageBox(NULL,"z80 bad op","error",MB_OK);
			}

			YM2151UpdateOne(0,sndbuf,sndbufsize);

			for (int i=0;i<sndbufsize;i++)
			{
				realsoundbuf[i+j*sndbufsize]=sndbuf[0][i];
				realsoundbuf[i+j*sndbufsize+1]=sndbuf[1][i];
			}
		}
		
		feedBuffer(realsoundbuf,sndbufsize*2);
		M68000_regs.irq=4;
	}
}

// player controls

void s16cpucore::update_controls()
{
	if (gui->ginput.coin1==1)
	{
		gencontrols=gencontrols&(~1);
	}
	else
	{
		gencontrols=gencontrols|1;
	}

	if (gui->ginput.start1==1)
	{
		gencontrols=gencontrols&(~0x10);
	}
	else
	{
		gencontrols=gencontrols|0x10;
	}

	if (gui->ginput.p1left==1)
	{
		p1controls=p1controls&(~0x80);
	}
	else
	{
		p1controls=p1controls|0x80;
	}

	if (gui->ginput.p1right==1)
	{
		p1controls=p1controls&(~0x40);
	}
	else
	{
		p1controls=p1controls|0x40;
	}

	if (gui->ginput.p1up==1)
	{
		p1controls=p1controls&(~0x20);
	}
	else
	{
		p1controls=p1controls|0x20;
	}

	if (gui->ginput.p1down==1)
	{
		p1controls=p1controls&(~0x10);
	}
	else
	{
		p1controls=p1controls|0x10;
	}

	if (gui->ginput.p1button[0]==1)
	{
		p1controls=p1controls&(~0x02);
	}
	else
	{
		p1controls=p1controls|0x02;
	}

	if (gui->ginput.p1button[1]==1)
	{
		p1controls=p1controls&(~0x04);
	}
	else
	{
		p1controls=p1controls|0x04;
	}

	if (gui->ginput.p1button[2]==1)
	{
		p1controls=p1controls&(~0x01);
	}
	else
	{
		p1controls=p1controls|0x01;
	}
}

// read-write handlers

static unsigned char  s16_m68kr8(int address)
{
	address&=0xffffff;

	if (address>=0xffc000)
	{
		return READ8(s16ram,address,0xffc000);
	}
	else if (address<=0x3ffff)
	{
		return READ8(rom68k,address,0);
	}
	else if (address==0xc41001)
	{
		//general controlz
		return gencontrols;
	}
	else if (address==0xc41003)
	{
		//p1
		return p1controls;
	}
	else if (address==0xc41007)
	{
		//p2
		return p2controls;
	}
	else if (address==0xc42001)
	{
		// dip switch 1
		return dsw1;
	}
	else if (address==0xc42003)
	{
		// dip switch 2
		return dsw2;
	}

	return 0;
}

static unsigned short int s16_m68kr16(int address)
{
	address&=0xffffff;

	if (address>=0xffc000)
	{
		return READ16(s16ram,address,0xffc000);
	}
	else if (address<=0x3ffff)
	{
		return READ16(rom68k,address,0);
	}

	return 0;
}

static unsigned long int s16_m68kr32(int address)
{
	address&=0xffffff;

	if (address>=0xffc000)
	{
		return READ32(s16ram,address,0xffc000);
	}
	else if (address<=0x3ffff)
	{
		return READ32(rom68k,address,0);
	}

	return 0;
}

static void s16_m68kw8(int address, unsigned char value)
{
	address&=0xffffff;

	if (address>=0xffc000)
	{
		WRITE8(s16ram,address,0xffc000,value);
	}
	else if ( (address>=0x410000) && (address<=0x410fff) )
	{
		WRITE8(textram,address,0x410000,value);
	}
    else if (address==0xfe0007)
	{
        sound_command=(value&0xff);
		mz80int(0x38);
		mz80exec(600);	
	}
}

static void s16_m68kw16(int address, unsigned short int value)
{
	address&=0xffffff;

	if (address>=0xffc000)
	{
		WRITE16(s16ram,address,0xffc000,value);
	}
	else if ( (address>=0x410000) && (address<=0x410fff) )
	{
		WRITE16(textram,address,0x410000,value);
	}
	else if ( (address>=0x400000) && (address<=0x40ffff) )
	{
		WRITE16(tileram,address,0x400000,value);
	}
	else if ( (address>=0x440000) && (address<=0x440fff) )
	{
		WRITE16(objram,address,0x440000,value);
	}
}

static void s16_m68kw32(int address, unsigned long int value)
{
	address&=0xffffff;

	if (address>=0xffc000)
	{
		WRITE32(s16ram,address,0xffc000,value);
	}
	else if ( (address>=0x400000) && (address<=0x40ffff) )
	{
		WRITE32(tileram,address,0x400000,value);
	}
	else if ( (address>=0x840000) && (address<=0x840fff) )
	{
		WRITE32(palram,address,0x840000,value);
	}
	else if ( (address>=0x440000) && (address<=0x440fff) )
	{
		WRITE32(objram,address,0x440000,value);
	}
	else if ( (address>=0x410000) && (address<=0x410fff) )
	{
		WRITE32(textram,address,0x410000,value);
	}
}

/* draw helpers */

void s16cpucore::draw_sprites(unsigned long int* video,tCharset* chset,int pri)
{
/*
   440000-4403FF: The sprites registers. There are 16 control bytes for each
                  sprite (64 sprites maximum). This is the function of the 16
                  bytes.

   00 : last sprite line on the screen
   01 : first sprites line on the screen (sprites height = [00]-[01])
   02 : high value of the sprite X position
   03 : low value of the sprite X position (X = [02]*256+[03])
   04 : horizontal sprite flipping flag (flip if value = 01)
   05 : sprite width (real width = [05]*2) (!! signed byte)
        if the bit7 is set to 1, the sprite is flipped vertically
   06 : high value of sprite rom position 
   07 : low value of sprite rom position
   08 : sprite bank selector
        bit7-bit4 are always at 1 if the sprite is active
        bit3-bit0 is the sprite bank number (0-F)
   09 : sprites priority and base color register selector
        bit7-bit6: 10 - foreground sprite
                   01 - background sprite
                   11 - sprite over everything
        bit5-bit0: base color registers (is 1024+16*value)
   0A : zoom function \
   0B : zoom function - unknown format (never used in Shinobi)
   0C to 0F are not used in current emulated games
*/

	int name;
    int count;
	int bank[16] = { 0, 0, 0, 0, 0, 0, 0, 6, 0, 0, 0, 4, 0, 2, 0, 0 };
	s16charset* pchs=(s16charset*)chset;

    for (count=0;count<0x80;count++)
    {
        unsigned char* p=&objram[count<<3];
        int start = p[0];		// first sprite line on screen
        int end = p[1];			// last sprite line on screen
		int priority=(p[8]&0xc0);

        if ( (start==0xff) && (end==0xff) ) return ;

		if (priority==pri)
		{
/*
   09 : sprites priority and base color register selector
        bit7-bit6: 10 - foreground sprite
                   01 - background sprite
                   11 - sprite over everything
        bit5-bit0: base color registers (is 1024+16*value)
*/

			int pal=(p[8]&0x3f)<<4;
     
			// xposition of the sprite
			// p[2] -> low byte
			// p[3] -> high byte

			int xpos=(p[3]<<8|p[2])-(160+24);
			int width=p[4]&0x7f;
			int height=(end-start);
			int flipx=p[5]&1;
			int flipy=p[4]&0x80;

			// p[4] -> vertical flip

			if (flipy)
			{
				width=0x80-width;
			}
			width*=4;

			// sprite code
			// p[6] -> low byte
			// p[7] -> high byte

			name = 4*( ((p[7]<<8|p[6])) | 
					 ((bank[p[9]&0xf])*0x8000) );

			if (p[9]&0xf0)		// sprite active
			{
				pchs->printSprite(video,name,pal,
					              xpos,start,width,height,
								  flipx,flipy,0,0);
			}
		}
    }
}

void s16cpucore::draw_bg(unsigned long int* video,tCharset* chset,int scrollx,int scrolly,unsigned short int *page)
{
    int xx, yy;
    int sx,sy,px,py,pg=0;
    int attr;
    int name;
    int pal;
    int ce=(scrollx&7)?41:40;

	s16charset* pch=(s16charset*)chset;

    for (yy=0;yy<(224+8);yy+=8)
    {
        sy = ((scrolly + yy) >> 3) & 0x3F;
        py = (sy & 0x1F);
        pg = (pg & 0x01) | ((sy >> 5) & 1) << 1;

        for (xx=0;xx<ce;xx++)
        {
            sx  = (24 + (scrollx >> 3) + xx);
            px  = (sx) & 0x3f;
            pg  = (pg & 0x02) | ((sx >> 5) & 1);

			// tile attribute
            attr = *(unsigned short int *)&tileram[(page[pg]) | (py << 7) | (px << 1)];

            // tile 
            name = (attr & 0x0FFF) | (((attr >> 12) & 1) << 12);

            // palette
            pal  = ((attr >> 6) & 0x7F) << 3;

			int sx2=( (((384-320)/2) - (scrollx & 7)) + (xx << 3) ) ;
			pch->printfgChar(video,name,pal,sx2,yy-(scrolly&7),0,0);
        }
    }
}

void s16cpucore::draw_text(unsigned long int* video,tCharset* chset)
{
    int xx,yy,name,cc;
	s16charset* pch=(s16charset*)chset;

    for (yy=0;yy<28;yy++)
    {
        for (xx=0;xx<40;xx++)
        {
            name = *(unsigned short int*)&textram[((yy << 6) + (xx + 24)) << 1];
            cc = (name >> 6) & 0x38;
            name = (name & 0x1FF);

			int sx=xx*8+((384-320)/2);
			int sy=(yy*8);
			pch->printfgChar(video,name,cc,sx,sy,0,0);
        }
    }
}

/* palette build */

void s16cpucore::build_palette(tCharset* chset)
{
    int index;
    unsigned short int temp;

    for (index = 0; index < 2048; index++)
    {
        int r, g, b;
        temp = palram[(index << 1) | (1)] << 8 | palram[(index << 1)];

        r = (temp & 0x000F) << 1;
        g = (temp & 0x00F0) >> 2;
        b = (temp & 0x0F00) >> 7;

        if(temp & 0x1000) r |= 1;
        if(temp & 0x4000) g |= 2;
        if(temp & 0x8000) g |= 1;
        if(temp & 0x2000) b |= 1;

        chset->palette[index*4+2] = (r<<3)&0xff;
        chset->palette[index*4+1] = (g<<2)&0xff;
        chset->palette[index*4+0] = (b<<3)&0xff;
        chset->palette[index*4+3] = 0;
    }
}

/* quick load-save handler */

static const int ssendmarker=  0;
static const int sss16ram=     1;
static const int sstileram=    2;
static const int sstextram=    3;
static const int ssobjram=     4;
static const int sspalram=     5;
static const int ssa68kstate=  6;
static const int sssoundlatch= 7;
static const int sssoundcomm=  8;
static const int ssz80prg=     9;
static const int ssz80context=10;
static const int ssym2151=    11;

/*  					   */

void s16cpucore::checkquickstate()
{
	int tag;
	FILE* savestate;
	char ssname[128];

	strcpy(ssname,"save//");
	strcat(ssname,romname);
	strcat(ssname,"qs.sav");

	if (gui->ginput.quicksavestate==1)
	{
		gui->ginput.quicksavestate=0;

		savestate=fopen(ssname,"wb");
		if (savestate==0)
		{
			gui->printMessage(". unable to quicksave state");
			return ;
		}

		tag=sss16ram;
		fwrite(&tag,4,1,savestate);
		fwrite(s16ram,0x4000,1,savestate);

		tag=sstileram;
		fwrite(&tag,4,1,savestate);
		fwrite(tileram,0x10000,1,savestate);

		tag=sstextram;
		fwrite(&tag,4,1,savestate);
		fwrite(textram,0x1000,1,savestate);

		tag=ssobjram;
		fwrite(&tag,4,1,savestate);
		fwrite(objram,0x1000,1,savestate);

		tag=sspalram;
		fwrite(&tag,4,1,savestate);
		fwrite(palram,0x1000,1,savestate);

		tag=ssa68kstate;
		fwrite(&tag,4,1,savestate);
		fwrite(&M68000_regs,sizeof(M68000_regs),1,savestate);

		tag=sssoundlatch;
		fwrite(&tag,4,1,savestate);
		fwrite(&ym2151_latch,4,1,savestate);

		tag=sssoundcomm;
		fwrite(&tag,4,1,savestate);
		fwrite(&sound_command,4,1,savestate);

		tag=ssz80prg;
		fwrite(&tag,4,1,savestate);
		fwrite(&z80rom[0x8000],0x8000,1,savestate);

		tag=ssz80context;
		fwrite(&tag,4,1,savestate);
		fwrite(&ctx1,sizeof(CONTEXTMZ80),1,savestate);

		tag=ssym2151;
		fwrite(&tag,4,1,savestate);
		YM2151savestatus(savestate);

		// end of savestate
		tag=ssendmarker;
		fwrite(&tag,4,1,savestate);

		fclose(savestate);

		gui->printMessage(". quickstate saved");
	}
	if (gui->ginput.quickloadstate==1)
	{
		int exitflag=0;

		gui->ginput.quickloadstate=0;

		savestate=fopen(ssname,"rb");
		if (savestate==0)
		{
			gui->printMessage(". unable to open quicksave");
			return ;
		}

		while(!exitflag)
		{
			int tag;

			fread(&tag,4,1,savestate);

			if (tag==ssendmarker)
			{
				exitflag=1;
			}
			else if (tag==sss16ram)
			{
				fread(s16ram,0x4000,1,savestate);
			}
			else if (tag==sstileram)
			{
				fread(tileram,0x10000,1,savestate);
			}
			else if (tag==sstextram)
			{
				fread(textram,0x1000,1,savestate);
			}
			else if (tag==ssobjram)
			{
				fread(objram,0x1000,1,savestate);
			}
			else if (tag==sspalram)
			{
				fread(palram,0x1000,1,savestate);
			}
			else if (tag==ssa68kstate)
			{
				fread(&M68000_regs,sizeof(M68000_regs),1,savestate);
				M68000_regs.IrqCallback=irqcall;
			}
			else if (tag==sssoundlatch)
			{
				fread(&ym2151_latch,4,1,savestate);
			}
			else if (tag==sssoundcomm)
			{
				fread(&sound_command,4,1,savestate);
			}
			else if (tag==ssz80prg)
			{
				fread(&z80rom[0x8000],0x8000,1,savestate);
			}
			else if (tag==ssz80context)
			{
				// it is important that ctx is read *after* z80prg
				fread(&ctx1,sizeof(CONTEXTMZ80),1,savestate);
				ctx1.z80Base=z80rom;
				ctx1.z80IoRead=z80ReadPorts;
				ctx1.z80IoWrite=z80WritePorts;
				ctx1.z80MemRead=z80readmem;
				ctx1.z80MemWrite=z80writemem;
			}
			else if (tag==ssym2151)
			{
				YM2151loadstatus(savestate);
			}
			else
			{
				exitflag=1;
			}
		}

		fclose(savestate);

		gui->printMessage(". quickstate loaded");
	}
}


/* main draw function */

void s16cpucore::drawScreen(unsigned long int* video,tCharset* chset)
{
	// build pal
	build_palette(chset);

	// update controls
	update_controls();

	// calculate page addresses

    fg_page[0] = ((textram[0xE80] >> 4) & 0x0F) << 12;
    fg_page[1] = ((textram[0xE80] >> 0) & 0x0F) << 12;
    fg_page[2] = ((textram[0xE81] >> 4) & 0x0F) << 12;
    fg_page[3] = ((textram[0xE81] >> 0) & 0x0F) << 12;

    bg_page[0] = ((textram[0xE82] >> 4) & 0x0F) << 12;
    bg_page[1] = ((textram[0xE82] >> 0) & 0x0F) << 12;
    bg_page[2] = ((textram[0xE83] >> 4) & 0x0F) << 12;
    bg_page[3] = ((textram[0xE83] >> 0) & 0x0F) << 12;

    fg_scrollx = 0x400-((textram[0xE98] | textram[0xE99] << 8) & 0x3FF);
    bg_scrollx = 0x400-((textram[0xE9A] | textram[0xE9B] << 8) & 0x3FF);
    fg_scrolly = (textram[0xE90] | textram[0xE91] << 8) & 0x1FF;
    bg_scrolly = (textram[0xE92] | textram[0xE93] << 8) & 0x1FF;

	// draw layers

	draw_bg(video,chset,bg_scrollx, bg_scrolly, bg_page);
	draw_sprites(video,chset,0x40);
	draw_bg(video,chset,fg_scrollx, fg_scrolly, fg_page);
	draw_sprites(video,chset,0x80);
	draw_text(video,chset);
	draw_sprites(video,chset,0xc0);

	// quick save/load
	checkquickstate();
}
