/*

  gui
  the gui class

*/


#ifndef GUI_H
#define GUI_H

#include <windows.h>
#include <vector>

#include "romspace.h"
#include "charset.h"
#include "component.h"
#include "message.h"


// the class
// gui is the glue between different parts
// of the emulation system
// everything should interact with the gui
// and the gui should interact with everything

struct iniInfo
{
	char rompath[1024];
	int showfps;
	int ngarcade;

	int soundon;
	int soundfreq;
	int interpolate;
	int usedsound;

	int autoframeskip;
	int frameskip;
	int videomode;
	int trimspeed;
	int usesysmemory;
};


class cpucore;

class tGui
{
private:

	unsigned long int* rotobuf;
	std::vector<tComponent*> compList;
	tMessage* msgs;
	int action;

	int state;			// is gui active?
	int guiXsize;
	int guiYsize;

	int openRom(char*);
	int ltrim(char* str);
	int rtrim(char* str);
	int readIniFile(char* inifile);
	int getParameter(char* filename,char* sectionname,
	                       char* paramname,char* paramvalue);

public:

	tFont* font8x8;
	unsigned long int* videoBuf;
	int emulating;
	int romsettype;
	tRomSpace* romset;
	tCharset* chset;
	cpucore* machine;
	guiInput ginput;
	iniInfo guiinfo;
	int frames;
	int orientation;
	int changingmode;
	int exitp;
	int initfailed;
	float onscreenfps;

	tGui(int s,int gxs,int gys,char* inifile);
	~tGui();
	void redraw(unsigned long int* vbuf);
	void update();		// reparse input
	int exitprogram();
	void printMessage(char*);

	void mouseMove(int x,int y);
	void lButtonDown();
	void lButtonUp();
	void rButtonDown();
	void rButtonUp();
};

#endif